# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/management/v1/management_types.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from continual.rpc.graphql import options_pb2 as continual_dot_rpc_dot_graphql_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/management/v1/management_types.proto',
  package='continual.rpc.management.v1',
  syntax='proto3',
  serialized_options=b'ZIgithub.com/continualiq/continual/continual/rpc/management/v1;managementv1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n2continual/rpc/management/v1/management_types.proto\x12\x1b\x63ontinual.rpc.management.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a#continual/rpc/graphql/options.proto\":\n\x14IngestionErrorDetail\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x02 \x01(\x05\"\xda\x03\n\tIngestion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12;\n\x05state\x18\x02 \x01(\x0e\x32,.continual.rpc.management.v1.Ingestion.State\x12\x1a\n\x12ingested_row_count\x18\x03 \x01(\x03\x12\x16\n\x0eingested_bytes\x18\x04 \x01(\x03\x12H\n\rerror_details\x18\x05 \x03(\x0b\x32\x31.continual.rpc.management.v1.IngestionErrorDetail\x12/\n\x0b\x63reate_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x66inish_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"r\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\x0e\n\nCANCELLING\x10\x03\x12\r\n\tSUCCEEDED\x10\x04\x12\n\n\x06\x46\x41ILED\x10\x05\x12\r\n\tCANCELLED\x10\x06\"\xe0\x02\n\x04User\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x16\n\x0e\x65mail_verified\x18\x03 \x01(\x08\x12\x11\n\tfull_name\x18\x04 \x01(\t\x12\x0b\n\x03\x62io\x18\x05 \x01(\t\x12\x10\n\x08location\x18\x06 \x01(\t\x12\x15\n\x08password\x18\x07 \x01(\tB\x03\xb8\x45\x00\x12\x1c\n\x0fservice_account\x18\x08 \x01(\x08\x42\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x10\n\x08\x64isabled\x18\x0b \x01(\x08\x12\x17\n\x0ftrial_available\x18\x0c \x01(\x08\x12\x12\n\nfirst_name\x18\r \x01(\t\x12\x11\n\tlast_name\x18\x0e \x01(\t\"\xf5\x01\n\x07Session\x12\x11\n\x04name\x18\x01 \x01(\tB\x03\xb8\x45\x01\x12\x14\n\x07subject\x18\x02 \x01(\tB\x03\xb8\x45\x01\x12\x1b\n\x0elast_access_ip\x18\x03 \x01(\tB\x03\xb8\x45\x01\x12\x39\n\x10last_access_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x15\n\x08login_ip\x18\x05 \x01(\tB\x03\xb8\x45\x01\x12\x33\n\nlogin_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1d\n\x10login_user_agent\x18\x07 \x01(\tB\x03\xb8\x45\x01\"d\n\x04Role\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x15\n\rresource_type\x18\x04 \x01(\t\x12\x13\n\x0bpermissions\x18\x05 \x03(\t\"\x90\x01\n\x0c\x41\x63\x63\x65ssPolicy\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x07subject\x18\x02 \x01(\tB\x07\xc2\x45\x04User\x12\x10\n\x08resource\x18\x03 \x01(\t\x12\x15\n\x04role\x18\x04 \x01(\tB\x07\xc2\x45\x04Role\x12/\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xc4\x01\n\x06Invite\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x06sender\x18\x02 \x01(\tB\n\xc2\x45\x04User\xb8\x45\x01\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x15\n\x04role\x18\x04 \x01(\tB\x07\xc2\x45\x04Role\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x65xpire_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\"\xb1\x03\n\x0cOrganization\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x45\n\x06status\x18\x06 \x01(\x0e\x32\x30.continual.rpc.management.v1.Organization.StatusB\x03\xb8\x45\x01\x12\x37\n\x0etrial_end_date\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1a\n\rtrial_credits\x18\x08 \x01(\x03\x42\x03\xb8\x45\x01\x12\x1f\n\x12trial_credits_used\x18\t \x01(\x03\x42\x03\xb8\x45\x01\x12\x11\n\tshow_plan\x18\n \x01(\x08\"A\n\x06Status\x12\x0b\n\x07\x45XPIRED\x10\x00\x12\x10\n\x0c\x41\x43TIVE_TRIAL\x10\x01\x12\n\n\x06\x41\x43TIVE\x10\x02\x12\x0c\n\x08\x44ISABLED\x10\x03\"\xfd\x01\n\nOnboarding\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x34\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x17\n\x0fproject_created\x18\x04 \x01(\x08\x12\x15\n\rmodel_created\x18\x05 \x01(\x08\x12\x1b\n\x13\x66\x65\x61ture_set_created\x18\x06 \x01(\x08\x12\x15\n\rbilling_setup\x18\x07 \x01(\x08\x12\x11\n\tdismissed\x18\x08 \x01(\x08\"\xbf\x02\n\x07Project\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12%\n\x0corganization\x18\x03 \x01(\tB\x0f\xc2\x45\x0cOrganization\x12\x41\n\x07summary\x18\x04 \x01(\x0b\x32+.continual.rpc.management.v1.ProjectSummaryB\x03\xb8\x45\x01\x12:\n\ndata_store\x18\x05 \x01(\x0b\x32&.continual.rpc.management.v1.DataStore\x12\x34\n\x0bupdate_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\"\xaf\x04\n\x0eProjectSummary\x12\x19\n\x11\x66\x65\x61ture_set_count\x18\x01 \x01(\x05\x12\x1d\n\x15\x66\x65\x61ture_set_row_count\x18\x02 \x01(\x03\x12\x19\n\x11\x66\x65\x61ture_set_bytes\x18\x03 \x01(\x03\x12\x15\n\rfeature_count\x18\x04 \x01(\x05\x12\x18\n\x10\x63onnection_count\x18\x05 \x01(\x05\x12\x13\n\x0bmodel_count\x18\x06 \x01(\x05\x12\x1b\n\x13model_version_count\x18\x07 \x01(\x05\x12\x18\n\x10\x65xperiment_count\x18\x08 \x01(\x03\x12\x18\n\x10prediction_count\x18\t \x01(\x03\x12\x45\n\x12\x66\x65\x61ture_set_health\x18\n \x01(\x0b\x32).continual.rpc.management.v1.HealthCounts\x12?\n\x0cmodel_health\x18\x0b \x01(\x0b\x32).continual.rpc.management.v1.HealthCounts\x12\x37\n\x13last_ingestion_time\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x38\n\x14last_deployment_time\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x36\n\x12last_training_time\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xc4\x0b\n\tDataStore\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x44\n\tbig_query\x18\x02 \x01(\x0b\x32/.continual.rpc.management.v1.DataStore.BigQueryH\x00\x12\x45\n\tsnowflake\x18\x03 \x01(\x0b\x32\x30.continual.rpc.management.v1.DataStore.SnowflakeH\x00\x12\x43\n\x08postgres\x18\x04 \x01(\x0b\x32/.continual.rpc.management.v1.DataStore.PostgresH\x00\x12\x43\n\x08redshift\x18\x05 \x01(\x0b\x32/.continual.rpc.management.v1.DataStore.RedshiftH\x00\x12G\n\ndatabricks\x18\x07 \x01(\x0b\x32\x31.continual.rpc.management.v1.DataStore.DatabricksH\x00\x12\x1a\n\x12is_demo_data_store\x18\x06 \x01(\x08\x1a\x46\n\x08\x42igQuery\x12\x11\n\tauth_file\x18\x01 \x01(\t\x12\x16\n\x0e\x61uth_file_name\x18\x02 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x03 \x01(\t\x1a\x8e\x02\n\tSnowflake\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12\x11\n\twarehouse\x18\x06 \x01(\t\x12\x0c\n\x04role\x18\x07 \x01(\t\x12L\n\x06params\x18\x08 \x03(\x0b\x32<.continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry\x12\x11\n\tdb_schema\x18\t \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xeb\x01\n\x08Postgres\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12K\n\x06params\x18\x06 \x03(\x0b\x32;.continual.rpc.management.v1.DataStore.Postgres.ParamsEntry\x12\x11\n\tdb_schema\x18\x07 \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xeb\x01\n\x08Redshift\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12K\n\x06params\x18\x06 \x03(\x0b\x32;.continual.rpc.management.v1.DataStore.Redshift.ParamsEntry\x12\x11\n\tdb_schema\x18\x07 \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xed\x01\n\nDatabricks\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\r\n\x05token\x18\x03 \x01(\t\x12\x11\n\thttp_path\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12M\n\x06params\x18\x06 \x03(\x0b\x32=.continual.rpc.management.v1.DataStore.Databricks.ParamsEntry\x12\x11\n\tdb_schema\x18\x07 \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x08\n\x06\x63onfig\"V\n\x0cHealthCounts\x12\x15\n\rhealthy_count\x18\x01 \x01(\x05\x12\x17\n\x0funhealthy_count\x18\x02 \x01(\x05\x12\x16\n\x0e\x63ritical_count\x18\x03 \x01(\x05\"\xb2\x02\n\x0b\x45nvironment\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x41\n\x07summary\x18\x02 \x01(\x0b\x32+.continual.rpc.management.v1.ProjectSummaryB\x03\xb8\x45\x01\x12:\n\ndata_store\x18\x03 \x01(\x0b\x32&.continual.rpc.management.v1.DataStore\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1a\n\x12scheduling_enabled\x18\x06 \x01(\x08\x12\x0e\n\x06source\x18\x07 \x01(\t\"\xe9\x02\n\x10\x46\x65\x61tureSetSchema\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06\x65ntity\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x0e\n\x06owners\x18\x06 \x03(\t\x12\x15\n\rdocumentation\x18\x07 \x01(\t\x12;\n\x07profile\x18\x08 \x01(\x0b\x32*.continual.rpc.management.v1.ProfileConfig\x12:\n\x07\x63olumns\x18\t \x03(\x0b\x32).continual.rpc.management.v1.ColumnConfig\x12\x17\n\x0f\x65xclude_columns\x18\n \x03(\t\x12\r\n\x05query\x18\x0b \x01(\t\x12\r\n\x05index\x18\x0c \x01(\t\x12\x12\n\ntime_index\x18\r \x01(\t\x12\r\n\x05split\x18\x0e \x01(\t\x12\r\n\x05table\x18\x0f \x01(\t\"\xb5\x04\n\x0bModelSchema\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x0e\n\x06owners\x18\x06 \x03(\t\x12\x15\n\rdocumentation\x18\x07 \x01(\t\x12\x0e\n\x06target\x18\x08 \x01(\t\x12\x37\n\x05train\x18\t \x01(\x0b\x32(.continual.rpc.management.v1.TrainConfig\x12;\n\x07predict\x18\n \x01(\x0b\x32*.continual.rpc.management.v1.PredictConfig\x12;\n\x07promote\x18\x0b \x01(\x0b\x32*.continual.rpc.management.v1.PromoteConfig\x12;\n\x07profile\x18\x0c \x01(\x0b\x32*.continual.rpc.management.v1.ProfileConfig\x12:\n\x07\x63olumns\x18\r \x03(\x0b\x32).continual.rpc.management.v1.ColumnConfig\x12\x17\n\x0f\x65xclude_columns\x18\x0e \x03(\t\x12\x1c\n\x14include_feature_sets\x18\x0f \x03(\t\x12\r\n\x05query\x18\x10 \x01(\t\x12\r\n\x05index\x18\x11 \x01(\t\x12\x12\n\ntime_index\x18\x12 \x01(\t\x12\r\n\x05split\x18\x13 \x01(\t\x12\r\n\x05table\x18\x14 \x01(\t\"\x86\x01\n\x0c\x43olumnConfig\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0e\n\x06\x65ntity\x18\x03 \x01(\t\x12\x34\n\x04type\x18\x04 \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\x12\r\n\x05\x64type\x18\x05 \x01(\t\"\x8f\x02\n\x0bTrainConfig\x12\x10\n\x08schedule\x18\x01 \x01(\t\x12\x0e\n\x06metric\x18\x02 \x01(\t\x12\x14\n\x0coptimization\x18\x03 \x01(\t\x12\x10\n\x08\x61\x64vanced\x18\x04 \x01(\t\x12\x1c\n\x14included_model_types\x18\x05 \x03(\t\x12\x1c\n\x14\x65xcluded_model_types\x18\x06 \x03(\t\x12\x0c\n\x04size\x18\x07 \x01(\t\x12\r\n\x05plots\x18\x08 \x03(\t\x12\x18\n\x10\x65xclude_ensemble\x18\t \x01(\x08\x12\x1b\n\x13\x64isable_data_checks\x18\n \x01(\x08\x12\x12\n\ntime_limit\x18\x0b \x01(\x03\x12\x12\n\nsize_limit\x18\x0c \x01(\x03\"6\n\rPredictConfig\x12\x10\n\x08schedule\x18\x01 \x01(\t\x12\x13\n\x0bincremental\x18\x02 \x01(\x08\"\x1f\n\rPromoteConfig\x12\x0e\n\x06policy\x18\x01 \x01(\t\"!\n\rProfileConfig\x12\x10\n\x08schedule\x18\x01 \x01(\t\"m\n\x0cSourceConfig\x12\x12\n\nconnection\x18\x01 \x01(\t\x12\r\n\x05query\x18\x02 \x01(\t\x12\x10\n\x08schedule\x18\x03 \x01(\t\x12\x0c\n\x04\x66ile\x18\x04 \x01(\t\x12\x1a\n\x12incremental_column\x18\x05 \x01(\t\"\xe4\x01\n\nFeatureSet\x12\x0c\n\x04name\x18\x01 \x01(\t\x12=\n\x06schema\x18\x02 \x01(\x0b\x32-.continual.rpc.management.v1.FeatureSetSchema\x12\x13\n\x0bschema_text\x18\x03 \x01(\t\x12/\n\x0b\x63reate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\x05table\x18\n \x01(\tB\x03\xb8\x45\x01\"\xca\x02\n\x0b\x43olumnStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x01\x12\x0b\n\x03min\x18\x03 \x01(\x01\x12\x0b\n\x03max\x18\x04 \x01(\x01\x12\n\n\x02sd\x18\x05 \x01(\x01\x12\x0c\n\x04mean\x18\x06 \x01(\x01\x12\r\n\x05\x63ount\x18\x07 \x01(\x03\x12\x12\n\nnull_count\x18\x08 \x01(\x03\x12\x12\n\nmin_length\x18\t \x01(\x03\x12\x12\n\nmax_length\x18\n \x01(\x03\x12,\n\x08\x65\x61rliest\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12*\n\x06latest\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x34\n\x04type\x18\r \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\"\xa8\x02\n\x0e\x43olumnSetStats\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\x0b\x63omputed_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x11\n\trow_count\x18\x03 \x01(\x03\x12\x13\n\x0btotal_bytes\x18\x04 \x01(\x03\x12\x39\n\x07\x63olumns\x18\x05 \x03(\x0b\x32(.continual.rpc.management.v1.ColumnStats\x12\x16\n\x0ehas_time_index\x18\x06 \x01(\x08\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x8d\x02\n\x05\x45vent\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x07subject\x18\x02 \x01(\tB\x07\xc2\x45\x04User\x12\r\n\x05\x65vent\x18\x03 \x01(\t\x12\x10\n\x08resource\x18\x04 \x01(\t\x12\x10\n\x08severity\x18\x05 \x01(\t\x12\x0f\n\x07message\x18\x06 \x01(\t\x12:\n\x04\x64\x61ta\x18\x07 \x03(\x0b\x32,.continual.rpc.management.v1.Event.DataEntry\x12/\n\x0b\x63reate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a+\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xef\x03\n\rServingMetric\x12\r\n\x05model\x18\x01 \x01(\t\x12\x0e\n\x06parent\x18\x02 \x01(\t\x12\x0b\n\x03tag\x18\x03 \x01(\t\x12\x36\n\x12request_start_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x34\n\x10model_start_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x33\n\x10request_duration\x18\x06 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x31\n\x0eroute_duration\x18\x07 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x33\n\x10\x66\x65\x61ture_duration\x18\x08 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x33\n\x10predict_duration\x18\t \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x31\n\x0emodel_duration\x18\n \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x15\n\rmodel_version\x18\x0b \x01(\t\x12\x13\n\x0bresult_code\x18\x0c \x01(\t\x12\x13\n\x0bresult_data\x18\r \x01(\t\"\xdd\x03\n\x05Model\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x37\n\x05state\x18\x02 \x01(\x0e\x32(.continual.rpc.management.v1.Model.State\x12\x38\n\x06schema\x18\x04 \x01(\x0b\x32(.continual.rpc.management.v1.ModelSchema\x12\x13\n\x0bschema_text\x18\x05 \x01(\t\x12-\n\x14latest_model_version\x18\x06 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12\x34\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x33\n\x17latest_batch_prediction\x18\n \x01(\tB\x12\xc2\x45\x0f\x42\x61tchPrediction\x12(\n\x0f\x63urrent_version\x18\x0b \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12\x12\n\x05table\x18\x0c \x01(\tB\x03\xb8\x45\x01\"0\n\x05State\x12\x0b\n\x07HEALTHY\x10\x00\x12\r\n\tUNHEALTHY\x10\x01\x12\x0b\n\x07OFFLINE\x10\x02\"\xaa\x07\n\tPromotion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x34\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x36\n\rpromoted_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x35\n\x0c\x64\x65moted_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12&\n\rmodel_version\x18\x05 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12+\n\x12\x62\x61se_model_version\x18\x06 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12@\n\x05state\x18\x07 \x01(\x0e\x32,.continual.rpc.management.v1.Promotion.StateB\x03\xb8\x45\x01\x12\x1a\n\x12improvement_metric\x18\x08 \x01(\t\x12 \n\x18improvement_metric_value\x18\t \x01(\x01\x12%\n\x1d\x62\x61se_improvement_metric_value\x18\n \x01(\x01\x12\x1f\n\x17improvement_metric_diff\x18\x0b \x01(\x01\x12\x41\n\x08\x64\x65\x63ision\x18\x0c \x01(\x0e\x32/.continual.rpc.management.v1.Promotion.Decision\x12=\n\x06reason\x18\r \x01(\x0e\x32-.continual.rpc.management.v1.Promotion.Reason\x12\x0f\n\x07message\x18\x0e \x01(\t\x12\x0e\n\x06manual\x18\x0f \x01(\x08\x12\x18\n\x07\x61\x63\x63ount\x18\x10 \x01(\tB\x07\xc2\x45\x04User\"A\n\x08\x44\x65\x63ision\x12\x18\n\x14\x44\x45\x43ISION_UNSPECIFIED\x10\x00\x12\x0b\n\x07PROMOTE\x10\x01\x12\x0e\n\nNO_PROMOTE\x10\x02\"\x91\x01\n\x06Reason\x12\x16\n\x12REASON_UNSPECIFIED\x10\x00\x12\n\n\x06UPLIFT\x10\x01\x12$\n NO_PROMOTED_MODEL_VERSION_EXISTS\x10\x02\x12\x11\n\rLATEST_POLICY\x10\x03\x12\x12\n\x0eUNDERPERFORMED\x10\x04\x12\n\n\x06MANUAL\x10\x05\x12\n\n\x06\x46ORCED\x10\x06\"9\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\r\n\tSUCCEEDED\x10\x01\x12\n\n\x06\x46\x41ILED\x10\x02\"\xfe\x0b\n\x0cModelVersion\x12\x11\n\x04name\x18\x01 \x01(\tB\x03\xb8\x45\x01\x12\x1d\n\x10model_version_id\x18\x02 \x01(\x03\x42\x03\xb8\x45\x02\x12\x1c\n\x0fmodel_signature\x18\x03 \x01(\tB\x03\xb8\x45\x02\x12!\n\nexperiment\x18\x04 \x01(\tB\r\xc2\x45\nExperiment\x12\x18\n\x0bstorage_uri\x18\x05 \x01(\tB\x03\xb8\x45\x02\x12\x1c\n\x14model_version_format\x18\x06 \x01(\t\x12\x43\n\x05state\x18\x07 \x01(\x0e\x32/.continual.rpc.management.v1.ModelVersion.StateB\x03\xb8\x45\x01\x12\x1a\n\rerror_message\x18\x08 \x01(\tB\x03\xb8\x45\x01\x12\x18\n\x0bstack_trace\x18\t \x01(\tB\x03\xb8\x45\x01\x12/\n\x0bsystem_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x36\n\x12visible_start_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x35\n\x11visible_stop_time\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x34\n\x0b\x63reate_time\x18\r \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12.\n\nstart_time\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x10 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1d\n\x10platform_version\x18\x11 \x01(\tB\x03\xb8\x45\x02\x12\x1f\n\x12training_row_count\x18\x12 \x01(\x04\x42\x03\xb8\x45\x01\x12!\n\x14validation_row_count\x18\x13 \x01(\x04\x42\x03\xb8\x45\x01\x12\x1b\n\x0etest_row_count\x18\x14 \x01(\x04\x42\x03\xb8\x45\x01\x12\x1f\n\x12performance_metric\x18\x15 \x01(\tB\x03\xb8\x45\x01\x12#\n\x16performance_metric_val\x18\x16 \x01(\x01\x42\x03\xb8\x45\x01\x12 \n\x13prediction_csv_path\x18\x17 \x01(\tB\x03\xb8\x45\x01\x12\x12\n\nfit_config\x18\x18 \x01(\t\x12\x1f\n\tpromotion\x18\x19 \x01(\tB\x0c\xc2\x45\tPromotion\x12\x10\n\x08promoted\x18\x1a \x01(\x08\x12S\n\x0elog_file_paths\x18\x1b \x03(\x0b\x32;.continual.rpc.management.v1.ModelVersion.LogFilePathsEntry\x12L\n\nplot_files\x18\x1c \x03(\x0b\x32\x38.continual.rpc.management.v1.ModelVersion.PlotFilesEntry\x12<\n\x0b\x64\x61ta_checks\x18\x1d \x01(\x0b\x32\'.continual.rpc.management.v1.DataChecks\x12\x14\n\x0c\x63redits_used\x18\x1e \x01(\x03\x12\x18\n\x0bmodel_query\x18\x1f \x01(\tB\x03\xb8\x45\x01\x1a\x33\n\x11LogFilePathsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x30\n\x0ePlotFilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xb6\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x11\n\rPREPROCESSING\x10\x03\x12\x0c\n\x08TRAINING\x10\x04\x12\x12\n\x0ePOSTPROCESSING\x10\x05\x12\x0e\n\nCANCELLING\x10\x06\x12\r\n\tSUCCEEDED\x10\x07\x12\n\n\x06\x46\x41ILED\x10\x08\x12\r\n\tCANCELLED\x10\t\x12\r\n\tPROFILING\x10\n\"\x84\x01\n\x0f\x43onfusionMatrix\x12\x14\n\x0c\x64isplay_name\x18\x03 \x03(\t\x12=\n\x03row\x18\x02 \x03(\x0b\x32\x30.continual.rpc.management.v1.ConfusionMatrix.Row\x1a\x1c\n\x03Row\x12\x15\n\rexample_count\x18\x01 \x03(\x05\"\xd5\x01\n\x18\x42ooleanEvaluationMetrics\x12\x10\n\x08log_loss\x18\x01 \x01(\x01\x12\x10\n\x08\x61\x63\x63uracy\x18\x02 \x01(\x01\x12\x0f\n\x07roc_auc\x18\x03 \x01(\x01\x12\r\n\x05\x65rror\x18\x04 \x01(\x01\x12\x46\n\x10\x63onfusion_matrix\x18\x05 \x01(\x0b\x32,.continual.rpc.management.v1.ConfusionMatrix\x12\x11\n\tprecision\x18\x06 \x01(\x01\x12\x0e\n\x06recall\x18\x07 \x01(\x01\x12\n\n\x02\x66\x31\x18\x08 \x01(\x01\"\x86\x01\n\x15TextEvaluationMetrics\x12\x10\n\x08log_loss\x18\x01 \x01(\x01\x12\n\n\x02\x66\x31\x18\x02 \x01(\x01\x12\x12\n\nperplexity\x18\x03 \x01(\x01\x12\x15\n\redit_distance\x18\x04 \x01(\x01\x12\x16\n\x0etoken_accuracy\x18\x05 \x01(\x01\x12\x0c\n\x04\x62leu\x18\x06 \x01(\x01\"\xe6\x03\n\x1c\x43\x61tegoricalEvaluationMetrics\x12\x10\n\x08log_loss\x18\x01 \x01(\x01\x12\x10\n\x08\x61\x63\x63uracy\x18\x02 \x01(\x01\x12\x0f\n\x07roc_auc\x18\x03 \x01(\x01\x12\x18\n\x10multiclass_error\x18\x04 \x01(\x01\x12\x46\n\x10\x63onfusion_matrix\x18\x05 \x01(\x0b\x32,.continual.rpc.management.v1.ConfusionMatrix\x12\x11\n\tprecision\x18\x06 \x01(\x01\x12\x17\n\x0fprecision_micro\x18\x07 \x01(\x01\x12\x17\n\x0fprecision_macro\x18\x08 \x01(\x01\x12\x1a\n\x12precision_weighted\x18\t \x01(\x01\x12\x0e\n\x06recall\x18\n \x01(\x01\x12\x14\n\x0crecall_micro\x18\x0b \x01(\x01\x12\x14\n\x0crecall_macro\x18\x0c \x01(\x01\x12\x17\n\x0frecall_weighted\x18\r \x01(\x01\x12\n\n\x02\x66\x31\x18\x0e \x01(\x01\x12\x10\n\x08\x66\x31_micro\x18\x0f \x01(\x01\x12\x10\n\x08\x66\x31_macro\x18\x10 \x01(\x01\x12\x13\n\x0b\x66\x31_weighted\x18\x11 \x01(\x01\x12\x19\n\x11\x62\x61lanced_accuracy\x18\x12 \x01(\x01\x12\x19\n\x11\x61verage_precision\x18\x13 \x01(\x01\"\xbb\x01\n\x17NumberEvaluationMetrics\x12\x0c\n\x04rmse\x18\x01 \x01(\x01\x12\x0b\n\x03mae\x18\x02 \x01(\x01\x12\n\n\x02r2\x18\x03 \x01(\x01\x12\x11\n\tprecision\x18\x04 \x01(\x01\x12\x0e\n\x06recall\x18\x05 \x01(\x01\x12\x0c\n\x04mape\x18\x06 \x01(\x01\x12\r\n\x05smape\x18\x07 \x01(\x01\x12\x1a\n\x12mean_squared_error\x18\x08 \x01(\x01\x12\x1d\n\x15median_absolute_error\x18\t \x01(\x01\"\xc6\x02\n\x11\x45valuationMetrics\x12H\n\x07\x62oolean\x18\x01 \x01(\x0b\x32\x35.continual.rpc.management.v1.BooleanEvaluationMetricsH\x00\x12\x46\n\x06number\x18\x02 \x01(\x0b\x32\x34.continual.rpc.management.v1.NumberEvaluationMetricsH\x00\x12\x42\n\x04text\x18\x03 \x01(\x0b\x32\x32.continual.rpc.management.v1.TextEvaluationMetricsH\x00\x12P\n\x0b\x63\x61tegorical\x18\x04 \x01(\x0b\x32\x39.continual.rpc.management.v1.CategoricalEvaluationMetricsH\x00\x42\t\n\x07metrics\"\xe2\x01\n\x07Metrics\x12\x45\n\rtrain_metrics\x18\x01 \x01(\x0b\x32..continual.rpc.management.v1.EvaluationMetrics\x12J\n\x12validation_metrics\x18\x02 \x01(\x0b\x32..continual.rpc.management.v1.EvaluationMetrics\x12\x44\n\x0ctest_metrics\x18\x03 \x01(\x0b\x32..continual.rpc.management.v1.EvaluationMetrics\"\x8c\x08\n\nExperiment\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x11\n\talgorithm\x18\x05 \x01(\t\x12<\n\x05state\x18\x06 \x01(\x0e\x32-.continual.rpc.management.v1.Experiment.State\x12/\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x66inish_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1a\n\x12performance_metric\x18\n \x01(\t\x12\x1e\n\x16performance_metric_val\x18\x0b \x01(\x01\x12\x35\n\x07metrics\x18\x0c \x01(\x0b\x32$.continual.rpc.management.v1.Metrics\x12:\n\nmodel_type\x18\r \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\x12Z\n\x12\x66\x65\x61ture_importance\x18\x0e \x03(\x0b\x32>.continual.rpc.management.v1.Experiment.FeatureImportanceEntry\x12\x17\n\x0ftraining_config\x18\x0f \x01(\t\x12\x15\n\rerror_message\x18\x10 \x01(\t\x12\x13\n\x0bstack_trace\x18\x11 \x01(\t\x12k\n\x1b\x66\x65\x61ture_importance_relative\x18\x12 \x03(\x0b\x32\x46.continual.rpc.management.v1.Experiment.FeatureImportanceRelativeEntry\x1a\x38\n\x16\x46\x65\x61tureImportanceEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01\x1a@\n\x1e\x46\x65\x61tureImportanceRelativeEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01\"\x9a\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\x11\n\rPREPROCESSING\x10\x02\x12\x0c\n\x08TRAINING\x10\x03\x12\x12\n\x0ePOSTPROCESSING\x10\x04\x12\x0e\n\nCANCELLING\x10\x05\x12\r\n\tSUCCEEDED\x10\x06\x12\n\n\x06\x46\x41ILED\x10\x07\x12\r\n\tCANCELLED\x10\x08\"\xea\x01\n\x0fTrainingMetrics\x12\x12\n\nexperiment\x18\x01 \x01(\t\x12\x0c\n\x04step\x18\x02 \x01(\x05\x12\x0f\n\x07trainer\x18\x03 \x01(\t\x12\x30\n\x0c\x63\x61pture_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12;\n\x0bmetric_type\x18\x05 \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\x12\x35\n\x07metrics\x18\x06 \x01(\x0b\x32$.continual.rpc.management.v1.Metrics\"Q\n\x06Metric\x12(\n\x04time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x0e\n\x06values\x18\x03 \x03(\x01\"\xc2\x01\n\x15\x42\x61tchPredictionMetric\x12\x0c\n\x04name\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x18\n\x10prediction_count\x18\x04 \x01(\x03\x12\x0f\n\x07success\x18\x05 \x01(\x08\x12\x12\n\nerror_code\x18\x06 \x01(\t\"\x1d\n\x0cTextSequence\x12\r\n\x05value\x18\x01 \x03(\t\"\x1f\n\x0eNumberSequence\x12\r\n\x05value\x18\x01 \x03(\x01\"`\n\x10SignatureFeature\x12\x16\n\x0equalified_name\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\"\x90\x02\n\x0cQueryFeature\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x16\n\x0equalified_name\x18\x02 \x01(\t\x12\x34\n\x08relation\x18\x03 \x01(\x0b\x32\".continual.rpc.management.v1.Query\x12\r\n\x05value\x18\x04 \x01(\t\x12\x34\n\x04type\x18\x05 \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\x12\r\n\x05\x64type\x18\x06 \x01(\t\x12\r\n\x05model\x18\x07 \x01(\x08\x12\r\n\x05index\x18\x08 \x01(\x08\x12\x12\n\ntime_index\x18\t \x01(\x08\x12\x0f\n\x07\x65xclude\x18\n \x01(\x08\x12\r\n\x05split\x18\x0b \x01(\x08\"\xac\x01\n\x05Query\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tentity_id\x18\x02 \x01(\t\x12\x12\n\ntable_name\x18\x03 \x01(\t\x12\x39\n\x06\x66ields\x18\x04 \x03(\x0b\x32).continual.rpc.management.v1.QueryFeature\x12\x33\n\x07related\x18\x05 \x03(\x0b\x32\".continual.rpc.management.v1.Query\"\x9b\x02\n\rChangeSummary\x12\x14\n\x0cmodels_added\x18\x01 \x03(\t\x12\x17\n\x0fmodels_modified\x18\x02 \x03(\t\x12\x16\n\x0emodels_deleted\x18\x03 \x03(\t\x12\x1e\n\x16model_versions_trained\x18\x04 \x03(\t\x12\x1a\n\x12\x66\x65\x61ture_sets_added\x18\x05 \x03(\t\x12\x1d\n\x15\x66\x65\x61ture_sets_modified\x18\x06 \x03(\t\x12\x1c\n\x14\x66\x65\x61ture_sets_deleted\x18\x07 \x03(\t\x12\x16\n\x0e\x66\x65\x61tures_added\x18\x08 \x03(\t\x12\x18\n\x10\x66\x65\x61tures_deleted\x18\t \x03(\t\x12\x18\n\x10predictions_made\x18\n \x01(\x04\"\xa8\x04\n\x06\x43hange\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x38\n\x05state\x18\x02 \x01(\x0e\x32).continual.rpc.management.v1.Change.State\x12\x33\n\x04plan\x18\x03 \x03(\x0b\x32%.continual.rpc.management.v1.PlanStep\x12/\n\x0b\x63reate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x66inish_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\ncreated_by\x18\x08 \x01(\t\x12\x14\n\x0c\x63\x61ncelled_by\x18\t \x01(\t\x12\x0f\n\x07message\x18\n \x01(\t\x12\x0f\n\x07trigger\x18\x0b \x01(\t\x12\x10\n\x08resource\x18\x0c \x01(\t\x12;\n\x07summary\x18\r \x01(\x0b\x32*.continual.rpc.management.v1.ChangeSummary\"t\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\r\n\tEXECUTING\x10\x02\x12\x0e\n\nCANCELLING\x10\x03\x12\r\n\tSUCCEEDED\x10\x04\x12\n\n\x06\x46\x41ILED\x10\x05\x12\r\n\tCANCELLED\x10\x06\"\xcc\"\n\x08PlanStep\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x38\n\x04type\x18\x03 \x01(\x0e\x32*.continual.rpc.management.v1.PlanStep.Type\x12\x42\n\toperation\x18\x04 \x01(\x0e\x32/.continual.rpc.management.v1.PlanStep.Operation\x12:\n\x05state\x18\x05 \x01(\x0e\x32+.continual.rpc.management.v1.PlanStep.State\x12\x15\n\rresource_name\x18\x06 \x01(\t\x12\x13\n\x0bobject_name\x18\x07 \x01(\t\x12\x14\n\x0c\x64\x65pendencies\x18\x08 \x03(\x03\x12\x16\n\x0e\x65rror_messages\x18\t \x03(\t\x12.\n\nstart_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x66inish_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0bschema_text\x18\x0c \x01(\t\x12\x1c\n\x14previous_schema_text\x18\r \x01(\t\x12T\n\x12\x63reate_feature_set\x18\x0e \x01(\x0b\x32\x36.continual.rpc.management.v1.PlanStep.CreateFeatureSetH\x00\x12I\n\x0c\x63reate_model\x18\x0f \x01(\x0b\x32\x31.continual.rpc.management.v1.PlanStep.CreateModelH\x00\x12T\n\x12update_feature_set\x18\x10 \x01(\x0b\x32\x36.continual.rpc.management.v1.PlanStep.UpdateFeatureSetH\x00\x12I\n\x0cupdate_model\x18\x11 \x01(\x0b\x32\x31.continual.rpc.management.v1.PlanStep.UpdateModelH\x00\x12V\n\x13profile_feature_set\x18\x12 \x01(\x0b\x32\x37.continual.rpc.management.v1.PlanStep.ProfileFeatureSetH\x00\x12K\n\rprofile_model\x18\x13 \x01(\x0b\x32\x32.continual.rpc.management.v1.PlanStep.ProfileModelH\x00\x12<\n\x05train\x18\x14 \x01(\x0b\x32+.continual.rpc.management.v1.PlanStep.TrainH\x00\x12@\n\x07promote\x18\x15 \x01(\x0b\x32-.continual.rpc.management.v1.PlanStep.PromoteH\x00\x12@\n\x07predict\x18\x16 \x01(\x0b\x32-.continual.rpc.management.v1.PlanStep.PredictH\x00\x12T\n\x12\x64\x65lete_feature_set\x18\x17 \x01(\x0b\x32\x36.continual.rpc.management.v1.PlanStep.DeleteFeatureSetH\x00\x12I\n\x0c\x64\x65lete_model\x18\x18 \x01(\x0b\x32\x31.continual.rpc.management.v1.PlanStep.DeleteModelH\x00\x1aj\n\x0b\x43reateModel\x12\x17\n\x05model\x18\x01 \x01(\tB\x08\xc2\x45\x05Model\x12\x13\n\x0bschema_text\x18\x02 \x01(\t\x12\x16\n\x0e\x61\x64\x64\x65\x64_features\x18\x03 \x03(\t\x12\x15\n\rcreated_views\x18\x04 \x03(\t\x1az\n\x10\x43reateFeatureSet\x12\"\n\x0b\x66\x65\x61ture_set\x18\x01 \x01(\tB\r\xc2\x45\nFeatureSet\x12\x13\n\x0bschema_text\x18\x02 \x01(\t\x12\x16\n\x0e\x61\x64\x64\x65\x64_features\x18\x03 \x03(\t\x12\x15\n\rcreated_views\x18\x04 \x03(\t\x1a\xa0\x04\n\x0bUpdateModel\x12\x17\n\x05model\x18\x01 \x01(\tB\x08\xc2\x45\x05Model\x12\x13\n\x0bschema_text\x18\x02 \x01(\t\x12\x1c\n\x14previous_schema_text\x18\x03 \x01(\t\x12\x16\n\x0e\x61\x64\x64\x65\x64_features\x18\x04 \x03(\t\x12\x18\n\x10updated_features\x18\x05 \x03(\t\x12\x18\n\x10removed_features\x18\x06 \x03(\t\x12\x15\n\rcreated_views\x18\x07 \x03(\t\x12\x15\n\rupdated_views\x18\x08 \x03(\t\x12\x15\n\rdeleted_views\x18\t \x03(\t\x12M\n\x07updates\x18\n \x03(\x0e\x32<.continual.rpc.management.v1.PlanStep.UpdateModel.UpdateType\"\xe4\x01\n\nUpdateType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0f\n\x0b\x44\x45SCRIPTION\x10\x01\x12\x11\n\rDOCUMENTATION\x10\x02\x12\t\n\x05QUERY\x10\x03\x12\t\n\x05TABLE\x10\x04\x12\t\n\x05INDEX\x10\x05\x12\x0e\n\nTIME_INDEX\x10\x06\x12\t\n\x05SPLIT\x10\x07\x12\x13\n\x0f\x45XCLUDE_COLUMNS\x10\x08\x12\x12\n\x0ePROFILE_CONFIG\x10\t\x12\x13\n\x0fTRAINING_CONFIG\x10\n\x12\x14\n\x10PROMOTION_CONFIG\x10\x0b\x12\x15\n\x11PREDICTION_CONFIG\x10\x0c\x1a\xff\x03\n\x10UpdateFeatureSet\x12\"\n\x0b\x66\x65\x61ture_set\x18\x01 \x01(\tB\r\xc2\x45\nFeatureSet\x12\x13\n\x0bschema_text\x18\x02 \x01(\t\x12\x1c\n\x14previous_schema_text\x18\x03 \x01(\t\x12\x16\n\x0e\x61\x64\x64\x65\x64_features\x18\x04 \x03(\t\x12\x18\n\x10updated_features\x18\x05 \x03(\t\x12\x18\n\x10removed_features\x18\x06 \x03(\t\x12\x15\n\rcreated_views\x18\x07 \x03(\t\x12\x15\n\rupdated_views\x18\x08 \x03(\t\x12\x15\n\rdeleted_views\x18\t \x03(\t\x12R\n\x07updates\x18\n \x03(\x0e\x32\x41.continual.rpc.management.v1.PlanStep.UpdateFeatureSet.UpdateType\"\xae\x01\n\nUpdateType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0f\n\x0b\x44\x45SCRIPTION\x10\x01\x12\x11\n\rDOCUMENTATION\x10\x02\x12\t\n\x05QUERY\x10\x03\x12\t\n\x05TABLE\x10\x04\x12\n\n\x06\x45NTITY\x10\x05\x12\t\n\x05INDEX\x10\x06\x12\x0e\n\nTIME_INDEX\x10\x07\x12\t\n\x05SPLIT\x10\x08\x12\x13\n\x0f\x45XCLUDE_COLUMNS\x10\t\x12\x12\n\x0ePROFILE_CONFIG\x10\n\x1a\xac\x01\n\x11ProfileFeatureSet\x12\"\n\x0b\x66\x65\x61ture_set\x18\x01 \x01(\tB\r\xc2\x45\nFeatureSet\x12\x15\n\rrows_profiled\x18\x02 \x01(\x03\x12\x1e\n\x16previous_rows_profiled\x18\x03 \x01(\x03\x12<\n\x0b\x64\x61ta_checks\x18\x04 \x01(\x0b\x32\'.continual.rpc.management.v1.DataChecks\x1a\x9c\x01\n\x0cProfileModel\x12\x17\n\x05model\x18\x01 \x01(\tB\x08\xc2\x45\x05Model\x12\x15\n\rrows_profiled\x18\x02 \x01(\x03\x12\x1e\n\x16previous_rows_profiled\x18\x03 \x01(\x03\x12<\n\x0b\x64\x61ta_checks\x18\x04 \x01(\x0b\x32\'.continual.rpc.management.v1.DataChecks\x1a\xab\x01\n\x07Predict\x12,\n\x10\x62\x61tch_prediction\x18\x01 \x01(\tB\x12\xc2\x45\x0f\x42\x61tchPrediction\x12\x35\n\x19previous_batch_prediction\x18\x02 \x01(\tB\x12\xc2\x45\x0f\x42\x61tchPrediction\x12\x18\n\x10prediction_count\x18\x03 \x01(\x03\x12!\n\x19previous_prediction_count\x18\x04 \x01(\x03\x1a\xc1\x04\n\x05Train\x12&\n\rmodel_version\x18\x01 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12+\n\x12\x62\x61se_model_version\x18\x02 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12\x61\n\x17\x62\x61se_model_version_type\x18\x03 \x01(\x0e\x32@.continual.rpc.management.v1.PlanStep.Train.BaseModelVersionType\x12\x42\n\x06reason\x18\x04 \x01(\x0e\x32\x32.continual.rpc.management.v1.PlanStep.Train.Reason\x12-\n\x14latest_model_version\x18\x05 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12/\n\x16promoted_model_version\x18\x06 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12\x31\n\x18production_model_version\x18\x07 \x01(\tB\x0f\xc2\x45\x0cModelVersion\"i\n\x14\x42\x61seModelVersionType\x12\'\n#BASE_MODEL_VERSION_TYPE_UNSPECIFIED\x10\x00\x12\x0c\n\x08PROMOTED\x10\x01\x12\n\n\x06LATEST\x10\x02\x12\x0e\n\nPRODUCTION\x10\x03\">\n\x06Reason\x12\x16\n\x12REASON_UNSPECIFIED\x10\x00\x12\x10\n\x0c\x44\x41TA_CHANGED\x10\x01\x12\n\n\x06\x46ORCED\x10\x02\x1a*\n\x07Promote\x12\x1f\n\tpromotion\x18\x01 \x01(\tB\x0c\xc2\x45\tPromotion\x1a>\n\x0b\x44\x65leteModel\x12\x15\n\rdeleted_views\x18\x01 \x03(\t\x12\x18\n\x10\x64\x65leted_features\x18\x02 \x03(\t\x1a\x43\n\x10\x44\x65leteFeatureSet\x12\x15\n\rdeleted_views\x18\x01 \x03(\t\x12\x18\n\x10\x64\x65leted_features\x18\x02 \x03(\t\"\x81\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\r\n\tEXECUTING\x10\x02\x12\r\n\tSUCCEEDED\x10\x03\x12\n\n\x06\x46\x41ILED\x10\x04\x12\r\n\tCANCELLED\x10\x05\x12\x0e\n\nCANCELLING\x10\x06\x12\x0b\n\x07SKIPPED\x10\x07\"\xa5\x01\n\tOperation\x12\x1b\n\x17OPERATION_UNSPECIFICIED\x10\x00\x12\x07\n\x03PUT\x10\x01\x12\n\n\x06\x44\x45LETE\x10\x02\x12\n\n\x06INGEST\x10\x03\x12\t\n\x05TRAIN\x10\x04\x12\x0b\n\x07PROMOTE\x10\x05\x12\x0b\n\x07PROFILE\x10\x06\x12\x10\n\x0c\x42\x41TCHPREDICT\x10\x07\x12\n\n\x06\x43REATE\x10\x08\x12\n\n\x06UPDATE\x10\t\x12\x0b\n\x07PREDICT\x10\n\"\xfd\x01\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x16\n\x12\x43REATE_FEATURE_SET\x10\x01\x12\x16\n\x12UPDATE_FEATURE_SET\x10\x02\x12\x17\n\x13PROFILE_FEATURE_SET\x10\x03\x12\x16\n\x12\x44\x45LETE_FEATURE_SET\x10\x04\x12\x10\n\x0c\x43REATE_MODEL\x10\x05\x12\x10\n\x0cUPDATE_MODEL\x10\x06\x12\x11\n\rPROFILE_MODEL\x10\x07\x12\x10\n\x0c\x44\x45LETE_MODEL\x10\x08\x12\x0f\n\x0bTRAIN_MODEL\x10\t\x12\x11\n\rPROMOTE_MODEL\x10\n\x12\x11\n\rPREDICT_MODEL\x10\x0b\x42\n\n\x08metadata\"\xae\x07\n\x0f\x42\x61tchPrediction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x41\n\x05state\x18\x02 \x01(\x0e\x32\x32.continual.rpc.management.v1.BatchPrediction.State\x12&\n\rmodel_version\x18\x03 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12J\n\x0bsource_type\x18\x04 \x01(\x0e\x32\x35.continual.rpc.management.v1.BatchPredictionStoreType\x12\x0e\n\x06source\x18\x05 \x01(\t\x12H\n\tdest_type\x18\x06 \x01(\x0e\x32\x35.continual.rpc.management.v1.BatchPredictionStoreType\x12\x13\n\x0bincremental\x18\x07 \x01(\x08\x12\x15\n\rcsv_file_path\x18\x08 \x01(\t\x12/\n\x0b\x63reate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x18\n\x10prediction_count\x18\x0c \x01(\x03\x12\x15\n\rerror_message\x18\r \x01(\t\x12\x13\n\x0bstack_trace\x18\x0e \x01(\t\x12V\n\x0elog_file_paths\x18\x0f \x03(\x0b\x32>.continual.rpc.management.v1.BatchPrediction.LogFilePathsEntry\x12\x18\n\x10prediction_model\x18\x10 \x01(\t\x12\x12\n\x05table\x18\x11 \x01(\tB\x03\xb8\x45\x01\x12\x14\n\x0c\x63redits_used\x18\x12 \x01(\x03\x1a\x33\n\x11LogFilePathsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa9\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x11\n\rPREPROCESSING\x10\x03\x12\x0e\n\nPROCESSING\x10\x04\x12\x12\n\x0ePOSTPROCESSING\x10\x05\x12\x0e\n\nCANCELLING\x10\x06\x12\r\n\tSUCCEEDED\x10\x07\x12\n\n\x06\x46\x41ILED\x10\x08\x12\r\n\tCANCELLED\x10\t\"\x1f\n\tSignedURL\x12\x12\n\nsigned_url\x18\x01 \x01(\t\"<\n\x06\x45ntity\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06models\x18\x02 \x03(\t\x12\x14\n\x0c\x66\x65\x61ture_sets\x18\x03 \x03(\t\"\xac\x01\n\x0c\x44\x61tasetStats\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x45\n\rdataset_stats\x18\x02 \x03(\x0b\x32..continual.rpc.management.v1.DatasetStatsEntry\x12\x30\n\x0cprofile_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x15\n\rresource_name\x18\x04 \x01(\t\"\x91\x04\n\x11\x44\x61tasetStatsEntry\x12\x12\n\nentry_name\x18\x01 \x01(\t\x12\x14\n\x0cnum_examples\x18\x02 \x01(\x04\x12J\n\x12\x63orrelation_matrix\x18\x03 \x01(\x0b\x32..continual.rpc.management.v1.CorrelationMatrix\x12>\n\x0cstring_stats\x18\x04 \x03(\x0b\x32(.continual.rpc.management.v1.StringStats\x12@\n\rnumeric_stats\x18\x05 \x03(\x0b\x32).continual.rpc.management.v1.NumericStats\x12\x44\n\x0ftimestamp_stats\x18\x06 \x03(\x0b\x32+.continual.rpc.management.v1.TimestampStats\x12H\n\x11\x63\x61tegorical_stats\x18\x07 \x03(\x0b\x32-.continual.rpc.management.v1.CategoricalStats\x12\x16\n\x0e\x64uplicate_rows\x18\x08 \x01(\x04\x12\x19\n\x11\x64uplicate_indexes\x18\t \x01(\x04\x12\x41\n\x0e\x64\x61taset_ranges\x18\n \x03(\x0b\x32).continual.rpc.management.v1.DatasetRange\"z\n\x0c\x44\x61tasetRange\x12\x0c\n\x04name\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x89\x01\n\x11\x43orrelationMatrix\x12\x15\n\rdisplay_names\x18\x01 \x03(\t\x12@\n\x04rows\x18\x02 \x03(\x0b\x32\x32.continual.rpc.management.v1.CorrelationMatrix.Row\x1a\x1b\n\x03Row\x12\x14\n\x0c\x63orrelations\x18\x01 \x03(\x02\"\xba\x01\n\tHistogram\x12\x0f\n\x07num_nan\x18\x01 \x01(\x04\x12\x15\n\rnum_undefined\x18\x02 \x01(\x04\x12>\n\x07\x62uckets\x18\x03 \x03(\x0b\x32-.continual.rpc.management.v1.Histogram.Bucket\x1a\x45\n\x06\x42ucket\x12\x11\n\tlow_value\x18\x01 \x01(\x01\x12\x12\n\nhigh_value\x18\x02 \x01(\x01\x12\x14\n\x0csample_count\x18\x04 \x01(\x03\"\x84\x02\n\x12TimestampHistogram\x12\x0f\n\x07num_nan\x18\x01 \x01(\x04\x12\x15\n\rnum_undefined\x18\x02 \x01(\x04\x12G\n\x07\x62uckets\x18\x03 \x03(\x0b\x32\x36.continual.rpc.management.v1.TimestampHistogram.Bucket\x1a}\n\x06\x42ucket\x12-\n\tlow_value\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nhigh_value\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0csample_count\x18\x04 \x01(\x03\"\x9d\x01\n\x17\x43\x61tegoricalDistribution\x12Q\n\ncategories\x18\x01 \x03(\x0b\x32=.continual.rpc.management.v1.CategoricalDistribution.Category\x1a/\n\x08\x43\x61tegory\x12\r\n\x05label\x18\x04 \x01(\t\x12\x14\n\x0csample_count\x18\x05 \x01(\x03\"w\n\x0bStringStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x12\n\nnull_count\x18\x03 \x01(\x03\x12\x12\n\nmin_length\x18\x04 \x01(\x03\x12\x12\n\nmax_length\x18\x05 \x01(\x03\"\xac\x01\n\x10\x43\x61tegoricalStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x12\n\nnull_count\x18\x03 \x01(\x03\x12V\n\x18\x63\x61tegorical_distribution\x18\x04 \x01(\x0b\x32\x34.continual.rpc.management.v1.CategoricalDistribution\"\xf2\x01\n\x0eTimestampStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x12\n\nnull_count\x18\x03 \x01(\x03\x12\'\n\x03min\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03max\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12L\n\x13timestamp_histogram\x18\x06 \x01(\x0b\x32/.continual.rpc.management.v1.TimestampHistogram\"\xac\x02\n\x0cNumericStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x0b\n\x03min\x18\x03 \x01(\x01\x12\x0b\n\x03max\x18\x04 \x01(\x01\x12\n\n\x02sd\x18\x05 \x01(\x01\x12\x0c\n\x04mean\x18\x06 \x01(\x01\x12\r\n\x05\x63ount\x18\x07 \x01(\x03\x12\x14\n\x0cpercentile25\x18\x08 \x01(\x01\x12\x14\n\x0cpercentile50\x18\t \x01(\x01\x12\x14\n\x0cpercentile75\x18\n \x01(\x01\x12\x12\n\nnull_count\x18\x0b \x01(\x03\x12\x0f\n\x07num_nan\x18\x0c \x01(\x03\x12\x44\n\x14num_values_histogram\x18\r \x01(\x0b\x32&.continual.rpc.management.v1.Histogram\"\xa7\x01\n\x0e\x42illingAccount\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x0c\n\x04memo\x18\x03 \x01(\t\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\"\xdb\x01\n\x0cSubscription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x17\n\x05price\x18\x02 \x01(\tB\x08\xc2\x45\x05Price\x12(\n\x0epayment_method\x18\x03 \x01(\tB\x10\xc2\x45\rPaymentMethod\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x0e\n\x06\x61\x63tive\x18\x06 \x01(\x08\"\xb8\x01\n\x12ProjectUsageRecord\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x63redits\x18\x02 \x01(\x03\x12Q\n\ttypeCount\x18\x03 \x03(\x0b\x32>.continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry\x1a\x30\n\x0eTypeCountEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"\x99\x01\n\x11\x43reditUsageRecord\x12*\n\x06period\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x63redits\x18\x02 \x01(\x03\x12G\n\x0eproject_usages\x18\x03 \x03(\x0b\x32/.continual.rpc.management.v1.ProjectUsageRecord\"\xc8\x01\n\x0fInvoiceLineItem\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x03\x12\x10\n\x08quantity\x18\x03 \x01(\x03\x12\x10\n\x08\x63urrency\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x30\n\x0cperiod_start\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nperiod_end\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x9a\x02\n\x07Invoice\x12\x0c\n\x04name\x18\x01 \x01(\t\x12(\n\x04\x64\x61te\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x03\x12\x30\n\x0cperiod_start\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nperiod_end\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06status\x18\x07 \x01(\t\x12\x13\n\x0binvoice_url\x18\x08 \x01(\t\x12@\n\nline_items\x18\t \x03(\x0b\x32,.continual.rpc.management.v1.InvoiceLineItem\"Q\n\x05Price\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x16\n\x0e\x64irect_invoice\x18\x04 \x01(\x08\"\xa2\x02\n\rPaymentMethod\x12\x0c\n\x04name\x18\x01 \x01(\t\x12R\n\x0f\x62illing_details\x18\x02 \x01(\x0b\x32\x39.continual.rpc.management.v1.PaymentMethod.BillingDetails\x12=\n\x04\x63\x61rd\x18\x03 \x01(\x0b\x32/.continual.rpc.management.v1.PaymentMethod.Card\x1a%\n\x0e\x42illingDetails\x12\x13\n\x0bpostal_code\x18\x01 \x01(\t\x1aI\n\x04\x43\x61rd\x12\r\n\x05\x62rand\x18\x01 \x01(\t\x12\r\n\x05last4\x18\x02 \x01(\t\x12\x10\n\x08\x65xp_year\x18\x03 \x01(\x03\x12\x11\n\texp_month\x18\x04 \x01(\x03\"\xd8\x01\n\nDataChecks\x12M\n\x07results\x18\x01 \x03(\x0b\x32<.continual.rpc.management.v1.DataChecks.DataValidationResult\x1a{\n\x14\x44\x61taValidationResult\x12\x0f\n\x07summary\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x10\n\x08\x64uration\x18\x03 \x01(\x02\x12\x0e\n\x06\x65rrors\x18\x04 \x03(\t\x12\x10\n\x08warnings\x18\x05 \x03(\t\x12\r\n\x05infos\x18\x06 \x03(\t*\xd5\x01\n\tFieldType\x12\x0b\n\x07INVALID\x10\x00\x12\n\n\x06NUMBER\x10\x01\x12\x0b\n\x07\x42OOLEAN\x10\x02\x12\x0f\n\x0b\x43\x41TEGORICAL\x10\x03\x12\x08\n\x04TEXT\x10\x04\x12\r\n\tTIMESTAMP\x10\x05\x12\x08\n\x04TIME\x10\x06\x12\x08\n\x04\x44\x41TE\x10\x07\x12\x0c\n\x08\x44\x41TETIME\x10\x08\x12\x0f\n\x0bNUMBER_LIST\x10\t\x12\x10\n\x0c\x42OOLEAN_LIST\x10\n\x12\r\n\tTEXT_LIST\x10\x0b\x12\t\n\x05INDEX\x10\x0c\x12\x0e\n\nTIME_INDEX\x10\r\x12\t\n\x05SPLIT\x10\x0f*?\n\x18\x42\x61tchPredictionStoreType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x07\n\x03\x43SV\x10\x01\x12\t\n\x05STORE\x10\x02\x42KZIgithub.com/continualiq/continual/continual/rpc/management/v1;managementv1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,continual_dot_rpc_dot_graphql_dot_options__pb2.DESCRIPTOR,])

_FIELDTYPE = _descriptor.EnumDescriptor(
  name='FieldType',
  full_name='continual.rpc.management.v1.FieldType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INVALID', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NUMBER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CATEGORICAL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIMESTAMP', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATE', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATETIME', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NUMBER_LIST', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN_LIST', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT_LIST', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_INDEX', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPLIT', index=14, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=25845,
  serialized_end=26058,
)
_sym_db.RegisterEnumDescriptor(_FIELDTYPE)

FieldType = enum_type_wrapper.EnumTypeWrapper(_FIELDTYPE)
_BATCHPREDICTIONSTORETYPE = _descriptor.EnumDescriptor(
  name='BatchPredictionStoreType',
  full_name='continual.rpc.management.v1.BatchPredictionStoreType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CSV', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STORE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=26060,
  serialized_end=26123,
)
_sym_db.RegisterEnumDescriptor(_BATCHPREDICTIONSTORETYPE)

BatchPredictionStoreType = enum_type_wrapper.EnumTypeWrapper(_BATCHPREDICTIONSTORETYPE)
INVALID = 0
NUMBER = 1
BOOLEAN = 2
CATEGORICAL = 3
TEXT = 4
TIMESTAMP = 5
TIME = 6
DATE = 7
DATETIME = 8
NUMBER_LIST = 9
BOOLEAN_LIST = 10
TEXT_LIST = 11
INDEX = 12
TIME_INDEX = 13
SPLIT = 15
UNSPECIFIED = 0
CSV = 1
STORE = 2


_INGESTION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Ingestion.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RUNNING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=606,
  serialized_end=720,
)
_sym_db.RegisterEnumDescriptor(_INGESTION_STATE)

_ORGANIZATION_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='continual.rpc.management.v1.Organization.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EXPIRED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE_TRIAL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DISABLED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2142,
  serialized_end=2207,
)
_sym_db.RegisterEnumDescriptor(_ORGANIZATION_STATUS)

_MODEL_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Model.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='HEALTHY', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNHEALTHY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='OFFLINE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=8866,
  serialized_end=8914,
)
_sym_db.RegisterEnumDescriptor(_MODEL_STATE)

_PROMOTION_DECISION = _descriptor.EnumDescriptor(
  name='Decision',
  full_name='continual.rpc.management.v1.Promotion.Decision',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DECISION_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROMOTE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NO_PROMOTE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=9583,
  serialized_end=9648,
)
_sym_db.RegisterEnumDescriptor(_PROMOTION_DECISION)

_PROMOTION_REASON = _descriptor.EnumDescriptor(
  name='Reason',
  full_name='continual.rpc.management.v1.Promotion.Reason',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REASON_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPLIFT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NO_PROMOTED_MODEL_VERSION_EXISTS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LATEST_POLICY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNDERPERFORMED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MANUAL', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FORCED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=9651,
  serialized_end=9796,
)
_sym_db.RegisterEnumDescriptor(_PROMOTION_REASON)

_PROMOTION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Promotion.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=9798,
  serialized_end=9855,
)
_sym_db.RegisterEnumDescriptor(_PROMOTION_STATE)

_MODELVERSION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.ModelVersion.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREPROCESSING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAINING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTPROCESSING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILING', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11210,
  serialized_end=11392,
)
_sym_db.RegisterEnumDescriptor(_MODELVERSION_STATE)

_EXPERIMENT_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Experiment.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREPROCESSING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAINING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTPROCESSING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=14002,
  serialized_end=14156,
)
_sym_db.RegisterEnumDescriptor(_EXPERIMENT_STATE)

_CHANGE_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Change.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXECUTING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=16010,
  serialized_end=16126,
)
_sym_db.RegisterEnumDescriptor(_CHANGE_STATE)

_PLANSTEP_UPDATEMODEL_UPDATETYPE = _descriptor.EnumDescriptor(
  name='UpdateType',
  full_name='continual.rpc.management.v1.PlanStep.UpdateModel.UpdateType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESCRIPTION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DOCUMENTATION', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='QUERY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TABLE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_INDEX', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPLIT', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXCLUDE_COLUMNS', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_CONFIG', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAINING_CONFIG', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROMOTION_CONFIG', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREDICTION_CONFIG', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=17982,
  serialized_end=18210,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_UPDATEMODEL_UPDATETYPE)

_PLANSTEP_UPDATEFEATURESET_UPDATETYPE = _descriptor.EnumDescriptor(
  name='UpdateType',
  full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.UpdateType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESCRIPTION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DOCUMENTATION', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='QUERY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TABLE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ENTITY', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_INDEX', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPLIT', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXCLUDE_COLUMNS', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_CONFIG', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=18550,
  serialized_end=18724,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_UPDATEFEATURESET_UPDATETYPE)

_PLANSTEP_TRAIN_BASEMODELVERSIONTYPE = _descriptor.EnumDescriptor(
  name='BaseModelVersionType',
  full_name='continual.rpc.management.v1.PlanStep.Train.BaseModelVersionType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BASE_MODEL_VERSION_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROMOTED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LATEST', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PRODUCTION', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19643,
  serialized_end=19748,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_TRAIN_BASEMODELVERSIONTYPE)

_PLANSTEP_TRAIN_REASON = _descriptor.EnumDescriptor(
  name='Reason',
  full_name='continual.rpc.management.v1.PlanStep.Train.Reason',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REASON_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_CHANGED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FORCED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19750,
  serialized_end=19812,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_TRAIN_REASON)

_PLANSTEP_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.PlanStep.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXECUTING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SKIPPED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19992,
  serialized_end=20121,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_STATE)

_PLANSTEP_OPERATION = _descriptor.EnumDescriptor(
  name='Operation',
  full_name='continual.rpc.management.v1.PlanStep.Operation',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OPERATION_UNSPECIFICIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PUT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DELETE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INGEST', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAIN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROMOTE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILE', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BATCHPREDICT', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATE', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPDATE', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREDICT', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=20124,
  serialized_end=20289,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_OPERATION)

_PLANSTEP_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='continual.rpc.management.v1.PlanStep.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATE_FEATURE_SET', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPDATE_FEATURE_SET', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_FEATURE_SET', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DELETE_FEATURE_SET', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATE_MODEL', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPDATE_MODEL', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_MODEL', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DELETE_MODEL', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAIN_MODEL', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROMOTE_MODEL', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREDICT_MODEL', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=20292,
  serialized_end=20545,
)
_sym_db.RegisterEnumDescriptor(_PLANSTEP_TYPE)

_BATCHPREDICTION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.BatchPrediction.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREPROCESSING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROCESSING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTPROCESSING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=21333,
  serialized_end=21502,
)
_sym_db.RegisterEnumDescriptor(_BATCHPREDICTION_STATE)


_INGESTIONERRORDETAIL = _descriptor.Descriptor(
  name='IngestionErrorDetail',
  full_name='continual.rpc.management.v1.IngestionErrorDetail',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.IngestionErrorDetail.description', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='continual.rpc.management.v1.IngestionErrorDetail.count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=185,
  serialized_end=243,
)


_INGESTION = _descriptor.Descriptor(
  name='Ingestion',
  full_name='continual.rpc.management.v1.Ingestion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Ingestion.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Ingestion.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ingested_row_count', full_name='continual.rpc.management.v1.Ingestion.ingested_row_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ingested_bytes', full_name='continual.rpc.management.v1.Ingestion.ingested_bytes', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_details', full_name='continual.rpc.management.v1.Ingestion.error_details', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Ingestion.create_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.Ingestion.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finish_time', full_name='continual.rpc.management.v1.Ingestion.finish_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _INGESTION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=246,
  serialized_end=720,
)


_USER = _descriptor.Descriptor(
  name='User',
  full_name='continual.rpc.management.v1.User',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.User.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.User.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email_verified', full_name='continual.rpc.management.v1.User.email_verified', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='full_name', full_name='continual.rpc.management.v1.User.full_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bio', full_name='continual.rpc.management.v1.User.bio', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='location', full_name='continual.rpc.management.v1.User.location', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.User.password', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_account', full_name='continual.rpc.management.v1.User.service_account', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.User.update_time', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.User.create_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disabled', full_name='continual.rpc.management.v1.User.disabled', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_available', full_name='continual.rpc.management.v1.User.trial_available', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='continual.rpc.management.v1.User.first_name', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_name', full_name='continual.rpc.management.v1.User.last_name', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=723,
  serialized_end=1075,
)


_SESSION = _descriptor.Descriptor(
  name='Session',
  full_name='continual.rpc.management.v1.Session',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Session.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='continual.rpc.management.v1.Session.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_access_ip', full_name='continual.rpc.management.v1.Session.last_access_ip', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_access_time', full_name='continual.rpc.management.v1.Session.last_access_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_ip', full_name='continual.rpc.management.v1.Session.login_ip', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_time', full_name='continual.rpc.management.v1.Session.login_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_user_agent', full_name='continual.rpc.management.v1.Session.login_user_agent', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1078,
  serialized_end=1323,
)


_ROLE = _descriptor.Descriptor(
  name='Role',
  full_name='continual.rpc.management.v1.Role',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Role.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.Role.title', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Role.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_type', full_name='continual.rpc.management.v1.Role.resource_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='permissions', full_name='continual.rpc.management.v1.Role.permissions', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1325,
  serialized_end=1425,
)


_ACCESSPOLICY = _descriptor.Descriptor(
  name='AccessPolicy',
  full_name='continual.rpc.management.v1.AccessPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.AccessPolicy.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='continual.rpc.management.v1.AccessPolicy.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.AccessPolicy.resource', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='continual.rpc.management.v1.AccessPolicy.role', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004Role', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.AccessPolicy.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1428,
  serialized_end=1572,
)


_INVITE = _descriptor.Descriptor(
  name='Invite',
  full_name='continual.rpc.management.v1.Invite',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Invite.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='continual.rpc.management.v1.Invite.sender', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.Invite.email', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='continual.rpc.management.v1.Invite.role', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004Role', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Invite.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expire_time', full_name='continual.rpc.management.v1.Invite.expire_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1575,
  serialized_end=1771,
)


_ORGANIZATION = _descriptor.Descriptor(
  name='Organization',
  full_name='continual.rpc.management.v1.Organization',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Organization.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Organization.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Organization.update_time', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Organization.create_time', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='continual.rpc.management.v1.Organization.status', index=4,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_end_date', full_name='continual.rpc.management.v1.Organization.trial_end_date', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_credits', full_name='continual.rpc.management.v1.Organization.trial_credits', index=6,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_credits_used', full_name='continual.rpc.management.v1.Organization.trial_credits_used', index=7,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='show_plan', full_name='continual.rpc.management.v1.Organization.show_plan', index=8,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ORGANIZATION_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1774,
  serialized_end=2207,
)


_ONBOARDING = _descriptor.Descriptor(
  name='Onboarding',
  full_name='continual.rpc.management.v1.Onboarding',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Onboarding.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Onboarding.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Onboarding.update_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_created', full_name='continual.rpc.management.v1.Onboarding.project_created', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_created', full_name='continual.rpc.management.v1.Onboarding.model_created', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_created', full_name='continual.rpc.management.v1.Onboarding.feature_set_created', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='billing_setup', full_name='continual.rpc.management.v1.Onboarding.billing_setup', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dismissed', full_name='continual.rpc.management.v1.Onboarding.dismissed', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2210,
  serialized_end=2463,
)


_PROJECT = _descriptor.Descriptor(
  name='Project',
  full_name='continual.rpc.management.v1.Project',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Project.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Project.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.Project.organization', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014Organization', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.Project.summary', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_store', full_name='continual.rpc.management.v1.Project.data_store', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Project.update_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Project.create_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2466,
  serialized_end=2785,
)


_PROJECTSUMMARY = _descriptor.Descriptor(
  name='ProjectSummary',
  full_name='continual.rpc.management.v1.ProjectSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set_count', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_row_count', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_row_count', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_bytes', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_bytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_count', full_name='continual.rpc.management.v1.ProjectSummary.feature_count', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='connection_count', full_name='continual.rpc.management.v1.ProjectSummary.connection_count', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_count', full_name='continual.rpc.management.v1.ProjectSummary.model_count', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version_count', full_name='continual.rpc.management.v1.ProjectSummary.model_version_count', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiment_count', full_name='continual.rpc.management.v1.ProjectSummary.experiment_count', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_count', full_name='continual.rpc.management.v1.ProjectSummary.prediction_count', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_health', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_health', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_health', full_name='continual.rpc.management.v1.ProjectSummary.model_health', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_ingestion_time', full_name='continual.rpc.management.v1.ProjectSummary.last_ingestion_time', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_deployment_time', full_name='continual.rpc.management.v1.ProjectSummary.last_deployment_time', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_training_time', full_name='continual.rpc.management.v1.ProjectSummary.last_training_time', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2788,
  serialized_end=3347,
)


_DATASTORE_BIGQUERY = _descriptor.Descriptor(
  name='BigQuery',
  full_name='continual.rpc.management.v1.DataStore.BigQuery',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='auth_file', full_name='continual.rpc.management.v1.DataStore.BigQuery.auth_file', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_file_name', full_name='continual.rpc.management.v1.DataStore.BigQuery.auth_file_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='continual.rpc.management.v1.DataStore.BigQuery.dataset', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3757,
  serialized_end=3827,
)

_DATASTORE_SNOWFLAKE_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4055,
  serialized_end=4100,
)

_DATASTORE_SNOWFLAKE = _descriptor.Descriptor(
  name='Snowflake',
  full_name='continual.rpc.management.v1.DataStore.Snowflake',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Snowflake.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Snowflake.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='username', full_name='continual.rpc.management.v1.DataStore.Snowflake.username', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.DataStore.Snowflake.password', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Snowflake.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse', full_name='continual.rpc.management.v1.DataStore.Snowflake.warehouse', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='continual.rpc.management.v1.DataStore.Snowflake.role', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Snowflake.params', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Snowflake.db_schema', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_SNOWFLAKE_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3830,
  serialized_end=4100,
)

_DATASTORE_POSTGRES_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Postgres.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Postgres.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Postgres.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4055,
  serialized_end=4100,
)

_DATASTORE_POSTGRES = _descriptor.Descriptor(
  name='Postgres',
  full_name='continual.rpc.management.v1.DataStore.Postgres',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Postgres.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Postgres.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='username', full_name='continual.rpc.management.v1.DataStore.Postgres.username', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.DataStore.Postgres.password', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Postgres.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Postgres.params', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Postgres.db_schema', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_POSTGRES_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4103,
  serialized_end=4338,
)

_DATASTORE_REDSHIFT_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Redshift.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Redshift.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Redshift.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4055,
  serialized_end=4100,
)

_DATASTORE_REDSHIFT = _descriptor.Descriptor(
  name='Redshift',
  full_name='continual.rpc.management.v1.DataStore.Redshift',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Redshift.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Redshift.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='username', full_name='continual.rpc.management.v1.DataStore.Redshift.username', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.DataStore.Redshift.password', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Redshift.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Redshift.params', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Redshift.db_schema', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_REDSHIFT_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4341,
  serialized_end=4576,
)

_DATASTORE_DATABRICKS_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Databricks.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Databricks.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Databricks.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4055,
  serialized_end=4100,
)

_DATASTORE_DATABRICKS = _descriptor.Descriptor(
  name='Databricks',
  full_name='continual.rpc.management.v1.DataStore.Databricks',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Databricks.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Databricks.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='token', full_name='continual.rpc.management.v1.DataStore.Databricks.token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='http_path', full_name='continual.rpc.management.v1.DataStore.Databricks.http_path', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Databricks.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Databricks.params', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Databricks.db_schema', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_DATABRICKS_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4579,
  serialized_end=4816,
)

_DATASTORE = _descriptor.Descriptor(
  name='DataStore',
  full_name='continual.rpc.management.v1.DataStore',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.DataStore.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='big_query', full_name='continual.rpc.management.v1.DataStore.big_query', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snowflake', full_name='continual.rpc.management.v1.DataStore.snowflake', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='postgres', full_name='continual.rpc.management.v1.DataStore.postgres', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='redshift', full_name='continual.rpc.management.v1.DataStore.redshift', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='databricks', full_name='continual.rpc.management.v1.DataStore.databricks', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_demo_data_store', full_name='continual.rpc.management.v1.DataStore.is_demo_data_store', index=6,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_BIGQUERY, _DATASTORE_SNOWFLAKE, _DATASTORE_POSTGRES, _DATASTORE_REDSHIFT, _DATASTORE_DATABRICKS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='config', full_name='continual.rpc.management.v1.DataStore.config',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3350,
  serialized_end=4826,
)


_HEALTHCOUNTS = _descriptor.Descriptor(
  name='HealthCounts',
  full_name='continual.rpc.management.v1.HealthCounts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='healthy_count', full_name='continual.rpc.management.v1.HealthCounts.healthy_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unhealthy_count', full_name='continual.rpc.management.v1.HealthCounts.unhealthy_count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='critical_count', full_name='continual.rpc.management.v1.HealthCounts.critical_count', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4828,
  serialized_end=4914,
)


_ENVIRONMENT = _descriptor.Descriptor(
  name='Environment',
  full_name='continual.rpc.management.v1.Environment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Environment.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.Environment.summary', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_store', full_name='continual.rpc.management.v1.Environment.data_store', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Environment.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Environment.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheduling_enabled', full_name='continual.rpc.management.v1.Environment.scheduling_enabled', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.Environment.source', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4917,
  serialized_end=5223,
)


_FEATURESETSCHEMA = _descriptor.Descriptor(
  name='FeatureSetSchema',
  full_name='continual.rpc.management.v1.FeatureSetSchema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.FeatureSetSchema.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.FeatureSetSchema.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity', full_name='continual.rpc.management.v1.FeatureSetSchema.entity', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.FeatureSetSchema.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.FeatureSetSchema.url', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='owners', full_name='continual.rpc.management.v1.FeatureSetSchema.owners', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='documentation', full_name='continual.rpc.management.v1.FeatureSetSchema.documentation', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='profile', full_name='continual.rpc.management.v1.FeatureSetSchema.profile', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='columns', full_name='continual.rpc.management.v1.FeatureSetSchema.columns', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exclude_columns', full_name='continual.rpc.management.v1.FeatureSetSchema.exclude_columns', index=9,
      number=10, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.FeatureSetSchema.query', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index', full_name='continual.rpc.management.v1.FeatureSetSchema.index', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_index', full_name='continual.rpc.management.v1.FeatureSetSchema.time_index', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='split', full_name='continual.rpc.management.v1.FeatureSetSchema.split', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.FeatureSetSchema.table', index=14,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5226,
  serialized_end=5587,
)


_MODELSCHEMA = _descriptor.Descriptor(
  name='ModelSchema',
  full_name='continual.rpc.management.v1.ModelSchema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.ModelSchema.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ModelSchema.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.ModelSchema.description', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.ModelSchema.url', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='owners', full_name='continual.rpc.management.v1.ModelSchema.owners', index=4,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='documentation', full_name='continual.rpc.management.v1.ModelSchema.documentation', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target', full_name='continual.rpc.management.v1.ModelSchema.target', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='train', full_name='continual.rpc.management.v1.ModelSchema.train', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='predict', full_name='continual.rpc.management.v1.ModelSchema.predict', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promote', full_name='continual.rpc.management.v1.ModelSchema.promote', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='profile', full_name='continual.rpc.management.v1.ModelSchema.profile', index=10,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='columns', full_name='continual.rpc.management.v1.ModelSchema.columns', index=11,
      number=13, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exclude_columns', full_name='continual.rpc.management.v1.ModelSchema.exclude_columns', index=12,
      number=14, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_feature_sets', full_name='continual.rpc.management.v1.ModelSchema.include_feature_sets', index=13,
      number=15, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.ModelSchema.query', index=14,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index', full_name='continual.rpc.management.v1.ModelSchema.index', index=15,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_index', full_name='continual.rpc.management.v1.ModelSchema.time_index', index=16,
      number=18, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='split', full_name='continual.rpc.management.v1.ModelSchema.split', index=17,
      number=19, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.ModelSchema.table', index=18,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5590,
  serialized_end=6155,
)


_COLUMNCONFIG = _descriptor.Descriptor(
  name='ColumnConfig',
  full_name='continual.rpc.management.v1.ColumnConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ColumnConfig.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.ColumnConfig.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity', full_name='continual.rpc.management.v1.ColumnConfig.entity', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.ColumnConfig.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='continual.rpc.management.v1.ColumnConfig.dtype', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6158,
  serialized_end=6292,
)


_TRAINCONFIG = _descriptor.Descriptor(
  name='TrainConfig',
  full_name='continual.rpc.management.v1.TrainConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schedule', full_name='continual.rpc.management.v1.TrainConfig.schedule', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metric', full_name='continual.rpc.management.v1.TrainConfig.metric', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='optimization', full_name='continual.rpc.management.v1.TrainConfig.optimization', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='advanced', full_name='continual.rpc.management.v1.TrainConfig.advanced', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='included_model_types', full_name='continual.rpc.management.v1.TrainConfig.included_model_types', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='excluded_model_types', full_name='continual.rpc.management.v1.TrainConfig.excluded_model_types', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='continual.rpc.management.v1.TrainConfig.size', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plots', full_name='continual.rpc.management.v1.TrainConfig.plots', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exclude_ensemble', full_name='continual.rpc.management.v1.TrainConfig.exclude_ensemble', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disable_data_checks', full_name='continual.rpc.management.v1.TrainConfig.disable_data_checks', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_limit', full_name='continual.rpc.management.v1.TrainConfig.time_limit', index=10,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size_limit', full_name='continual.rpc.management.v1.TrainConfig.size_limit', index=11,
      number=12, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6295,
  serialized_end=6566,
)


_PREDICTCONFIG = _descriptor.Descriptor(
  name='PredictConfig',
  full_name='continual.rpc.management.v1.PredictConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schedule', full_name='continual.rpc.management.v1.PredictConfig.schedule', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='incremental', full_name='continual.rpc.management.v1.PredictConfig.incremental', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6568,
  serialized_end=6622,
)


_PROMOTECONFIG = _descriptor.Descriptor(
  name='PromoteConfig',
  full_name='continual.rpc.management.v1.PromoteConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='policy', full_name='continual.rpc.management.v1.PromoteConfig.policy', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6624,
  serialized_end=6655,
)


_PROFILECONFIG = _descriptor.Descriptor(
  name='ProfileConfig',
  full_name='continual.rpc.management.v1.ProfileConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schedule', full_name='continual.rpc.management.v1.ProfileConfig.schedule', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6657,
  serialized_end=6690,
)


_SOURCECONFIG = _descriptor.Descriptor(
  name='SourceConfig',
  full_name='continual.rpc.management.v1.SourceConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='connection', full_name='continual.rpc.management.v1.SourceConfig.connection', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.SourceConfig.query', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schedule', full_name='continual.rpc.management.v1.SourceConfig.schedule', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file', full_name='continual.rpc.management.v1.SourceConfig.file', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='incremental_column', full_name='continual.rpc.management.v1.SourceConfig.incremental_column', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6692,
  serialized_end=6801,
)


_FEATURESET = _descriptor.Descriptor(
  name='FeatureSet',
  full_name='continual.rpc.management.v1.FeatureSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.FeatureSet.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='continual.rpc.management.v1.FeatureSet.schema', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.FeatureSet.schema_text', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.FeatureSet.create_time', index=3,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.FeatureSet.update_time', index=4,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.FeatureSet.table', index=5,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6804,
  serialized_end=7032,
)


_COLUMNSTATS = _descriptor.Descriptor(
  name='ColumnStats',
  full_name='continual.rpc.management.v1.ColumnStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.ColumnStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.ColumnStats.distinct_values', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='continual.rpc.management.v1.ColumnStats.min', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='continual.rpc.management.v1.ColumnStats.max', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sd', full_name='continual.rpc.management.v1.ColumnStats.sd', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mean', full_name='continual.rpc.management.v1.ColumnStats.mean', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='continual.rpc.management.v1.ColumnStats.count', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.ColumnStats.null_count', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_length', full_name='continual.rpc.management.v1.ColumnStats.min_length', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_length', full_name='continual.rpc.management.v1.ColumnStats.max_length', index=9,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='earliest', full_name='continual.rpc.management.v1.ColumnStats.earliest', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ColumnStats.latest', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.ColumnStats.type', index=12,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7035,
  serialized_end=7365,
)


_COLUMNSETSTATS = _descriptor.Descriptor(
  name='ColumnSetStats',
  full_name='continual.rpc.management.v1.ColumnSetStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ColumnSetStats.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='computed_at', full_name='continual.rpc.management.v1.ColumnSetStats.computed_at', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='row_count', full_name='continual.rpc.management.v1.ColumnSetStats.row_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='total_bytes', full_name='continual.rpc.management.v1.ColumnSetStats.total_bytes', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='columns', full_name='continual.rpc.management.v1.ColumnSetStats.columns', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_time_index', full_name='continual.rpc.management.v1.ColumnSetStats.has_time_index', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ColumnSetStats.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ColumnSetStats.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7368,
  serialized_end=7664,
)


_EVENT_DATAENTRY = _descriptor.Descriptor(
  name='DataEntry',
  full_name='continual.rpc.management.v1.Event.DataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.Event.DataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.Event.DataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7893,
  serialized_end=7936,
)

_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='continual.rpc.management.v1.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Event.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='continual.rpc.management.v1.Event.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event', full_name='continual.rpc.management.v1.Event.event', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.Event.resource', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='severity', full_name='continual.rpc.management.v1.Event.severity', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.Event.message', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.Event.data', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Event.create_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EVENT_DATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7667,
  serialized_end=7936,
)


_SERVINGMETRIC = _descriptor.Descriptor(
  name='ServingMetric',
  full_name='continual.rpc.management.v1.ServingMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.ServingMetric.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ServingMetric.parent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tag', full_name='continual.rpc.management.v1.ServingMetric.tag', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_start_time', full_name='continual.rpc.management.v1.ServingMetric.request_start_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_start_time', full_name='continual.rpc.management.v1.ServingMetric.model_start_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_duration', full_name='continual.rpc.management.v1.ServingMetric.request_duration', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='route_duration', full_name='continual.rpc.management.v1.ServingMetric.route_duration', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_duration', full_name='continual.rpc.management.v1.ServingMetric.feature_duration', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='predict_duration', full_name='continual.rpc.management.v1.ServingMetric.predict_duration', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_duration', full_name='continual.rpc.management.v1.ServingMetric.model_duration', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.ServingMetric.model_version', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result_code', full_name='continual.rpc.management.v1.ServingMetric.result_code', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result_data', full_name='continual.rpc.management.v1.ServingMetric.result_data', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7939,
  serialized_end=8434,
)


_MODEL = _descriptor.Descriptor(
  name='Model',
  full_name='continual.rpc.management.v1.Model',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Model.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Model.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='continual.rpc.management.v1.Model.schema', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.Model.schema_text', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_model_version', full_name='continual.rpc.management.v1.Model.latest_model_version', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Model.create_time', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Model.update_time', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_batch_prediction', full_name='continual.rpc.management.v1.Model.latest_batch_prediction', index=7,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\017BatchPrediction', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_version', full_name='continual.rpc.management.v1.Model.current_version', index=8,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.Model.table', index=9,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MODEL_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8437,
  serialized_end=8914,
)


_PROMOTION = _descriptor.Descriptor(
  name='Promotion',
  full_name='continual.rpc.management.v1.Promotion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Promotion.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Promotion.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promoted_time', full_name='continual.rpc.management.v1.Promotion.promoted_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='demoted_time', full_name='continual.rpc.management.v1.Promotion.demoted_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.Promotion.model_version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_model_version', full_name='continual.rpc.management.v1.Promotion.base_model_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Promotion.state', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='improvement_metric', full_name='continual.rpc.management.v1.Promotion.improvement_metric', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='improvement_metric_value', full_name='continual.rpc.management.v1.Promotion.improvement_metric_value', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_improvement_metric_value', full_name='continual.rpc.management.v1.Promotion.base_improvement_metric_value', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='improvement_metric_diff', full_name='continual.rpc.management.v1.Promotion.improvement_metric_diff', index=10,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='decision', full_name='continual.rpc.management.v1.Promotion.decision', index=11,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='continual.rpc.management.v1.Promotion.reason', index=12,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.Promotion.message', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='manual', full_name='continual.rpc.management.v1.Promotion.manual', index=14,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account', full_name='continual.rpc.management.v1.Promotion.account', index=15,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PROMOTION_DECISION,
    _PROMOTION_REASON,
    _PROMOTION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8917,
  serialized_end=9855,
)


_MODELVERSION_LOGFILEPATHSENTRY = _descriptor.Descriptor(
  name='LogFilePathsEntry',
  full_name='continual.rpc.management.v1.ModelVersion.LogFilePathsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.ModelVersion.LogFilePathsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.ModelVersion.LogFilePathsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11106,
  serialized_end=11157,
)

_MODELVERSION_PLOTFILESENTRY = _descriptor.Descriptor(
  name='PlotFilesEntry',
  full_name='continual.rpc.management.v1.ModelVersion.PlotFilesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.ModelVersion.PlotFilesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.ModelVersion.PlotFilesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11159,
  serialized_end=11207,
)

_MODELVERSION = _descriptor.Descriptor(
  name='ModelVersion',
  full_name='continual.rpc.management.v1.ModelVersion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ModelVersion.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version_id', full_name='continual.rpc.management.v1.ModelVersion.model_version_id', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_signature', full_name='continual.rpc.management.v1.ModelVersion.model_signature', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiment', full_name='continual.rpc.management.v1.ModelVersion.experiment', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\nExperiment', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='storage_uri', full_name='continual.rpc.management.v1.ModelVersion.storage_uri', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version_format', full_name='continual.rpc.management.v1.ModelVersion.model_version_format', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.ModelVersion.state', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.ModelVersion.error_message', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace', full_name='continual.rpc.management.v1.ModelVersion.stack_trace', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='system_time', full_name='continual.rpc.management.v1.ModelVersion.system_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visible_start_time', full_name='continual.rpc.management.v1.ModelVersion.visible_start_time', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visible_stop_time', full_name='continual.rpc.management.v1.ModelVersion.visible_stop_time', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.ModelVersion.create_time', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.ModelVersion.update_time', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ModelVersion.start_time', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ModelVersion.end_time', index=15,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='platform_version', full_name='continual.rpc.management.v1.ModelVersion.platform_version', index=16,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training_row_count', full_name='continual.rpc.management.v1.ModelVersion.training_row_count', index=17,
      number=18, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='validation_row_count', full_name='continual.rpc.management.v1.ModelVersion.validation_row_count', index=18,
      number=19, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_row_count', full_name='continual.rpc.management.v1.ModelVersion.test_row_count', index=19,
      number=20, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric', full_name='continual.rpc.management.v1.ModelVersion.performance_metric', index=20,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric_val', full_name='continual.rpc.management.v1.ModelVersion.performance_metric_val', index=21,
      number=22, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_csv_path', full_name='continual.rpc.management.v1.ModelVersion.prediction_csv_path', index=22,
      number=23, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fit_config', full_name='continual.rpc.management.v1.ModelVersion.fit_config', index=23,
      number=24, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promotion', full_name='continual.rpc.management.v1.ModelVersion.promotion', index=24,
      number=25, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\tPromotion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promoted', full_name='continual.rpc.management.v1.ModelVersion.promoted', index=25,
      number=26, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='log_file_paths', full_name='continual.rpc.management.v1.ModelVersion.log_file_paths', index=26,
      number=27, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plot_files', full_name='continual.rpc.management.v1.ModelVersion.plot_files', index=27,
      number=28, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_checks', full_name='continual.rpc.management.v1.ModelVersion.data_checks', index=28,
      number=29, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credits_used', full_name='continual.rpc.management.v1.ModelVersion.credits_used', index=29,
      number=30, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_query', full_name='continual.rpc.management.v1.ModelVersion.model_query', index=30,
      number=31, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MODELVERSION_LOGFILEPATHSENTRY, _MODELVERSION_PLOTFILESENTRY, ],
  enum_types=[
    _MODELVERSION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9858,
  serialized_end=11392,
)


_CONFUSIONMATRIX_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='continual.rpc.management.v1.ConfusionMatrix.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='example_count', full_name='continual.rpc.management.v1.ConfusionMatrix.Row.example_count', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11499,
  serialized_end=11527,
)

_CONFUSIONMATRIX = _descriptor.Descriptor(
  name='ConfusionMatrix',
  full_name='continual.rpc.management.v1.ConfusionMatrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.ConfusionMatrix.display_name', index=0,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='row', full_name='continual.rpc.management.v1.ConfusionMatrix.row', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CONFUSIONMATRIX_ROW, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11395,
  serialized_end=11527,
)


_BOOLEANEVALUATIONMETRICS = _descriptor.Descriptor(
  name='BooleanEvaluationMetrics',
  full_name='continual.rpc.management.v1.BooleanEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_loss', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.log_loss', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='accuracy', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.accuracy', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roc_auc', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.roc_auc', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.error', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='confusion_matrix', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.confusion_matrix', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.precision', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.recall', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.f1', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11530,
  serialized_end=11743,
)


_TEXTEVALUATIONMETRICS = _descriptor.Descriptor(
  name='TextEvaluationMetrics',
  full_name='continual.rpc.management.v1.TextEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_loss', full_name='continual.rpc.management.v1.TextEvaluationMetrics.log_loss', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1', full_name='continual.rpc.management.v1.TextEvaluationMetrics.f1', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perplexity', full_name='continual.rpc.management.v1.TextEvaluationMetrics.perplexity', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='edit_distance', full_name='continual.rpc.management.v1.TextEvaluationMetrics.edit_distance', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='token_accuracy', full_name='continual.rpc.management.v1.TextEvaluationMetrics.token_accuracy', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bleu', full_name='continual.rpc.management.v1.TextEvaluationMetrics.bleu', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11746,
  serialized_end=11880,
)


_CATEGORICALEVALUATIONMETRICS = _descriptor.Descriptor(
  name='CategoricalEvaluationMetrics',
  full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_loss', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.log_loss', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='accuracy', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.accuracy', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roc_auc', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.roc_auc', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='multiclass_error', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.multiclass_error', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='confusion_matrix', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.confusion_matrix', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_micro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision_micro', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_macro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision_macro', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_weighted', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision_weighted', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall_micro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall_micro', index=10,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall_macro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall_macro', index=11,
      number=12, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall_weighted', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall_weighted', index=12,
      number=13, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1', index=13,
      number=14, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1_micro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1_micro', index=14,
      number=15, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1_macro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1_macro', index=15,
      number=16, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1_weighted', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1_weighted', index=16,
      number=17, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='balanced_accuracy', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.balanced_accuracy', index=17,
      number=18, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='average_precision', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.average_precision', index=18,
      number=19, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11883,
  serialized_end=12369,
)


_NUMBEREVALUATIONMETRICS = _descriptor.Descriptor(
  name='NumberEvaluationMetrics',
  full_name='continual.rpc.management.v1.NumberEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='rmse', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.rmse', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mae', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.mae', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='r2', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.r2', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.precision', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.recall', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mape', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.mape', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='smape', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.smape', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mean_squared_error', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.mean_squared_error', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='median_absolute_error', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.median_absolute_error', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12372,
  serialized_end=12559,
)


_EVALUATIONMETRICS = _descriptor.Descriptor(
  name='EvaluationMetrics',
  full_name='continual.rpc.management.v1.EvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='boolean', full_name='continual.rpc.management.v1.EvaluationMetrics.boolean', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='number', full_name='continual.rpc.management.v1.EvaluationMetrics.number', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='continual.rpc.management.v1.EvaluationMetrics.text', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='categorical', full_name='continual.rpc.management.v1.EvaluationMetrics.categorical', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='metrics', full_name='continual.rpc.management.v1.EvaluationMetrics.metrics',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=12562,
  serialized_end=12888,
)


_METRICS = _descriptor.Descriptor(
  name='Metrics',
  full_name='continual.rpc.management.v1.Metrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='train_metrics', full_name='continual.rpc.management.v1.Metrics.train_metrics', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='validation_metrics', full_name='continual.rpc.management.v1.Metrics.validation_metrics', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_metrics', full_name='continual.rpc.management.v1.Metrics.test_metrics', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12891,
  serialized_end=13117,
)


_EXPERIMENT_FEATUREIMPORTANCEENTRY = _descriptor.Descriptor(
  name='FeatureImportanceEntry',
  full_name='continual.rpc.management.v1.Experiment.FeatureImportanceEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.Experiment.FeatureImportanceEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.Experiment.FeatureImportanceEntry.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13877,
  serialized_end=13933,
)

_EXPERIMENT_FEATUREIMPORTANCERELATIVEENTRY = _descriptor.Descriptor(
  name='FeatureImportanceRelativeEntry',
  full_name='continual.rpc.management.v1.Experiment.FeatureImportanceRelativeEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.Experiment.FeatureImportanceRelativeEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.Experiment.FeatureImportanceRelativeEntry.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13935,
  serialized_end=13999,
)

_EXPERIMENT = _descriptor.Descriptor(
  name='Experiment',
  full_name='continual.rpc.management.v1.Experiment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Experiment.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Experiment.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Experiment.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.Experiment.type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='algorithm', full_name='continual.rpc.management.v1.Experiment.algorithm', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Experiment.state', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Experiment.create_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.Experiment.start_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finish_time', full_name='continual.rpc.management.v1.Experiment.finish_time', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric', full_name='continual.rpc.management.v1.Experiment.performance_metric', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric_val', full_name='continual.rpc.management.v1.Experiment.performance_metric_val', index=10,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='continual.rpc.management.v1.Experiment.metrics', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_type', full_name='continual.rpc.management.v1.Experiment.model_type', index=12,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_importance', full_name='continual.rpc.management.v1.Experiment.feature_importance', index=13,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training_config', full_name='continual.rpc.management.v1.Experiment.training_config', index=14,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.Experiment.error_message', index=15,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace', full_name='continual.rpc.management.v1.Experiment.stack_trace', index=16,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_importance_relative', full_name='continual.rpc.management.v1.Experiment.feature_importance_relative', index=17,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EXPERIMENT_FEATUREIMPORTANCEENTRY, _EXPERIMENT_FEATUREIMPORTANCERELATIVEENTRY, ],
  enum_types=[
    _EXPERIMENT_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13120,
  serialized_end=14156,
)


_TRAININGMETRICS = _descriptor.Descriptor(
  name='TrainingMetrics',
  full_name='continual.rpc.management.v1.TrainingMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment', full_name='continual.rpc.management.v1.TrainingMetrics.experiment', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='step', full_name='continual.rpc.management.v1.TrainingMetrics.step', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trainer', full_name='continual.rpc.management.v1.TrainingMetrics.trainer', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='capture_time', full_name='continual.rpc.management.v1.TrainingMetrics.capture_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metric_type', full_name='continual.rpc.management.v1.TrainingMetrics.metric_type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='continual.rpc.management.v1.TrainingMetrics.metrics', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14159,
  serialized_end=14393,
)


_METRIC = _descriptor.Descriptor(
  name='Metric',
  full_name='continual.rpc.management.v1.Metric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='time', full_name='continual.rpc.management.v1.Metric.time', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.Metric.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='continual.rpc.management.v1.Metric.values', index=2,
      number=3, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14395,
  serialized_end=14476,
)


_BATCHPREDICTIONMETRIC = _descriptor.Descriptor(
  name='BatchPredictionMetric',
  full_name='continual.rpc.management.v1.BatchPredictionMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BatchPredictionMetric.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.BatchPredictionMetric.start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.BatchPredictionMetric.end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_count', full_name='continual.rpc.management.v1.BatchPredictionMetric.prediction_count', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.BatchPredictionMetric.success', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_code', full_name='continual.rpc.management.v1.BatchPredictionMetric.error_code', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14479,
  serialized_end=14673,
)


_TEXTSEQUENCE = _descriptor.Descriptor(
  name='TextSequence',
  full_name='continual.rpc.management.v1.TextSequence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.TextSequence.value', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14675,
  serialized_end=14704,
)


_NUMBERSEQUENCE = _descriptor.Descriptor(
  name='NumberSequence',
  full_name='continual.rpc.management.v1.NumberSequence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.NumberSequence.value', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14706,
  serialized_end=14737,
)


_SIGNATUREFEATURE = _descriptor.Descriptor(
  name='SignatureFeature',
  full_name='continual.rpc.management.v1.SignatureFeature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='qualified_name', full_name='continual.rpc.management.v1.SignatureFeature.qualified_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.SignatureFeature.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14739,
  serialized_end=14835,
)


_QUERYFEATURE = _descriptor.Descriptor(
  name='QueryFeature',
  full_name='continual.rpc.management.v1.QueryFeature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.QueryFeature.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qualified_name', full_name='continual.rpc.management.v1.QueryFeature.qualified_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='relation', full_name='continual.rpc.management.v1.QueryFeature.relation', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.QueryFeature.value', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.QueryFeature.type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='continual.rpc.management.v1.QueryFeature.dtype', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.QueryFeature.model', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index', full_name='continual.rpc.management.v1.QueryFeature.index', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_index', full_name='continual.rpc.management.v1.QueryFeature.time_index', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exclude', full_name='continual.rpc.management.v1.QueryFeature.exclude', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='split', full_name='continual.rpc.management.v1.QueryFeature.split', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14838,
  serialized_end=15110,
)


_QUERY = _descriptor.Descriptor(
  name='Query',
  full_name='continual.rpc.management.v1.Query',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Query.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity_id', full_name='continual.rpc.management.v1.Query.entity_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table_name', full_name='continual.rpc.management.v1.Query.table_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fields', full_name='continual.rpc.management.v1.Query.fields', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='related', full_name='continual.rpc.management.v1.Query.related', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15113,
  serialized_end=15285,
)


_CHANGESUMMARY = _descriptor.Descriptor(
  name='ChangeSummary',
  full_name='continual.rpc.management.v1.ChangeSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='models_added', full_name='continual.rpc.management.v1.ChangeSummary.models_added', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='models_modified', full_name='continual.rpc.management.v1.ChangeSummary.models_modified', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='models_deleted', full_name='continual.rpc.management.v1.ChangeSummary.models_deleted', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_versions_trained', full_name='continual.rpc.management.v1.ChangeSummary.model_versions_trained', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_sets_added', full_name='continual.rpc.management.v1.ChangeSummary.feature_sets_added', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_sets_modified', full_name='continual.rpc.management.v1.ChangeSummary.feature_sets_modified', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_sets_deleted', full_name='continual.rpc.management.v1.ChangeSummary.feature_sets_deleted', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='features_added', full_name='continual.rpc.management.v1.ChangeSummary.features_added', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='features_deleted', full_name='continual.rpc.management.v1.ChangeSummary.features_deleted', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='predictions_made', full_name='continual.rpc.management.v1.ChangeSummary.predictions_made', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15288,
  serialized_end=15571,
)


_CHANGE = _descriptor.Descriptor(
  name='Change',
  full_name='continual.rpc.management.v1.Change',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Change.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Change.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plan', full_name='continual.rpc.management.v1.Change.plan', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Change.create_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.Change.start_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finish_time', full_name='continual.rpc.management.v1.Change.finish_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by', full_name='continual.rpc.management.v1.Change.created_by', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cancelled_by', full_name='continual.rpc.management.v1.Change.cancelled_by', index=7,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.Change.message', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trigger', full_name='continual.rpc.management.v1.Change.trigger', index=9,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.Change.resource', index=10,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.Change.summary', index=11,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CHANGE_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15574,
  serialized_end=16126,
)


_PLANSTEP_CREATEMODEL = _descriptor.Descriptor(
  name='CreateModel',
  full_name='continual.rpc.management.v1.PlanStep.CreateModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.PlanStep.CreateModel.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\005Model', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.PlanStep.CreateModel.schema_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='added_features', full_name='continual.rpc.management.v1.PlanStep.CreateModel.added_features', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_views', full_name='continual.rpc.management.v1.PlanStep.CreateModel.created_views', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17433,
  serialized_end=17539,
)

_PLANSTEP_CREATEFEATURESET = _descriptor.Descriptor(
  name='CreateFeatureSet',
  full_name='continual.rpc.management.v1.PlanStep.CreateFeatureSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set', full_name='continual.rpc.management.v1.PlanStep.CreateFeatureSet.feature_set', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\nFeatureSet', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.PlanStep.CreateFeatureSet.schema_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='added_features', full_name='continual.rpc.management.v1.PlanStep.CreateFeatureSet.added_features', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_views', full_name='continual.rpc.management.v1.PlanStep.CreateFeatureSet.created_views', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17541,
  serialized_end=17663,
)

_PLANSTEP_UPDATEMODEL = _descriptor.Descriptor(
  name='UpdateModel',
  full_name='continual.rpc.management.v1.PlanStep.UpdateModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\005Model', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.schema_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_schema_text', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.previous_schema_text', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='added_features', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.added_features', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_features', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.updated_features', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='removed_features', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.removed_features', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_views', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.created_views', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_views', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.updated_views', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deleted_views', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.deleted_views', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updates', full_name='continual.rpc.management.v1.PlanStep.UpdateModel.updates', index=9,
      number=10, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PLANSTEP_UPDATEMODEL_UPDATETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17666,
  serialized_end=18210,
)

_PLANSTEP_UPDATEFEATURESET = _descriptor.Descriptor(
  name='UpdateFeatureSet',
  full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.feature_set', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\nFeatureSet', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.schema_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_schema_text', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.previous_schema_text', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='added_features', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.added_features', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_features', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.updated_features', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='removed_features', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.removed_features', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_views', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.created_views', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updated_views', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.updated_views', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deleted_views', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.deleted_views', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='updates', full_name='continual.rpc.management.v1.PlanStep.UpdateFeatureSet.updates', index=9,
      number=10, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PLANSTEP_UPDATEFEATURESET_UPDATETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18213,
  serialized_end=18724,
)

_PLANSTEP_PROFILEFEATURESET = _descriptor.Descriptor(
  name='ProfileFeatureSet',
  full_name='continual.rpc.management.v1.PlanStep.ProfileFeatureSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set', full_name='continual.rpc.management.v1.PlanStep.ProfileFeatureSet.feature_set', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\nFeatureSet', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rows_profiled', full_name='continual.rpc.management.v1.PlanStep.ProfileFeatureSet.rows_profiled', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_rows_profiled', full_name='continual.rpc.management.v1.PlanStep.ProfileFeatureSet.previous_rows_profiled', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_checks', full_name='continual.rpc.management.v1.PlanStep.ProfileFeatureSet.data_checks', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18727,
  serialized_end=18899,
)

_PLANSTEP_PROFILEMODEL = _descriptor.Descriptor(
  name='ProfileModel',
  full_name='continual.rpc.management.v1.PlanStep.ProfileModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.PlanStep.ProfileModel.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\005Model', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rows_profiled', full_name='continual.rpc.management.v1.PlanStep.ProfileModel.rows_profiled', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_rows_profiled', full_name='continual.rpc.management.v1.PlanStep.ProfileModel.previous_rows_profiled', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_checks', full_name='continual.rpc.management.v1.PlanStep.ProfileModel.data_checks', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18902,
  serialized_end=19058,
)

_PLANSTEP_PREDICT = _descriptor.Descriptor(
  name='Predict',
  full_name='continual.rpc.management.v1.PlanStep.Predict',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_prediction', full_name='continual.rpc.management.v1.PlanStep.Predict.batch_prediction', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\017BatchPrediction', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_batch_prediction', full_name='continual.rpc.management.v1.PlanStep.Predict.previous_batch_prediction', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\017BatchPrediction', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_count', full_name='continual.rpc.management.v1.PlanStep.Predict.prediction_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_prediction_count', full_name='continual.rpc.management.v1.PlanStep.Predict.previous_prediction_count', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19061,
  serialized_end=19232,
)

_PLANSTEP_TRAIN = _descriptor.Descriptor(
  name='Train',
  full_name='continual.rpc.management.v1.PlanStep.Train',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.PlanStep.Train.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_model_version', full_name='continual.rpc.management.v1.PlanStep.Train.base_model_version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_model_version_type', full_name='continual.rpc.management.v1.PlanStep.Train.base_model_version_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='continual.rpc.management.v1.PlanStep.Train.reason', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_model_version', full_name='continual.rpc.management.v1.PlanStep.Train.latest_model_version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promoted_model_version', full_name='continual.rpc.management.v1.PlanStep.Train.promoted_model_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='production_model_version', full_name='continual.rpc.management.v1.PlanStep.Train.production_model_version', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PLANSTEP_TRAIN_BASEMODELVERSIONTYPE,
    _PLANSTEP_TRAIN_REASON,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19235,
  serialized_end=19812,
)

_PLANSTEP_PROMOTE = _descriptor.Descriptor(
  name='Promote',
  full_name='continual.rpc.management.v1.PlanStep.Promote',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='promotion', full_name='continual.rpc.management.v1.PlanStep.Promote.promotion', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\tPromotion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19814,
  serialized_end=19856,
)

_PLANSTEP_DELETEMODEL = _descriptor.Descriptor(
  name='DeleteModel',
  full_name='continual.rpc.management.v1.PlanStep.DeleteModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='deleted_views', full_name='continual.rpc.management.v1.PlanStep.DeleteModel.deleted_views', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deleted_features', full_name='continual.rpc.management.v1.PlanStep.DeleteModel.deleted_features', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19858,
  serialized_end=19920,
)

_PLANSTEP_DELETEFEATURESET = _descriptor.Descriptor(
  name='DeleteFeatureSet',
  full_name='continual.rpc.management.v1.PlanStep.DeleteFeatureSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='deleted_views', full_name='continual.rpc.management.v1.PlanStep.DeleteFeatureSet.deleted_views', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deleted_features', full_name='continual.rpc.management.v1.PlanStep.DeleteFeatureSet.deleted_features', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19922,
  serialized_end=19989,
)

_PLANSTEP = _descriptor.Descriptor(
  name='PlanStep',
  full_name='continual.rpc.management.v1.PlanStep',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='continual.rpc.management.v1.PlanStep.id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.PlanStep.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.PlanStep.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='operation', full_name='continual.rpc.management.v1.PlanStep.operation', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.PlanStep.state', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_name', full_name='continual.rpc.management.v1.PlanStep.resource_name', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_name', full_name='continual.rpc.management.v1.PlanStep.object_name', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dependencies', full_name='continual.rpc.management.v1.PlanStep.dependencies', index=7,
      number=8, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_messages', full_name='continual.rpc.management.v1.PlanStep.error_messages', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.PlanStep.start_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finish_time', full_name='continual.rpc.management.v1.PlanStep.finish_time', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.PlanStep.schema_text', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previous_schema_text', full_name='continual.rpc.management.v1.PlanStep.previous_schema_text', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_feature_set', full_name='continual.rpc.management.v1.PlanStep.create_feature_set', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_model', full_name='continual.rpc.management.v1.PlanStep.create_model', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_feature_set', full_name='continual.rpc.management.v1.PlanStep.update_feature_set', index=15,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_model', full_name='continual.rpc.management.v1.PlanStep.update_model', index=16,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='profile_feature_set', full_name='continual.rpc.management.v1.PlanStep.profile_feature_set', index=17,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='profile_model', full_name='continual.rpc.management.v1.PlanStep.profile_model', index=18,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='train', full_name='continual.rpc.management.v1.PlanStep.train', index=19,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promote', full_name='continual.rpc.management.v1.PlanStep.promote', index=20,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='predict', full_name='continual.rpc.management.v1.PlanStep.predict', index=21,
      number=22, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delete_feature_set', full_name='continual.rpc.management.v1.PlanStep.delete_feature_set', index=22,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delete_model', full_name='continual.rpc.management.v1.PlanStep.delete_model', index=23,
      number=24, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PLANSTEP_CREATEMODEL, _PLANSTEP_CREATEFEATURESET, _PLANSTEP_UPDATEMODEL, _PLANSTEP_UPDATEFEATURESET, _PLANSTEP_PROFILEFEATURESET, _PLANSTEP_PROFILEMODEL, _PLANSTEP_PREDICT, _PLANSTEP_TRAIN, _PLANSTEP_PROMOTE, _PLANSTEP_DELETEMODEL, _PLANSTEP_DELETEFEATURESET, ],
  enum_types=[
    _PLANSTEP_STATE,
    _PLANSTEP_OPERATION,
    _PLANSTEP_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='metadata', full_name='continual.rpc.management.v1.PlanStep.metadata',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=16129,
  serialized_end=20557,
)


_BATCHPREDICTION_LOGFILEPATHSENTRY = _descriptor.Descriptor(
  name='LogFilePathsEntry',
  full_name='continual.rpc.management.v1.BatchPrediction.LogFilePathsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.BatchPrediction.LogFilePathsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.BatchPrediction.LogFilePathsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11106,
  serialized_end=11157,
)

_BATCHPREDICTION = _descriptor.Descriptor(
  name='BatchPrediction',
  full_name='continual.rpc.management.v1.BatchPrediction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BatchPrediction.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.BatchPrediction.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.BatchPrediction.model_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_type', full_name='continual.rpc.management.v1.BatchPrediction.source_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.BatchPrediction.source', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dest_type', full_name='continual.rpc.management.v1.BatchPrediction.dest_type', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='incremental', full_name='continual.rpc.management.v1.BatchPrediction.incremental', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='csv_file_path', full_name='continual.rpc.management.v1.BatchPrediction.csv_file_path', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.BatchPrediction.create_time', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.BatchPrediction.start_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.BatchPrediction.end_time', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_count', full_name='continual.rpc.management.v1.BatchPrediction.prediction_count', index=11,
      number=12, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.BatchPrediction.error_message', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace', full_name='continual.rpc.management.v1.BatchPrediction.stack_trace', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='log_file_paths', full_name='continual.rpc.management.v1.BatchPrediction.log_file_paths', index=14,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_model', full_name='continual.rpc.management.v1.BatchPrediction.prediction_model', index=15,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.BatchPrediction.table', index=16,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credits_used', full_name='continual.rpc.management.v1.BatchPrediction.credits_used', index=17,
      number=18, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_BATCHPREDICTION_LOGFILEPATHSENTRY, ],
  enum_types=[
    _BATCHPREDICTION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20560,
  serialized_end=21502,
)


_SIGNEDURL = _descriptor.Descriptor(
  name='SignedURL',
  full_name='continual.rpc.management.v1.SignedURL',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signed_url', full_name='continual.rpc.management.v1.SignedURL.signed_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21504,
  serialized_end=21535,
)


_ENTITY = _descriptor.Descriptor(
  name='Entity',
  full_name='continual.rpc.management.v1.Entity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Entity.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='models', full_name='continual.rpc.management.v1.Entity.models', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_sets', full_name='continual.rpc.management.v1.Entity.feature_sets', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21537,
  serialized_end=21597,
)


_DATASETSTATS = _descriptor.Descriptor(
  name='DatasetStats',
  full_name='continual.rpc.management.v1.DatasetStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DatasetStats.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_stats', full_name='continual.rpc.management.v1.DatasetStats.dataset_stats', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='profile_time', full_name='continual.rpc.management.v1.DatasetStats.profile_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_name', full_name='continual.rpc.management.v1.DatasetStats.resource_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21600,
  serialized_end=21772,
)


_DATASETSTATSENTRY = _descriptor.Descriptor(
  name='DatasetStatsEntry',
  full_name='continual.rpc.management.v1.DatasetStatsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entry_name', full_name='continual.rpc.management.v1.DatasetStatsEntry.entry_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_examples', full_name='continual.rpc.management.v1.DatasetStatsEntry.num_examples', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='correlation_matrix', full_name='continual.rpc.management.v1.DatasetStatsEntry.correlation_matrix', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.string_stats', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='numeric_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.numeric_stats', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.timestamp_stats', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='categorical_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.categorical_stats', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duplicate_rows', full_name='continual.rpc.management.v1.DatasetStatsEntry.duplicate_rows', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duplicate_indexes', full_name='continual.rpc.management.v1.DatasetStatsEntry.duplicate_indexes', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_ranges', full_name='continual.rpc.management.v1.DatasetStatsEntry.dataset_ranges', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21775,
  serialized_end=22304,
)


_DATASETRANGE = _descriptor.Descriptor(
  name='DatasetRange',
  full_name='continual.rpc.management.v1.DatasetRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DatasetRange.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.DatasetRange.start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.DatasetRange.end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22306,
  serialized_end=22428,
)


_CORRELATIONMATRIX_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='continual.rpc.management.v1.CorrelationMatrix.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='correlations', full_name='continual.rpc.management.v1.CorrelationMatrix.Row.correlations', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22541,
  serialized_end=22568,
)

_CORRELATIONMATRIX = _descriptor.Descriptor(
  name='CorrelationMatrix',
  full_name='continual.rpc.management.v1.CorrelationMatrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_names', full_name='continual.rpc.management.v1.CorrelationMatrix.display_names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rows', full_name='continual.rpc.management.v1.CorrelationMatrix.rows', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CORRELATIONMATRIX_ROW, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22431,
  serialized_end=22568,
)


_HISTOGRAM_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='continual.rpc.management.v1.Histogram.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='low_value', full_name='continual.rpc.management.v1.Histogram.Bucket.low_value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='high_value', full_name='continual.rpc.management.v1.Histogram.Bucket.high_value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_count', full_name='continual.rpc.management.v1.Histogram.Bucket.sample_count', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22688,
  serialized_end=22757,
)

_HISTOGRAM = _descriptor.Descriptor(
  name='Histogram',
  full_name='continual.rpc.management.v1.Histogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_nan', full_name='continual.rpc.management.v1.Histogram.num_nan', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_undefined', full_name='continual.rpc.management.v1.Histogram.num_undefined', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='buckets', full_name='continual.rpc.management.v1.Histogram.buckets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_HISTOGRAM_BUCKET, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22571,
  serialized_end=22757,
)


_TIMESTAMPHISTOGRAM_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='continual.rpc.management.v1.TimestampHistogram.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='low_value', full_name='continual.rpc.management.v1.TimestampHistogram.Bucket.low_value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='high_value', full_name='continual.rpc.management.v1.TimestampHistogram.Bucket.high_value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_count', full_name='continual.rpc.management.v1.TimestampHistogram.Bucket.sample_count', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22895,
  serialized_end=23020,
)

_TIMESTAMPHISTOGRAM = _descriptor.Descriptor(
  name='TimestampHistogram',
  full_name='continual.rpc.management.v1.TimestampHistogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_nan', full_name='continual.rpc.management.v1.TimestampHistogram.num_nan', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_undefined', full_name='continual.rpc.management.v1.TimestampHistogram.num_undefined', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='buckets', full_name='continual.rpc.management.v1.TimestampHistogram.buckets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TIMESTAMPHISTOGRAM_BUCKET, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22760,
  serialized_end=23020,
)


_CATEGORICALDISTRIBUTION_CATEGORY = _descriptor.Descriptor(
  name='Category',
  full_name='continual.rpc.management.v1.CategoricalDistribution.Category',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='continual.rpc.management.v1.CategoricalDistribution.Category.label', index=0,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_count', full_name='continual.rpc.management.v1.CategoricalDistribution.Category.sample_count', index=1,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23133,
  serialized_end=23180,
)

_CATEGORICALDISTRIBUTION = _descriptor.Descriptor(
  name='CategoricalDistribution',
  full_name='continual.rpc.management.v1.CategoricalDistribution',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='categories', full_name='continual.rpc.management.v1.CategoricalDistribution.categories', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CATEGORICALDISTRIBUTION_CATEGORY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23023,
  serialized_end=23180,
)


_STRINGSTATS = _descriptor.Descriptor(
  name='StringStats',
  full_name='continual.rpc.management.v1.StringStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.StringStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.StringStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.StringStats.null_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_length', full_name='continual.rpc.management.v1.StringStats.min_length', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_length', full_name='continual.rpc.management.v1.StringStats.max_length', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23182,
  serialized_end=23301,
)


_CATEGORICALSTATS = _descriptor.Descriptor(
  name='CategoricalStats',
  full_name='continual.rpc.management.v1.CategoricalStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.CategoricalStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.CategoricalStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.CategoricalStats.null_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='categorical_distribution', full_name='continual.rpc.management.v1.CategoricalStats.categorical_distribution', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23304,
  serialized_end=23476,
)


_TIMESTAMPSTATS = _descriptor.Descriptor(
  name='TimestampStats',
  full_name='continual.rpc.management.v1.TimestampStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.TimestampStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.TimestampStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.TimestampStats.null_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='continual.rpc.management.v1.TimestampStats.min', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='continual.rpc.management.v1.TimestampStats.max', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_histogram', full_name='continual.rpc.management.v1.TimestampStats.timestamp_histogram', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23479,
  serialized_end=23721,
)


_NUMERICSTATS = _descriptor.Descriptor(
  name='NumericStats',
  full_name='continual.rpc.management.v1.NumericStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.NumericStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.NumericStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='continual.rpc.management.v1.NumericStats.min', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='continual.rpc.management.v1.NumericStats.max', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sd', full_name='continual.rpc.management.v1.NumericStats.sd', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mean', full_name='continual.rpc.management.v1.NumericStats.mean', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='continual.rpc.management.v1.NumericStats.count', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentile25', full_name='continual.rpc.management.v1.NumericStats.percentile25', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentile50', full_name='continual.rpc.management.v1.NumericStats.percentile50', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentile75', full_name='continual.rpc.management.v1.NumericStats.percentile75', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.NumericStats.null_count', index=10,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_nan', full_name='continual.rpc.management.v1.NumericStats.num_nan', index=11,
      number=12, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_values_histogram', full_name='continual.rpc.management.v1.NumericStats.num_values_histogram', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23724,
  serialized_end=24024,
)


_BILLINGACCOUNT = _descriptor.Descriptor(
  name='BillingAccount',
  full_name='continual.rpc.management.v1.BillingAccount',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BillingAccount.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.BillingAccount.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='memo', full_name='continual.rpc.management.v1.BillingAccount.memo', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.BillingAccount.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.BillingAccount.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24027,
  serialized_end=24194,
)


_SUBSCRIPTION = _descriptor.Descriptor(
  name='Subscription',
  full_name='continual.rpc.management.v1.Subscription',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Subscription.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='price', full_name='continual.rpc.management.v1.Subscription.price', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\005Price', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payment_method', full_name='continual.rpc.management.v1.Subscription.payment_method', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\rPaymentMethod', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Subscription.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Subscription.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active', full_name='continual.rpc.management.v1.Subscription.active', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24197,
  serialized_end=24416,
)


_PROJECTUSAGERECORD_TYPECOUNTENTRY = _descriptor.Descriptor(
  name='TypeCountEntry',
  full_name='continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry.value', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24555,
  serialized_end=24603,
)

_PROJECTUSAGERECORD = _descriptor.Descriptor(
  name='ProjectUsageRecord',
  full_name='continual.rpc.management.v1.ProjectUsageRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ProjectUsageRecord.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credits', full_name='continual.rpc.management.v1.ProjectUsageRecord.credits', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='typeCount', full_name='continual.rpc.management.v1.ProjectUsageRecord.typeCount', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PROJECTUSAGERECORD_TYPECOUNTENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24419,
  serialized_end=24603,
)


_CREDITUSAGERECORD = _descriptor.Descriptor(
  name='CreditUsageRecord',
  full_name='continual.rpc.management.v1.CreditUsageRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='period', full_name='continual.rpc.management.v1.CreditUsageRecord.period', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credits', full_name='continual.rpc.management.v1.CreditUsageRecord.credits', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_usages', full_name='continual.rpc.management.v1.CreditUsageRecord.project_usages', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24606,
  serialized_end=24759,
)


_INVOICELINEITEM = _descriptor.Descriptor(
  name='InvoiceLineItem',
  full_name='continual.rpc.management.v1.InvoiceLineItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='continual.rpc.management.v1.InvoiceLineItem.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='amount', full_name='continual.rpc.management.v1.InvoiceLineItem.amount', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='quantity', full_name='continual.rpc.management.v1.InvoiceLineItem.quantity', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='currency', full_name='continual.rpc.management.v1.InvoiceLineItem.currency', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.InvoiceLineItem.description', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_start', full_name='continual.rpc.management.v1.InvoiceLineItem.period_start', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_end', full_name='continual.rpc.management.v1.InvoiceLineItem.period_end', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24762,
  serialized_end=24962,
)


_INVOICE = _descriptor.Descriptor(
  name='Invoice',
  full_name='continual.rpc.management.v1.Invoice',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Invoice.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date', full_name='continual.rpc.management.v1.Invoice.date', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='amount', full_name='continual.rpc.management.v1.Invoice.amount', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_start', full_name='continual.rpc.management.v1.Invoice.period_start', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_end', full_name='continual.rpc.management.v1.Invoice.period_end', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='continual.rpc.management.v1.Invoice.status', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invoice_url', full_name='continual.rpc.management.v1.Invoice.invoice_url', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_items', full_name='continual.rpc.management.v1.Invoice.line_items', index=7,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24965,
  serialized_end=25247,
)


_PRICE = _descriptor.Descriptor(
  name='Price',
  full_name='continual.rpc.management.v1.Price',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Price.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.Price.title', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Price.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='direct_invoice', full_name='continual.rpc.management.v1.Price.direct_invoice', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25249,
  serialized_end=25330,
)


_PAYMENTMETHOD_BILLINGDETAILS = _descriptor.Descriptor(
  name='BillingDetails',
  full_name='continual.rpc.management.v1.PaymentMethod.BillingDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='postal_code', full_name='continual.rpc.management.v1.PaymentMethod.BillingDetails.postal_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25511,
  serialized_end=25548,
)

_PAYMENTMETHOD_CARD = _descriptor.Descriptor(
  name='Card',
  full_name='continual.rpc.management.v1.PaymentMethod.Card',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='brand', full_name='continual.rpc.management.v1.PaymentMethod.Card.brand', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last4', full_name='continual.rpc.management.v1.PaymentMethod.Card.last4', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exp_year', full_name='continual.rpc.management.v1.PaymentMethod.Card.exp_year', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exp_month', full_name='continual.rpc.management.v1.PaymentMethod.Card.exp_month', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25550,
  serialized_end=25623,
)

_PAYMENTMETHOD = _descriptor.Descriptor(
  name='PaymentMethod',
  full_name='continual.rpc.management.v1.PaymentMethod',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.PaymentMethod.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='billing_details', full_name='continual.rpc.management.v1.PaymentMethod.billing_details', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='card', full_name='continual.rpc.management.v1.PaymentMethod.card', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PAYMENTMETHOD_BILLINGDETAILS, _PAYMENTMETHOD_CARD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25333,
  serialized_end=25623,
)


_DATACHECKS_DATAVALIDATIONRESULT = _descriptor.Descriptor(
  name='DataValidationResult',
  full_name='continual.rpc.management.v1.DataChecks.DataValidationResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.DataChecks.DataValidationResult.summary', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.DataChecks.DataValidationResult.success', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duration', full_name='continual.rpc.management.v1.DataChecks.DataValidationResult.duration', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors', full_name='continual.rpc.management.v1.DataChecks.DataValidationResult.errors', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warnings', full_name='continual.rpc.management.v1.DataChecks.DataValidationResult.warnings', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='infos', full_name='continual.rpc.management.v1.DataChecks.DataValidationResult.infos', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25719,
  serialized_end=25842,
)

_DATACHECKS = _descriptor.Descriptor(
  name='DataChecks',
  full_name='continual.rpc.management.v1.DataChecks',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='results', full_name='continual.rpc.management.v1.DataChecks.results', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATACHECKS_DATAVALIDATIONRESULT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25626,
  serialized_end=25842,
)

_INGESTION.fields_by_name['state'].enum_type = _INGESTION_STATE
_INGESTION.fields_by_name['error_details'].message_type = _INGESTIONERRORDETAIL
_INGESTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INGESTION.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INGESTION.fields_by_name['finish_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INGESTION_STATE.containing_type = _INGESTION
_USER.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USER.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SESSION.fields_by_name['last_access_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SESSION.fields_by_name['login_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ACCESSPOLICY.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVITE.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVITE.fields_by_name['expire_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION.fields_by_name['status'].enum_type = _ORGANIZATION_STATUS
_ORGANIZATION.fields_by_name['trial_end_date'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION_STATUS.containing_type = _ORGANIZATION
_ONBOARDING.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ONBOARDING.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECT.fields_by_name['summary'].message_type = _PROJECTSUMMARY
_PROJECT.fields_by_name['data_store'].message_type = _DATASTORE
_PROJECT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTSUMMARY.fields_by_name['feature_set_health'].message_type = _HEALTHCOUNTS
_PROJECTSUMMARY.fields_by_name['model_health'].message_type = _HEALTHCOUNTS
_PROJECTSUMMARY.fields_by_name['last_ingestion_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTSUMMARY.fields_by_name['last_deployment_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTSUMMARY.fields_by_name['last_training_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASTORE_BIGQUERY.containing_type = _DATASTORE
_DATASTORE_SNOWFLAKE_PARAMSENTRY.containing_type = _DATASTORE_SNOWFLAKE
_DATASTORE_SNOWFLAKE.fields_by_name['params'].message_type = _DATASTORE_SNOWFLAKE_PARAMSENTRY
_DATASTORE_SNOWFLAKE.containing_type = _DATASTORE
_DATASTORE_POSTGRES_PARAMSENTRY.containing_type = _DATASTORE_POSTGRES
_DATASTORE_POSTGRES.fields_by_name['params'].message_type = _DATASTORE_POSTGRES_PARAMSENTRY
_DATASTORE_POSTGRES.containing_type = _DATASTORE
_DATASTORE_REDSHIFT_PARAMSENTRY.containing_type = _DATASTORE_REDSHIFT
_DATASTORE_REDSHIFT.fields_by_name['params'].message_type = _DATASTORE_REDSHIFT_PARAMSENTRY
_DATASTORE_REDSHIFT.containing_type = _DATASTORE
_DATASTORE_DATABRICKS_PARAMSENTRY.containing_type = _DATASTORE_DATABRICKS
_DATASTORE_DATABRICKS.fields_by_name['params'].message_type = _DATASTORE_DATABRICKS_PARAMSENTRY
_DATASTORE_DATABRICKS.containing_type = _DATASTORE
_DATASTORE.fields_by_name['big_query'].message_type = _DATASTORE_BIGQUERY
_DATASTORE.fields_by_name['snowflake'].message_type = _DATASTORE_SNOWFLAKE
_DATASTORE.fields_by_name['postgres'].message_type = _DATASTORE_POSTGRES
_DATASTORE.fields_by_name['redshift'].message_type = _DATASTORE_REDSHIFT
_DATASTORE.fields_by_name['databricks'].message_type = _DATASTORE_DATABRICKS
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['big_query'])
_DATASTORE.fields_by_name['big_query'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['snowflake'])
_DATASTORE.fields_by_name['snowflake'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['postgres'])
_DATASTORE.fields_by_name['postgres'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['redshift'])
_DATASTORE.fields_by_name['redshift'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['databricks'])
_DATASTORE.fields_by_name['databricks'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_ENVIRONMENT.fields_by_name['summary'].message_type = _PROJECTSUMMARY
_ENVIRONMENT.fields_by_name['data_store'].message_type = _DATASTORE
_ENVIRONMENT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ENVIRONMENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FEATURESETSCHEMA.fields_by_name['profile'].message_type = _PROFILECONFIG
_FEATURESETSCHEMA.fields_by_name['columns'].message_type = _COLUMNCONFIG
_MODELSCHEMA.fields_by_name['train'].message_type = _TRAINCONFIG
_MODELSCHEMA.fields_by_name['predict'].message_type = _PREDICTCONFIG
_MODELSCHEMA.fields_by_name['promote'].message_type = _PROMOTECONFIG
_MODELSCHEMA.fields_by_name['profile'].message_type = _PROFILECONFIG
_MODELSCHEMA.fields_by_name['columns'].message_type = _COLUMNCONFIG
_COLUMNCONFIG.fields_by_name['type'].enum_type = _FIELDTYPE
_FEATURESET.fields_by_name['schema'].message_type = _FEATURESETSCHEMA
_FEATURESET.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FEATURESET.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COLUMNSTATS.fields_by_name['earliest'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COLUMNSTATS.fields_by_name['latest'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COLUMNSTATS.fields_by_name['type'].enum_type = _FIELDTYPE
_COLUMNSETSTATS.fields_by_name['computed_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COLUMNSETSTATS.fields_by_name['columns'].message_type = _COLUMNSTATS
_COLUMNSETSTATS.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COLUMNSETSTATS.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EVENT_DATAENTRY.containing_type = _EVENT
_EVENT.fields_by_name['data'].message_type = _EVENT_DATAENTRY
_EVENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGMETRIC.fields_by_name['request_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGMETRIC.fields_by_name['model_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGMETRIC.fields_by_name['request_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['route_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['feature_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['predict_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['model_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_MODEL.fields_by_name['state'].enum_type = _MODEL_STATE
_MODEL.fields_by_name['schema'].message_type = _MODELSCHEMA
_MODEL.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODEL.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODEL_STATE.containing_type = _MODEL
_PROMOTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['promoted_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['demoted_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['state'].enum_type = _PROMOTION_STATE
_PROMOTION.fields_by_name['decision'].enum_type = _PROMOTION_DECISION
_PROMOTION.fields_by_name['reason'].enum_type = _PROMOTION_REASON
_PROMOTION_DECISION.containing_type = _PROMOTION
_PROMOTION_REASON.containing_type = _PROMOTION
_PROMOTION_STATE.containing_type = _PROMOTION
_MODELVERSION_LOGFILEPATHSENTRY.containing_type = _MODELVERSION
_MODELVERSION_PLOTFILESENTRY.containing_type = _MODELVERSION
_MODELVERSION.fields_by_name['state'].enum_type = _MODELVERSION_STATE
_MODELVERSION.fields_by_name['system_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['visible_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['visible_stop_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['log_file_paths'].message_type = _MODELVERSION_LOGFILEPATHSENTRY
_MODELVERSION.fields_by_name['plot_files'].message_type = _MODELVERSION_PLOTFILESENTRY
_MODELVERSION.fields_by_name['data_checks'].message_type = _DATACHECKS
_MODELVERSION_STATE.containing_type = _MODELVERSION
_CONFUSIONMATRIX_ROW.containing_type = _CONFUSIONMATRIX
_CONFUSIONMATRIX.fields_by_name['row'].message_type = _CONFUSIONMATRIX_ROW
_BOOLEANEVALUATIONMETRICS.fields_by_name['confusion_matrix'].message_type = _CONFUSIONMATRIX
_CATEGORICALEVALUATIONMETRICS.fields_by_name['confusion_matrix'].message_type = _CONFUSIONMATRIX
_EVALUATIONMETRICS.fields_by_name['boolean'].message_type = _BOOLEANEVALUATIONMETRICS
_EVALUATIONMETRICS.fields_by_name['number'].message_type = _NUMBEREVALUATIONMETRICS
_EVALUATIONMETRICS.fields_by_name['text'].message_type = _TEXTEVALUATIONMETRICS
_EVALUATIONMETRICS.fields_by_name['categorical'].message_type = _CATEGORICALEVALUATIONMETRICS
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['boolean'])
_EVALUATIONMETRICS.fields_by_name['boolean'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['number'])
_EVALUATIONMETRICS.fields_by_name['number'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['text'])
_EVALUATIONMETRICS.fields_by_name['text'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['categorical'])
_EVALUATIONMETRICS.fields_by_name['categorical'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_METRICS.fields_by_name['train_metrics'].message_type = _EVALUATIONMETRICS
_METRICS.fields_by_name['validation_metrics'].message_type = _EVALUATIONMETRICS
_METRICS.fields_by_name['test_metrics'].message_type = _EVALUATIONMETRICS
_EXPERIMENT_FEATUREIMPORTANCEENTRY.containing_type = _EXPERIMENT
_EXPERIMENT_FEATUREIMPORTANCERELATIVEENTRY.containing_type = _EXPERIMENT
_EXPERIMENT.fields_by_name['state'].enum_type = _EXPERIMENT_STATE
_EXPERIMENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT.fields_by_name['finish_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT.fields_by_name['metrics'].message_type = _METRICS
_EXPERIMENT.fields_by_name['model_type'].enum_type = _FIELDTYPE
_EXPERIMENT.fields_by_name['feature_importance'].message_type = _EXPERIMENT_FEATUREIMPORTANCEENTRY
_EXPERIMENT.fields_by_name['feature_importance_relative'].message_type = _EXPERIMENT_FEATUREIMPORTANCERELATIVEENTRY
_EXPERIMENT_STATE.containing_type = _EXPERIMENT
_TRAININGMETRICS.fields_by_name['capture_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TRAININGMETRICS.fields_by_name['metric_type'].enum_type = _FIELDTYPE
_TRAININGMETRICS.fields_by_name['metrics'].message_type = _METRICS
_METRIC.fields_by_name['time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTIONMETRIC.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTIONMETRIC.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SIGNATUREFEATURE.fields_by_name['type'].enum_type = _FIELDTYPE
_QUERYFEATURE.fields_by_name['relation'].message_type = _QUERY
_QUERYFEATURE.fields_by_name['type'].enum_type = _FIELDTYPE
_QUERY.fields_by_name['fields'].message_type = _QUERYFEATURE
_QUERY.fields_by_name['related'].message_type = _QUERY
_CHANGE.fields_by_name['state'].enum_type = _CHANGE_STATE
_CHANGE.fields_by_name['plan'].message_type = _PLANSTEP
_CHANGE.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CHANGE.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CHANGE.fields_by_name['finish_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CHANGE.fields_by_name['summary'].message_type = _CHANGESUMMARY
_CHANGE_STATE.containing_type = _CHANGE
_PLANSTEP_CREATEMODEL.containing_type = _PLANSTEP
_PLANSTEP_CREATEFEATURESET.containing_type = _PLANSTEP
_PLANSTEP_UPDATEMODEL.fields_by_name['updates'].enum_type = _PLANSTEP_UPDATEMODEL_UPDATETYPE
_PLANSTEP_UPDATEMODEL.containing_type = _PLANSTEP
_PLANSTEP_UPDATEMODEL_UPDATETYPE.containing_type = _PLANSTEP_UPDATEMODEL
_PLANSTEP_UPDATEFEATURESET.fields_by_name['updates'].enum_type = _PLANSTEP_UPDATEFEATURESET_UPDATETYPE
_PLANSTEP_UPDATEFEATURESET.containing_type = _PLANSTEP
_PLANSTEP_UPDATEFEATURESET_UPDATETYPE.containing_type = _PLANSTEP_UPDATEFEATURESET
_PLANSTEP_PROFILEFEATURESET.fields_by_name['data_checks'].message_type = _DATACHECKS
_PLANSTEP_PROFILEFEATURESET.containing_type = _PLANSTEP
_PLANSTEP_PROFILEMODEL.fields_by_name['data_checks'].message_type = _DATACHECKS
_PLANSTEP_PROFILEMODEL.containing_type = _PLANSTEP
_PLANSTEP_PREDICT.containing_type = _PLANSTEP
_PLANSTEP_TRAIN.fields_by_name['base_model_version_type'].enum_type = _PLANSTEP_TRAIN_BASEMODELVERSIONTYPE
_PLANSTEP_TRAIN.fields_by_name['reason'].enum_type = _PLANSTEP_TRAIN_REASON
_PLANSTEP_TRAIN.containing_type = _PLANSTEP
_PLANSTEP_TRAIN_BASEMODELVERSIONTYPE.containing_type = _PLANSTEP_TRAIN
_PLANSTEP_TRAIN_REASON.containing_type = _PLANSTEP_TRAIN
_PLANSTEP_PROMOTE.containing_type = _PLANSTEP
_PLANSTEP_DELETEMODEL.containing_type = _PLANSTEP
_PLANSTEP_DELETEFEATURESET.containing_type = _PLANSTEP
_PLANSTEP.fields_by_name['type'].enum_type = _PLANSTEP_TYPE
_PLANSTEP.fields_by_name['operation'].enum_type = _PLANSTEP_OPERATION
_PLANSTEP.fields_by_name['state'].enum_type = _PLANSTEP_STATE
_PLANSTEP.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PLANSTEP.fields_by_name['finish_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PLANSTEP.fields_by_name['create_feature_set'].message_type = _PLANSTEP_CREATEFEATURESET
_PLANSTEP.fields_by_name['create_model'].message_type = _PLANSTEP_CREATEMODEL
_PLANSTEP.fields_by_name['update_feature_set'].message_type = _PLANSTEP_UPDATEFEATURESET
_PLANSTEP.fields_by_name['update_model'].message_type = _PLANSTEP_UPDATEMODEL
_PLANSTEP.fields_by_name['profile_feature_set'].message_type = _PLANSTEP_PROFILEFEATURESET
_PLANSTEP.fields_by_name['profile_model'].message_type = _PLANSTEP_PROFILEMODEL
_PLANSTEP.fields_by_name['train'].message_type = _PLANSTEP_TRAIN
_PLANSTEP.fields_by_name['promote'].message_type = _PLANSTEP_PROMOTE
_PLANSTEP.fields_by_name['predict'].message_type = _PLANSTEP_PREDICT
_PLANSTEP.fields_by_name['delete_feature_set'].message_type = _PLANSTEP_DELETEFEATURESET
_PLANSTEP.fields_by_name['delete_model'].message_type = _PLANSTEP_DELETEMODEL
_PLANSTEP_STATE.containing_type = _PLANSTEP
_PLANSTEP_OPERATION.containing_type = _PLANSTEP
_PLANSTEP_TYPE.containing_type = _PLANSTEP
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['create_feature_set'])
_PLANSTEP.fields_by_name['create_feature_set'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['create_model'])
_PLANSTEP.fields_by_name['create_model'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['update_feature_set'])
_PLANSTEP.fields_by_name['update_feature_set'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['update_model'])
_PLANSTEP.fields_by_name['update_model'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['profile_feature_set'])
_PLANSTEP.fields_by_name['profile_feature_set'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['profile_model'])
_PLANSTEP.fields_by_name['profile_model'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['train'])
_PLANSTEP.fields_by_name['train'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['promote'])
_PLANSTEP.fields_by_name['promote'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['predict'])
_PLANSTEP.fields_by_name['predict'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['delete_feature_set'])
_PLANSTEP.fields_by_name['delete_feature_set'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_PLANSTEP.oneofs_by_name['metadata'].fields.append(
  _PLANSTEP.fields_by_name['delete_model'])
_PLANSTEP.fields_by_name['delete_model'].containing_oneof = _PLANSTEP.oneofs_by_name['metadata']
_BATCHPREDICTION_LOGFILEPATHSENTRY.containing_type = _BATCHPREDICTION
_BATCHPREDICTION.fields_by_name['state'].enum_type = _BATCHPREDICTION_STATE
_BATCHPREDICTION.fields_by_name['source_type'].enum_type = _BATCHPREDICTIONSTORETYPE
_BATCHPREDICTION.fields_by_name['dest_type'].enum_type = _BATCHPREDICTIONSTORETYPE
_BATCHPREDICTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTION.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTION.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTION.fields_by_name['log_file_paths'].message_type = _BATCHPREDICTION_LOGFILEPATHSENTRY
_BATCHPREDICTION_STATE.containing_type = _BATCHPREDICTION
_DATASETSTATS.fields_by_name['dataset_stats'].message_type = _DATASETSTATSENTRY
_DATASETSTATS.fields_by_name['profile_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETSTATSENTRY.fields_by_name['correlation_matrix'].message_type = _CORRELATIONMATRIX
_DATASETSTATSENTRY.fields_by_name['string_stats'].message_type = _STRINGSTATS
_DATASETSTATSENTRY.fields_by_name['numeric_stats'].message_type = _NUMERICSTATS
_DATASETSTATSENTRY.fields_by_name['timestamp_stats'].message_type = _TIMESTAMPSTATS
_DATASETSTATSENTRY.fields_by_name['categorical_stats'].message_type = _CATEGORICALSTATS
_DATASETSTATSENTRY.fields_by_name['dataset_ranges'].message_type = _DATASETRANGE
_DATASETRANGE.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETRANGE.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CORRELATIONMATRIX_ROW.containing_type = _CORRELATIONMATRIX
_CORRELATIONMATRIX.fields_by_name['rows'].message_type = _CORRELATIONMATRIX_ROW
_HISTOGRAM_BUCKET.containing_type = _HISTOGRAM
_HISTOGRAM.fields_by_name['buckets'].message_type = _HISTOGRAM_BUCKET
_TIMESTAMPHISTOGRAM_BUCKET.fields_by_name['low_value'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPHISTOGRAM_BUCKET.fields_by_name['high_value'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPHISTOGRAM_BUCKET.containing_type = _TIMESTAMPHISTOGRAM
_TIMESTAMPHISTOGRAM.fields_by_name['buckets'].message_type = _TIMESTAMPHISTOGRAM_BUCKET
_CATEGORICALDISTRIBUTION_CATEGORY.containing_type = _CATEGORICALDISTRIBUTION
_CATEGORICALDISTRIBUTION.fields_by_name['categories'].message_type = _CATEGORICALDISTRIBUTION_CATEGORY
_CATEGORICALSTATS.fields_by_name['categorical_distribution'].message_type = _CATEGORICALDISTRIBUTION
_TIMESTAMPSTATS.fields_by_name['min'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPSTATS.fields_by_name['max'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPSTATS.fields_by_name['timestamp_histogram'].message_type = _TIMESTAMPHISTOGRAM
_NUMERICSTATS.fields_by_name['num_values_histogram'].message_type = _HISTOGRAM
_BILLINGACCOUNT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BILLINGACCOUNT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SUBSCRIPTION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SUBSCRIPTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTUSAGERECORD_TYPECOUNTENTRY.containing_type = _PROJECTUSAGERECORD
_PROJECTUSAGERECORD.fields_by_name['typeCount'].message_type = _PROJECTUSAGERECORD_TYPECOUNTENTRY
_CREDITUSAGERECORD.fields_by_name['period'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CREDITUSAGERECORD.fields_by_name['project_usages'].message_type = _PROJECTUSAGERECORD
_INVOICELINEITEM.fields_by_name['period_start'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICELINEITEM.fields_by_name['period_end'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['date'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['period_start'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['period_end'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['line_items'].message_type = _INVOICELINEITEM
_PAYMENTMETHOD_BILLINGDETAILS.containing_type = _PAYMENTMETHOD
_PAYMENTMETHOD_CARD.containing_type = _PAYMENTMETHOD
_PAYMENTMETHOD.fields_by_name['billing_details'].message_type = _PAYMENTMETHOD_BILLINGDETAILS
_PAYMENTMETHOD.fields_by_name['card'].message_type = _PAYMENTMETHOD_CARD
_DATACHECKS_DATAVALIDATIONRESULT.containing_type = _DATACHECKS
_DATACHECKS.fields_by_name['results'].message_type = _DATACHECKS_DATAVALIDATIONRESULT
DESCRIPTOR.message_types_by_name['IngestionErrorDetail'] = _INGESTIONERRORDETAIL
DESCRIPTOR.message_types_by_name['Ingestion'] = _INGESTION
DESCRIPTOR.message_types_by_name['User'] = _USER
DESCRIPTOR.message_types_by_name['Session'] = _SESSION
DESCRIPTOR.message_types_by_name['Role'] = _ROLE
DESCRIPTOR.message_types_by_name['AccessPolicy'] = _ACCESSPOLICY
DESCRIPTOR.message_types_by_name['Invite'] = _INVITE
DESCRIPTOR.message_types_by_name['Organization'] = _ORGANIZATION
DESCRIPTOR.message_types_by_name['Onboarding'] = _ONBOARDING
DESCRIPTOR.message_types_by_name['Project'] = _PROJECT
DESCRIPTOR.message_types_by_name['ProjectSummary'] = _PROJECTSUMMARY
DESCRIPTOR.message_types_by_name['DataStore'] = _DATASTORE
DESCRIPTOR.message_types_by_name['HealthCounts'] = _HEALTHCOUNTS
DESCRIPTOR.message_types_by_name['Environment'] = _ENVIRONMENT
DESCRIPTOR.message_types_by_name['FeatureSetSchema'] = _FEATURESETSCHEMA
DESCRIPTOR.message_types_by_name['ModelSchema'] = _MODELSCHEMA
DESCRIPTOR.message_types_by_name['ColumnConfig'] = _COLUMNCONFIG
DESCRIPTOR.message_types_by_name['TrainConfig'] = _TRAINCONFIG
DESCRIPTOR.message_types_by_name['PredictConfig'] = _PREDICTCONFIG
DESCRIPTOR.message_types_by_name['PromoteConfig'] = _PROMOTECONFIG
DESCRIPTOR.message_types_by_name['ProfileConfig'] = _PROFILECONFIG
DESCRIPTOR.message_types_by_name['SourceConfig'] = _SOURCECONFIG
DESCRIPTOR.message_types_by_name['FeatureSet'] = _FEATURESET
DESCRIPTOR.message_types_by_name['ColumnStats'] = _COLUMNSTATS
DESCRIPTOR.message_types_by_name['ColumnSetStats'] = _COLUMNSETSTATS
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['ServingMetric'] = _SERVINGMETRIC
DESCRIPTOR.message_types_by_name['Model'] = _MODEL
DESCRIPTOR.message_types_by_name['Promotion'] = _PROMOTION
DESCRIPTOR.message_types_by_name['ModelVersion'] = _MODELVERSION
DESCRIPTOR.message_types_by_name['ConfusionMatrix'] = _CONFUSIONMATRIX
DESCRIPTOR.message_types_by_name['BooleanEvaluationMetrics'] = _BOOLEANEVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['TextEvaluationMetrics'] = _TEXTEVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['CategoricalEvaluationMetrics'] = _CATEGORICALEVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['NumberEvaluationMetrics'] = _NUMBEREVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['EvaluationMetrics'] = _EVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['Metrics'] = _METRICS
DESCRIPTOR.message_types_by_name['Experiment'] = _EXPERIMENT
DESCRIPTOR.message_types_by_name['TrainingMetrics'] = _TRAININGMETRICS
DESCRIPTOR.message_types_by_name['Metric'] = _METRIC
DESCRIPTOR.message_types_by_name['BatchPredictionMetric'] = _BATCHPREDICTIONMETRIC
DESCRIPTOR.message_types_by_name['TextSequence'] = _TEXTSEQUENCE
DESCRIPTOR.message_types_by_name['NumberSequence'] = _NUMBERSEQUENCE
DESCRIPTOR.message_types_by_name['SignatureFeature'] = _SIGNATUREFEATURE
DESCRIPTOR.message_types_by_name['QueryFeature'] = _QUERYFEATURE
DESCRIPTOR.message_types_by_name['Query'] = _QUERY
DESCRIPTOR.message_types_by_name['ChangeSummary'] = _CHANGESUMMARY
DESCRIPTOR.message_types_by_name['Change'] = _CHANGE
DESCRIPTOR.message_types_by_name['PlanStep'] = _PLANSTEP
DESCRIPTOR.message_types_by_name['BatchPrediction'] = _BATCHPREDICTION
DESCRIPTOR.message_types_by_name['SignedURL'] = _SIGNEDURL
DESCRIPTOR.message_types_by_name['Entity'] = _ENTITY
DESCRIPTOR.message_types_by_name['DatasetStats'] = _DATASETSTATS
DESCRIPTOR.message_types_by_name['DatasetStatsEntry'] = _DATASETSTATSENTRY
DESCRIPTOR.message_types_by_name['DatasetRange'] = _DATASETRANGE
DESCRIPTOR.message_types_by_name['CorrelationMatrix'] = _CORRELATIONMATRIX
DESCRIPTOR.message_types_by_name['Histogram'] = _HISTOGRAM
DESCRIPTOR.message_types_by_name['TimestampHistogram'] = _TIMESTAMPHISTOGRAM
DESCRIPTOR.message_types_by_name['CategoricalDistribution'] = _CATEGORICALDISTRIBUTION
DESCRIPTOR.message_types_by_name['StringStats'] = _STRINGSTATS
DESCRIPTOR.message_types_by_name['CategoricalStats'] = _CATEGORICALSTATS
DESCRIPTOR.message_types_by_name['TimestampStats'] = _TIMESTAMPSTATS
DESCRIPTOR.message_types_by_name['NumericStats'] = _NUMERICSTATS
DESCRIPTOR.message_types_by_name['BillingAccount'] = _BILLINGACCOUNT
DESCRIPTOR.message_types_by_name['Subscription'] = _SUBSCRIPTION
DESCRIPTOR.message_types_by_name['ProjectUsageRecord'] = _PROJECTUSAGERECORD
DESCRIPTOR.message_types_by_name['CreditUsageRecord'] = _CREDITUSAGERECORD
DESCRIPTOR.message_types_by_name['InvoiceLineItem'] = _INVOICELINEITEM
DESCRIPTOR.message_types_by_name['Invoice'] = _INVOICE
DESCRIPTOR.message_types_by_name['Price'] = _PRICE
DESCRIPTOR.message_types_by_name['PaymentMethod'] = _PAYMENTMETHOD
DESCRIPTOR.message_types_by_name['DataChecks'] = _DATACHECKS
DESCRIPTOR.enum_types_by_name['FieldType'] = _FIELDTYPE
DESCRIPTOR.enum_types_by_name['BatchPredictionStoreType'] = _BATCHPREDICTIONSTORETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

IngestionErrorDetail = _reflection.GeneratedProtocolMessageType('IngestionErrorDetail', (_message.Message,), {
  'DESCRIPTOR' : _INGESTIONERRORDETAIL,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.IngestionErrorDetail)
  })
_sym_db.RegisterMessage(IngestionErrorDetail)

Ingestion = _reflection.GeneratedProtocolMessageType('Ingestion', (_message.Message,), {
  'DESCRIPTOR' : _INGESTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Ingestion)
  })
_sym_db.RegisterMessage(Ingestion)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.User)
  })
_sym_db.RegisterMessage(User)

Session = _reflection.GeneratedProtocolMessageType('Session', (_message.Message,), {
  'DESCRIPTOR' : _SESSION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Session)
  })
_sym_db.RegisterMessage(Session)

Role = _reflection.GeneratedProtocolMessageType('Role', (_message.Message,), {
  'DESCRIPTOR' : _ROLE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Role)
  })
_sym_db.RegisterMessage(Role)

AccessPolicy = _reflection.GeneratedProtocolMessageType('AccessPolicy', (_message.Message,), {
  'DESCRIPTOR' : _ACCESSPOLICY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.AccessPolicy)
  })
_sym_db.RegisterMessage(AccessPolicy)

Invite = _reflection.GeneratedProtocolMessageType('Invite', (_message.Message,), {
  'DESCRIPTOR' : _INVITE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Invite)
  })
_sym_db.RegisterMessage(Invite)

Organization = _reflection.GeneratedProtocolMessageType('Organization', (_message.Message,), {
  'DESCRIPTOR' : _ORGANIZATION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Organization)
  })
_sym_db.RegisterMessage(Organization)

Onboarding = _reflection.GeneratedProtocolMessageType('Onboarding', (_message.Message,), {
  'DESCRIPTOR' : _ONBOARDING,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Onboarding)
  })
_sym_db.RegisterMessage(Onboarding)

Project = _reflection.GeneratedProtocolMessageType('Project', (_message.Message,), {
  'DESCRIPTOR' : _PROJECT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Project)
  })
_sym_db.RegisterMessage(Project)

ProjectSummary = _reflection.GeneratedProtocolMessageType('ProjectSummary', (_message.Message,), {
  'DESCRIPTOR' : _PROJECTSUMMARY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectSummary)
  })
_sym_db.RegisterMessage(ProjectSummary)

DataStore = _reflection.GeneratedProtocolMessageType('DataStore', (_message.Message,), {

  'BigQuery' : _reflection.GeneratedProtocolMessageType('BigQuery', (_message.Message,), {
    'DESCRIPTOR' : _DATASTORE_BIGQUERY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.BigQuery)
    })
  ,

  'Snowflake' : _reflection.GeneratedProtocolMessageType('Snowflake', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_SNOWFLAKE_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_SNOWFLAKE,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Snowflake)
    })
  ,

  'Postgres' : _reflection.GeneratedProtocolMessageType('Postgres', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_POSTGRES_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Postgres.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_POSTGRES,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Postgres)
    })
  ,

  'Redshift' : _reflection.GeneratedProtocolMessageType('Redshift', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_REDSHIFT_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Redshift.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_REDSHIFT,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Redshift)
    })
  ,

  'Databricks' : _reflection.GeneratedProtocolMessageType('Databricks', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_DATABRICKS_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Databricks.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_DATABRICKS,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Databricks)
    })
  ,
  'DESCRIPTOR' : _DATASTORE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore)
  })
_sym_db.RegisterMessage(DataStore)
_sym_db.RegisterMessage(DataStore.BigQuery)
_sym_db.RegisterMessage(DataStore.Snowflake)
_sym_db.RegisterMessage(DataStore.Snowflake.ParamsEntry)
_sym_db.RegisterMessage(DataStore.Postgres)
_sym_db.RegisterMessage(DataStore.Postgres.ParamsEntry)
_sym_db.RegisterMessage(DataStore.Redshift)
_sym_db.RegisterMessage(DataStore.Redshift.ParamsEntry)
_sym_db.RegisterMessage(DataStore.Databricks)
_sym_db.RegisterMessage(DataStore.Databricks.ParamsEntry)

HealthCounts = _reflection.GeneratedProtocolMessageType('HealthCounts', (_message.Message,), {
  'DESCRIPTOR' : _HEALTHCOUNTS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.HealthCounts)
  })
_sym_db.RegisterMessage(HealthCounts)

Environment = _reflection.GeneratedProtocolMessageType('Environment', (_message.Message,), {
  'DESCRIPTOR' : _ENVIRONMENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Environment)
  })
_sym_db.RegisterMessage(Environment)

FeatureSetSchema = _reflection.GeneratedProtocolMessageType('FeatureSetSchema', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESETSCHEMA,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.FeatureSetSchema)
  })
_sym_db.RegisterMessage(FeatureSetSchema)

ModelSchema = _reflection.GeneratedProtocolMessageType('ModelSchema', (_message.Message,), {
  'DESCRIPTOR' : _MODELSCHEMA,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ModelSchema)
  })
_sym_db.RegisterMessage(ModelSchema)

ColumnConfig = _reflection.GeneratedProtocolMessageType('ColumnConfig', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNCONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ColumnConfig)
  })
_sym_db.RegisterMessage(ColumnConfig)

TrainConfig = _reflection.GeneratedProtocolMessageType('TrainConfig', (_message.Message,), {
  'DESCRIPTOR' : _TRAINCONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TrainConfig)
  })
_sym_db.RegisterMessage(TrainConfig)

PredictConfig = _reflection.GeneratedProtocolMessageType('PredictConfig', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTCONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PredictConfig)
  })
_sym_db.RegisterMessage(PredictConfig)

PromoteConfig = _reflection.GeneratedProtocolMessageType('PromoteConfig', (_message.Message,), {
  'DESCRIPTOR' : _PROMOTECONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PromoteConfig)
  })
_sym_db.RegisterMessage(PromoteConfig)

ProfileConfig = _reflection.GeneratedProtocolMessageType('ProfileConfig', (_message.Message,), {
  'DESCRIPTOR' : _PROFILECONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProfileConfig)
  })
_sym_db.RegisterMessage(ProfileConfig)

SourceConfig = _reflection.GeneratedProtocolMessageType('SourceConfig', (_message.Message,), {
  'DESCRIPTOR' : _SOURCECONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SourceConfig)
  })
_sym_db.RegisterMessage(SourceConfig)

FeatureSet = _reflection.GeneratedProtocolMessageType('FeatureSet', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESET,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.FeatureSet)
  })
_sym_db.RegisterMessage(FeatureSet)

ColumnStats = _reflection.GeneratedProtocolMessageType('ColumnStats', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ColumnStats)
  })
_sym_db.RegisterMessage(ColumnStats)

ColumnSetStats = _reflection.GeneratedProtocolMessageType('ColumnSetStats', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNSETSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ColumnSetStats)
  })
_sym_db.RegisterMessage(ColumnSetStats)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {

  'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
    'DESCRIPTOR' : _EVENT_DATAENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Event.DataEntry)
    })
  ,
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Event)
  })
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.DataEntry)

ServingMetric = _reflection.GeneratedProtocolMessageType('ServingMetric', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGMETRIC,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ServingMetric)
  })
_sym_db.RegisterMessage(ServingMetric)

Model = _reflection.GeneratedProtocolMessageType('Model', (_message.Message,), {
  'DESCRIPTOR' : _MODEL,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Model)
  })
_sym_db.RegisterMessage(Model)

Promotion = _reflection.GeneratedProtocolMessageType('Promotion', (_message.Message,), {
  'DESCRIPTOR' : _PROMOTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Promotion)
  })
_sym_db.RegisterMessage(Promotion)

ModelVersion = _reflection.GeneratedProtocolMessageType('ModelVersion', (_message.Message,), {

  'LogFilePathsEntry' : _reflection.GeneratedProtocolMessageType('LogFilePathsEntry', (_message.Message,), {
    'DESCRIPTOR' : _MODELVERSION_LOGFILEPATHSENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ModelVersion.LogFilePathsEntry)
    })
  ,

  'PlotFilesEntry' : _reflection.GeneratedProtocolMessageType('PlotFilesEntry', (_message.Message,), {
    'DESCRIPTOR' : _MODELVERSION_PLOTFILESENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ModelVersion.PlotFilesEntry)
    })
  ,
  'DESCRIPTOR' : _MODELVERSION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ModelVersion)
  })
_sym_db.RegisterMessage(ModelVersion)
_sym_db.RegisterMessage(ModelVersion.LogFilePathsEntry)
_sym_db.RegisterMessage(ModelVersion.PlotFilesEntry)

ConfusionMatrix = _reflection.GeneratedProtocolMessageType('ConfusionMatrix', (_message.Message,), {

  'Row' : _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {
    'DESCRIPTOR' : _CONFUSIONMATRIX_ROW,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ConfusionMatrix.Row)
    })
  ,
  'DESCRIPTOR' : _CONFUSIONMATRIX,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ConfusionMatrix)
  })
_sym_db.RegisterMessage(ConfusionMatrix)
_sym_db.RegisterMessage(ConfusionMatrix.Row)

BooleanEvaluationMetrics = _reflection.GeneratedProtocolMessageType('BooleanEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _BOOLEANEVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BooleanEvaluationMetrics)
  })
_sym_db.RegisterMessage(BooleanEvaluationMetrics)

TextEvaluationMetrics = _reflection.GeneratedProtocolMessageType('TextEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _TEXTEVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TextEvaluationMetrics)
  })
_sym_db.RegisterMessage(TextEvaluationMetrics)

CategoricalEvaluationMetrics = _reflection.GeneratedProtocolMessageType('CategoricalEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _CATEGORICALEVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalEvaluationMetrics)
  })
_sym_db.RegisterMessage(CategoricalEvaluationMetrics)

NumberEvaluationMetrics = _reflection.GeneratedProtocolMessageType('NumberEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _NUMBEREVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NumberEvaluationMetrics)
  })
_sym_db.RegisterMessage(NumberEvaluationMetrics)

EvaluationMetrics = _reflection.GeneratedProtocolMessageType('EvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _EVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.EvaluationMetrics)
  })
_sym_db.RegisterMessage(EvaluationMetrics)

Metrics = _reflection.GeneratedProtocolMessageType('Metrics', (_message.Message,), {
  'DESCRIPTOR' : _METRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Metrics)
  })
_sym_db.RegisterMessage(Metrics)

Experiment = _reflection.GeneratedProtocolMessageType('Experiment', (_message.Message,), {

  'FeatureImportanceEntry' : _reflection.GeneratedProtocolMessageType('FeatureImportanceEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPERIMENT_FEATUREIMPORTANCEENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Experiment.FeatureImportanceEntry)
    })
  ,

  'FeatureImportanceRelativeEntry' : _reflection.GeneratedProtocolMessageType('FeatureImportanceRelativeEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXPERIMENT_FEATUREIMPORTANCERELATIVEENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Experiment.FeatureImportanceRelativeEntry)
    })
  ,
  'DESCRIPTOR' : _EXPERIMENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Experiment)
  })
_sym_db.RegisterMessage(Experiment)
_sym_db.RegisterMessage(Experiment.FeatureImportanceEntry)
_sym_db.RegisterMessage(Experiment.FeatureImportanceRelativeEntry)

TrainingMetrics = _reflection.GeneratedProtocolMessageType('TrainingMetrics', (_message.Message,), {
  'DESCRIPTOR' : _TRAININGMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TrainingMetrics)
  })
_sym_db.RegisterMessage(TrainingMetrics)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Metric)
  })
_sym_db.RegisterMessage(Metric)

BatchPredictionMetric = _reflection.GeneratedProtocolMessageType('BatchPredictionMetric', (_message.Message,), {
  'DESCRIPTOR' : _BATCHPREDICTIONMETRIC,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchPredictionMetric)
  })
_sym_db.RegisterMessage(BatchPredictionMetric)

TextSequence = _reflection.GeneratedProtocolMessageType('TextSequence', (_message.Message,), {
  'DESCRIPTOR' : _TEXTSEQUENCE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TextSequence)
  })
_sym_db.RegisterMessage(TextSequence)

NumberSequence = _reflection.GeneratedProtocolMessageType('NumberSequence', (_message.Message,), {
  'DESCRIPTOR' : _NUMBERSEQUENCE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NumberSequence)
  })
_sym_db.RegisterMessage(NumberSequence)

SignatureFeature = _reflection.GeneratedProtocolMessageType('SignatureFeature', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATUREFEATURE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SignatureFeature)
  })
_sym_db.RegisterMessage(SignatureFeature)

QueryFeature = _reflection.GeneratedProtocolMessageType('QueryFeature', (_message.Message,), {
  'DESCRIPTOR' : _QUERYFEATURE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.QueryFeature)
  })
_sym_db.RegisterMessage(QueryFeature)

Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), {
  'DESCRIPTOR' : _QUERY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Query)
  })
_sym_db.RegisterMessage(Query)

ChangeSummary = _reflection.GeneratedProtocolMessageType('ChangeSummary', (_message.Message,), {
  'DESCRIPTOR' : _CHANGESUMMARY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ChangeSummary)
  })
_sym_db.RegisterMessage(ChangeSummary)

Change = _reflection.GeneratedProtocolMessageType('Change', (_message.Message,), {
  'DESCRIPTOR' : _CHANGE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Change)
  })
_sym_db.RegisterMessage(Change)

PlanStep = _reflection.GeneratedProtocolMessageType('PlanStep', (_message.Message,), {

  'CreateModel' : _reflection.GeneratedProtocolMessageType('CreateModel', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_CREATEMODEL,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.CreateModel)
    })
  ,

  'CreateFeatureSet' : _reflection.GeneratedProtocolMessageType('CreateFeatureSet', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_CREATEFEATURESET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.CreateFeatureSet)
    })
  ,

  'UpdateModel' : _reflection.GeneratedProtocolMessageType('UpdateModel', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_UPDATEMODEL,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.UpdateModel)
    })
  ,

  'UpdateFeatureSet' : _reflection.GeneratedProtocolMessageType('UpdateFeatureSet', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_UPDATEFEATURESET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.UpdateFeatureSet)
    })
  ,

  'ProfileFeatureSet' : _reflection.GeneratedProtocolMessageType('ProfileFeatureSet', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_PROFILEFEATURESET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.ProfileFeatureSet)
    })
  ,

  'ProfileModel' : _reflection.GeneratedProtocolMessageType('ProfileModel', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_PROFILEMODEL,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.ProfileModel)
    })
  ,

  'Predict' : _reflection.GeneratedProtocolMessageType('Predict', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_PREDICT,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.Predict)
    })
  ,

  'Train' : _reflection.GeneratedProtocolMessageType('Train', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_TRAIN,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.Train)
    })
  ,

  'Promote' : _reflection.GeneratedProtocolMessageType('Promote', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_PROMOTE,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.Promote)
    })
  ,

  'DeleteModel' : _reflection.GeneratedProtocolMessageType('DeleteModel', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_DELETEMODEL,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.DeleteModel)
    })
  ,

  'DeleteFeatureSet' : _reflection.GeneratedProtocolMessageType('DeleteFeatureSet', (_message.Message,), {
    'DESCRIPTOR' : _PLANSTEP_DELETEFEATURESET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep.DeleteFeatureSet)
    })
  ,
  'DESCRIPTOR' : _PLANSTEP,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PlanStep)
  })
_sym_db.RegisterMessage(PlanStep)
_sym_db.RegisterMessage(PlanStep.CreateModel)
_sym_db.RegisterMessage(PlanStep.CreateFeatureSet)
_sym_db.RegisterMessage(PlanStep.UpdateModel)
_sym_db.RegisterMessage(PlanStep.UpdateFeatureSet)
_sym_db.RegisterMessage(PlanStep.ProfileFeatureSet)
_sym_db.RegisterMessage(PlanStep.ProfileModel)
_sym_db.RegisterMessage(PlanStep.Predict)
_sym_db.RegisterMessage(PlanStep.Train)
_sym_db.RegisterMessage(PlanStep.Promote)
_sym_db.RegisterMessage(PlanStep.DeleteModel)
_sym_db.RegisterMessage(PlanStep.DeleteFeatureSet)

BatchPrediction = _reflection.GeneratedProtocolMessageType('BatchPrediction', (_message.Message,), {

  'LogFilePathsEntry' : _reflection.GeneratedProtocolMessageType('LogFilePathsEntry', (_message.Message,), {
    'DESCRIPTOR' : _BATCHPREDICTION_LOGFILEPATHSENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchPrediction.LogFilePathsEntry)
    })
  ,
  'DESCRIPTOR' : _BATCHPREDICTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchPrediction)
  })
_sym_db.RegisterMessage(BatchPrediction)
_sym_db.RegisterMessage(BatchPrediction.LogFilePathsEntry)

SignedURL = _reflection.GeneratedProtocolMessageType('SignedURL', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDURL,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SignedURL)
  })
_sym_db.RegisterMessage(SignedURL)

Entity = _reflection.GeneratedProtocolMessageType('Entity', (_message.Message,), {
  'DESCRIPTOR' : _ENTITY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Entity)
  })
_sym_db.RegisterMessage(Entity)

DatasetStats = _reflection.GeneratedProtocolMessageType('DatasetStats', (_message.Message,), {
  'DESCRIPTOR' : _DATASETSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetStats)
  })
_sym_db.RegisterMessage(DatasetStats)

DatasetStatsEntry = _reflection.GeneratedProtocolMessageType('DatasetStatsEntry', (_message.Message,), {
  'DESCRIPTOR' : _DATASETSTATSENTRY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetStatsEntry)
  })
_sym_db.RegisterMessage(DatasetStatsEntry)

DatasetRange = _reflection.GeneratedProtocolMessageType('DatasetRange', (_message.Message,), {
  'DESCRIPTOR' : _DATASETRANGE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetRange)
  })
_sym_db.RegisterMessage(DatasetRange)

CorrelationMatrix = _reflection.GeneratedProtocolMessageType('CorrelationMatrix', (_message.Message,), {

  'Row' : _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {
    'DESCRIPTOR' : _CORRELATIONMATRIX_ROW,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CorrelationMatrix.Row)
    })
  ,
  'DESCRIPTOR' : _CORRELATIONMATRIX,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CorrelationMatrix)
  })
_sym_db.RegisterMessage(CorrelationMatrix)
_sym_db.RegisterMessage(CorrelationMatrix.Row)

Histogram = _reflection.GeneratedProtocolMessageType('Histogram', (_message.Message,), {

  'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
    'DESCRIPTOR' : _HISTOGRAM_BUCKET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Histogram.Bucket)
    })
  ,
  'DESCRIPTOR' : _HISTOGRAM,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Histogram)
  })
_sym_db.RegisterMessage(Histogram)
_sym_db.RegisterMessage(Histogram.Bucket)

TimestampHistogram = _reflection.GeneratedProtocolMessageType('TimestampHistogram', (_message.Message,), {

  'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
    'DESCRIPTOR' : _TIMESTAMPHISTOGRAM_BUCKET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TimestampHistogram.Bucket)
    })
  ,
  'DESCRIPTOR' : _TIMESTAMPHISTOGRAM,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TimestampHistogram)
  })
_sym_db.RegisterMessage(TimestampHistogram)
_sym_db.RegisterMessage(TimestampHistogram.Bucket)

CategoricalDistribution = _reflection.GeneratedProtocolMessageType('CategoricalDistribution', (_message.Message,), {

  'Category' : _reflection.GeneratedProtocolMessageType('Category', (_message.Message,), {
    'DESCRIPTOR' : _CATEGORICALDISTRIBUTION_CATEGORY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalDistribution.Category)
    })
  ,
  'DESCRIPTOR' : _CATEGORICALDISTRIBUTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalDistribution)
  })
_sym_db.RegisterMessage(CategoricalDistribution)
_sym_db.RegisterMessage(CategoricalDistribution.Category)

StringStats = _reflection.GeneratedProtocolMessageType('StringStats', (_message.Message,), {
  'DESCRIPTOR' : _STRINGSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.StringStats)
  })
_sym_db.RegisterMessage(StringStats)

CategoricalStats = _reflection.GeneratedProtocolMessageType('CategoricalStats', (_message.Message,), {
  'DESCRIPTOR' : _CATEGORICALSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalStats)
  })
_sym_db.RegisterMessage(CategoricalStats)

TimestampStats = _reflection.GeneratedProtocolMessageType('TimestampStats', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TimestampStats)
  })
_sym_db.RegisterMessage(TimestampStats)

NumericStats = _reflection.GeneratedProtocolMessageType('NumericStats', (_message.Message,), {
  'DESCRIPTOR' : _NUMERICSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NumericStats)
  })
_sym_db.RegisterMessage(NumericStats)

BillingAccount = _reflection.GeneratedProtocolMessageType('BillingAccount', (_message.Message,), {
  'DESCRIPTOR' : _BILLINGACCOUNT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BillingAccount)
  })
_sym_db.RegisterMessage(BillingAccount)

Subscription = _reflection.GeneratedProtocolMessageType('Subscription', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Subscription)
  })
_sym_db.RegisterMessage(Subscription)

ProjectUsageRecord = _reflection.GeneratedProtocolMessageType('ProjectUsageRecord', (_message.Message,), {

  'TypeCountEntry' : _reflection.GeneratedProtocolMessageType('TypeCountEntry', (_message.Message,), {
    'DESCRIPTOR' : _PROJECTUSAGERECORD_TYPECOUNTENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry)
    })
  ,
  'DESCRIPTOR' : _PROJECTUSAGERECORD,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectUsageRecord)
  })
_sym_db.RegisterMessage(ProjectUsageRecord)
_sym_db.RegisterMessage(ProjectUsageRecord.TypeCountEntry)

CreditUsageRecord = _reflection.GeneratedProtocolMessageType('CreditUsageRecord', (_message.Message,), {
  'DESCRIPTOR' : _CREDITUSAGERECORD,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreditUsageRecord)
  })
_sym_db.RegisterMessage(CreditUsageRecord)

InvoiceLineItem = _reflection.GeneratedProtocolMessageType('InvoiceLineItem', (_message.Message,), {
  'DESCRIPTOR' : _INVOICELINEITEM,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.InvoiceLineItem)
  })
_sym_db.RegisterMessage(InvoiceLineItem)

Invoice = _reflection.GeneratedProtocolMessageType('Invoice', (_message.Message,), {
  'DESCRIPTOR' : _INVOICE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Invoice)
  })
_sym_db.RegisterMessage(Invoice)

Price = _reflection.GeneratedProtocolMessageType('Price', (_message.Message,), {
  'DESCRIPTOR' : _PRICE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Price)
  })
_sym_db.RegisterMessage(Price)

PaymentMethod = _reflection.GeneratedProtocolMessageType('PaymentMethod', (_message.Message,), {

  'BillingDetails' : _reflection.GeneratedProtocolMessageType('BillingDetails', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_BILLINGDETAILS,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PaymentMethod.BillingDetails)
    })
  ,

  'Card' : _reflection.GeneratedProtocolMessageType('Card', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_CARD,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PaymentMethod.Card)
    })
  ,
  'DESCRIPTOR' : _PAYMENTMETHOD,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PaymentMethod)
  })
_sym_db.RegisterMessage(PaymentMethod)
_sym_db.RegisterMessage(PaymentMethod.BillingDetails)
_sym_db.RegisterMessage(PaymentMethod.Card)

DataChecks = _reflection.GeneratedProtocolMessageType('DataChecks', (_message.Message,), {

  'DataValidationResult' : _reflection.GeneratedProtocolMessageType('DataValidationResult', (_message.Message,), {
    'DESCRIPTOR' : _DATACHECKS_DATAVALIDATIONRESULT,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataChecks.DataValidationResult)
    })
  ,
  'DESCRIPTOR' : _DATACHECKS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataChecks)
  })
_sym_db.RegisterMessage(DataChecks)
_sym_db.RegisterMessage(DataChecks.DataValidationResult)


DESCRIPTOR._options = None
_USER.fields_by_name['password']._options = None
_USER.fields_by_name['service_account']._options = None
_USER.fields_by_name['update_time']._options = None
_USER.fields_by_name['create_time']._options = None
_SESSION.fields_by_name['name']._options = None
_SESSION.fields_by_name['subject']._options = None
_SESSION.fields_by_name['last_access_ip']._options = None
_SESSION.fields_by_name['last_access_time']._options = None
_SESSION.fields_by_name['login_ip']._options = None
_SESSION.fields_by_name['login_time']._options = None
_SESSION.fields_by_name['login_user_agent']._options = None
_ACCESSPOLICY.fields_by_name['subject']._options = None
_ACCESSPOLICY.fields_by_name['role']._options = None
_INVITE.fields_by_name['sender']._options = None
_INVITE.fields_by_name['role']._options = None
_INVITE.fields_by_name['create_time']._options = None
_INVITE.fields_by_name['expire_time']._options = None
_ORGANIZATION.fields_by_name['update_time']._options = None
_ORGANIZATION.fields_by_name['create_time']._options = None
_ORGANIZATION.fields_by_name['status']._options = None
_ORGANIZATION.fields_by_name['trial_end_date']._options = None
_ORGANIZATION.fields_by_name['trial_credits']._options = None
_ORGANIZATION.fields_by_name['trial_credits_used']._options = None
_ONBOARDING.fields_by_name['create_time']._options = None
_ONBOARDING.fields_by_name['update_time']._options = None
_PROJECT.fields_by_name['organization']._options = None
_PROJECT.fields_by_name['summary']._options = None
_PROJECT.fields_by_name['update_time']._options = None
_PROJECT.fields_by_name['create_time']._options = None
_DATASTORE_SNOWFLAKE_PARAMSENTRY._options = None
_DATASTORE_POSTGRES_PARAMSENTRY._options = None
_DATASTORE_REDSHIFT_PARAMSENTRY._options = None
_DATASTORE_DATABRICKS_PARAMSENTRY._options = None
_ENVIRONMENT.fields_by_name['summary']._options = None
_ENVIRONMENT.fields_by_name['update_time']._options = None
_ENVIRONMENT.fields_by_name['create_time']._options = None
_FEATURESET.fields_by_name['table']._options = None
_EVENT_DATAENTRY._options = None
_EVENT.fields_by_name['subject']._options = None
_MODEL.fields_by_name['latest_model_version']._options = None
_MODEL.fields_by_name['create_time']._options = None
_MODEL.fields_by_name['update_time']._options = None
_MODEL.fields_by_name['latest_batch_prediction']._options = None
_MODEL.fields_by_name['current_version']._options = None
_MODEL.fields_by_name['table']._options = None
_PROMOTION.fields_by_name['create_time']._options = None
_PROMOTION.fields_by_name['promoted_time']._options = None
_PROMOTION.fields_by_name['demoted_time']._options = None
_PROMOTION.fields_by_name['model_version']._options = None
_PROMOTION.fields_by_name['base_model_version']._options = None
_PROMOTION.fields_by_name['state']._options = None
_PROMOTION.fields_by_name['account']._options = None
_MODELVERSION_LOGFILEPATHSENTRY._options = None
_MODELVERSION_PLOTFILESENTRY._options = None
_MODELVERSION.fields_by_name['name']._options = None
_MODELVERSION.fields_by_name['model_version_id']._options = None
_MODELVERSION.fields_by_name['model_signature']._options = None
_MODELVERSION.fields_by_name['experiment']._options = None
_MODELVERSION.fields_by_name['storage_uri']._options = None
_MODELVERSION.fields_by_name['state']._options = None
_MODELVERSION.fields_by_name['error_message']._options = None
_MODELVERSION.fields_by_name['stack_trace']._options = None
_MODELVERSION.fields_by_name['create_time']._options = None
_MODELVERSION.fields_by_name['update_time']._options = None
_MODELVERSION.fields_by_name['platform_version']._options = None
_MODELVERSION.fields_by_name['training_row_count']._options = None
_MODELVERSION.fields_by_name['validation_row_count']._options = None
_MODELVERSION.fields_by_name['test_row_count']._options = None
_MODELVERSION.fields_by_name['performance_metric']._options = None
_MODELVERSION.fields_by_name['performance_metric_val']._options = None
_MODELVERSION.fields_by_name['prediction_csv_path']._options = None
_MODELVERSION.fields_by_name['promotion']._options = None
_MODELVERSION.fields_by_name['model_query']._options = None
_EXPERIMENT_FEATUREIMPORTANCEENTRY._options = None
_EXPERIMENT_FEATUREIMPORTANCERELATIVEENTRY._options = None
_PLANSTEP_CREATEMODEL.fields_by_name['model']._options = None
_PLANSTEP_CREATEFEATURESET.fields_by_name['feature_set']._options = None
_PLANSTEP_UPDATEMODEL.fields_by_name['model']._options = None
_PLANSTEP_UPDATEFEATURESET.fields_by_name['feature_set']._options = None
_PLANSTEP_PROFILEFEATURESET.fields_by_name['feature_set']._options = None
_PLANSTEP_PROFILEMODEL.fields_by_name['model']._options = None
_PLANSTEP_PREDICT.fields_by_name['batch_prediction']._options = None
_PLANSTEP_PREDICT.fields_by_name['previous_batch_prediction']._options = None
_PLANSTEP_TRAIN.fields_by_name['model_version']._options = None
_PLANSTEP_TRAIN.fields_by_name['base_model_version']._options = None
_PLANSTEP_TRAIN.fields_by_name['latest_model_version']._options = None
_PLANSTEP_TRAIN.fields_by_name['promoted_model_version']._options = None
_PLANSTEP_TRAIN.fields_by_name['production_model_version']._options = None
_PLANSTEP_PROMOTE.fields_by_name['promotion']._options = None
_BATCHPREDICTION_LOGFILEPATHSENTRY._options = None
_BATCHPREDICTION.fields_by_name['model_version']._options = None
_BATCHPREDICTION.fields_by_name['table']._options = None
_BILLINGACCOUNT.fields_by_name['update_time']._options = None
_BILLINGACCOUNT.fields_by_name['create_time']._options = None
_SUBSCRIPTION.fields_by_name['price']._options = None
_SUBSCRIPTION.fields_by_name['payment_method']._options = None
_SUBSCRIPTION.fields_by_name['update_time']._options = None
_SUBSCRIPTION.fields_by_name['create_time']._options = None
_PROJECTUSAGERECORD_TYPECOUNTENTRY._options = None
# @@protoc_insertion_point(module_scope)
