# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/management/v1/management.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2
from continual.rpc.management.v1 import management_types_pb2 as continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2
from continual.rpc.graphql import options_pb2 as continual_dot_rpc_dot_graphql_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/management/v1/management.proto',
  package='continual.rpc.management.v1',
  syntax='proto3',
  serialized_options=b'ZIgithub.com/continualiq/continual/continual/rpc/management/v1;managementv1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,continual/rpc/management/v1/management.proto\x12\x1b\x63ontinual.rpc.management.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/api/annotations.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a google/protobuf/field_mask.proto\x1a\x32\x63ontinual/rpc/management/v1/management_types.proto\x1a#continual/rpc/graphql/options.proto\"u\n\x11UpdateUserRequest\x12/\n\x04user\x18\x01 \x01(\x0b\x32!.continual.rpc.management.v1.User\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"\x1e\n\x0eGetUserRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"%\n\x14\x42\x61tchGetUsersRequest\x12\r\n\x05names\x18\x01 \x03(\t\"I\n\x15\x42\x61tchGetUsersResponse\x12\x30\n\x05users\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.User\"!\n\x11\x44\x65leteUserRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"9\n\x10ListUsersRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12\x12\n\npage_token\x18\x02 \x01(\t\"^\n\x11ListUsersResponse\x12\x30\n\x05users\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.User\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"l\n\x0fRegisterRequest\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x10\n\x08password\x18\x03 \x01(\t\x12\x12\n\nfirst_name\x18\x04 \x01(\t\x12\x11\n\tlast_name\x18\x05 \x01(\t\"W\n\x10RegisterResponse\x12/\n\x04user\x18\x01 \x01(\x0b\x32!.continual.rpc.management.v1.User\x12\x12\n\nauth_token\x18\x02 \x01(\t\"/\n\x0cLoginRequest\x12\r\n\x05\x65mail\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\"T\n\rLoginResponse\x12/\n\x04user\x18\x01 \x01(\x0b\x32!.continual.rpc.management.v1.User\x12\x12\n\nauth_token\x18\x02 \x01(\t\",\n\x1cSendEmailVerificationRequest\x12\x0c\n\x04user\x18\x01 \x01(\t\"0\n\x12VerifyEmailRequest\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\"&\n\x13VerifyEmailResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\")\n\x18SendPasswordResetRequest\x12\r\n\x05\x65mail\x18\x01 \x01(\t\"\'\n\x17GetPasswordResetRequest\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\"6\n\x14ResetPasswordRequest\x12\x10\n\x08password\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\"(\n\x15ResetPasswordResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"!\n\x11GetSessionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"L\n\x13ListSessionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"g\n\x14ListSessionsResponse\x12\x36\n\x08sessions\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Session\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"$\n\x14\x44\x65leteSessionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"*\n\x18\x44\x65leteAllSessionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\"9\n\x10ListRolesRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12\x12\n\npage_token\x18\x02 \x01(\t\"^\n\x11ListRolesResponse\x12\x30\n\x05roles\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.Role\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x1e\n\x0eGetRoleRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"5\n\x14\x42\x61tchGetRolesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"I\n\x15\x42\x61tchGetRolesResponse\x12\x30\n\x05roles\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.Role\"R\n\x19ListAccessPoliciesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"y\n\x1aListAccessPoliciesResponse\x12\x42\n\x0f\x61\x63\x63\x65ss_policies\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.AccessPolicy\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"m\n\x19\x43reateAccessPolicyRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12@\n\raccess_policy\x18\x02 \x01(\x0b\x32).continual.rpc.management.v1.AccessPolicy\"&\n\x16GetAccessPolicyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x8e\x01\n\x19UpdateAccessPolicyRequest\x12@\n\raccess_policy\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.AccessPolicy\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\")\n\x19\x44\x65leteAccessPolicyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"Z\n\x13\x43reateInviteRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x33\n\x06invite\x18\x02 \x01(\x0b\x32#.continual.rpc.management.v1.Invite\" \n\x10GetInviteRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"K\n\x12ListInvitesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\"d\n\x13ListInvitesResponse\x12\x34\n\x07invites\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.Invite\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"{\n\x13UpdateInviteRequest\x12\x33\n\x06invite\x18\x01 \x01(\x0b\x32#.continual.rpc.management.v1.Invite\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"#\n\x13\x44\x65leteInviteRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"#\n\x13\x41\x63\x63\x65ptInviteRequest\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\"#\n\x13ResendInviteRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"?\n\x16TestPermissionsRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x13\n\x0bpermissions\x18\x02 \x03(\t\".\n\x17TestPermissionsResponse\x12\x13\n\x0bpermissions\x18\x01 \x03(\t\"\\\n\x19\x43reateOrganizationRequest\x12?\n\x0corganization\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.Organization\"\x8d\x01\n\x19UpdateOrganizationRequest\x12?\n\x0corganization\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.Organization\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"&\n\x16GetOrganizationRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"-\n\x1c\x42\x61tchGetOrganizationsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"a\n\x1d\x42\x61tchGetOrganizationsResponse\x12@\n\rorganizations\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.Organization\")\n\x19\x44\x65leteOrganizationRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"A\n\x18ListOrganizationsRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12\x12\n\npage_token\x18\x02 \x01(\t\"v\n\x19ListOrganizationsResponse\x12@\n\rorganizations\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.Organization\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"$\n\x14GetOnboardingRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x87\x01\n\x17UpdateOnboardingRequest\x12;\n\nonboarding\x18\x01 \x01(\x0b\x32\'.continual.rpc.management.v1.Onboarding\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"a\n\x14\x43reateProjectRequest\x12\x35\n\x07project\x18\x01 \x01(\x0b\x32$.continual.rpc.management.v1.Project\x12\x12\n\nproject_id\x18\x02 \x01(\t\"~\n\x14UpdateProjectRequest\x12\x35\n\x07project\x18\x01 \x01(\x0b\x32$.continual.rpc.management.v1.Project\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"!\n\x11GetProjectRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"(\n\x17\x42\x61tchGetProjectsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"R\n\x18\x42\x61tchGetProjectsResponse\x12\x36\n\x08projects\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Project\";\n\x14\x44\x65leteProjectRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rdelete_schema\x18\x02 \x01(\x08\"\xd3\x01\n\x13ListProjectsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"g\n\x14ListProjectsResponse\x12\x36\n\x08projects\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Project\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"+\n\x1bTestProjectDataStoreRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"F\n\x1cTestProjectDataStoreResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"R\n\x14TestDataStoreRequest\x12:\n\ndata_store\x18\x01 \x01(\x0b\x32&.continual.rpc.management.v1.DataStore\"?\n\x15TestDataStoreResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"$\n\x14GetFeatureSetRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xdd\x01\n\x13\x43reateChangeRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0f\n\x07schemas\x18\x02 \x03(\t\x12\x11\n\tplan_only\x18\x03 \x01(\x08\x12\r\n\x05purge\x18\x04 \x01(\x08\x12\x12\n\nupdate_all\x18\x05 \x01(\x08\x12\x12\n\ntrain_none\x18\x06 \x01(\x08\x12\x11\n\ttrain_all\x18\x07 \x01(\x08\x12\x14\n\x0cprofile_none\x18\x08 \x01(\x08\x12\x0f\n\x07message\x18\t \x01(\t\x12\x0f\n\x07trigger\x18\n \x01(\t\x12\x10\n\x08resource\x18\x0b \x01(\t\">\n\x0cReRunRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x06 \x01(\t\x12\x0f\n\x07trigger\x18\x07 \x01(\t\" \n\x10GetChangeRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"V\n\x11GetChangeResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x33\n\x04plan\x18\x02 \x03(\x0b\x32%.continual.rpc.management.v1.PlanStep\"\xf8\x01\n\x12ListChangesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\"b\n\x13ListChangesResponse\x12\x32\n\x05plans\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.Change\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"#\n\x13\x43\x61ncelChangeRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xec\x01\n\x16ListFeatureSetsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\x08 \x01(\x08\"q\n\x17ListFeatureSetsResponse\x12=\n\x0c\x66\x65\x61ture_sets\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.FeatureSet\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"F\n\x1bGetEntityFeatureSetsRequest\x12\x14\n\x0cproject_name\x18\x01 \x01(\t\x12\x11\n\tentity_id\x18\x02 \x01(\t\"]\n\x1cGetEntityFeatureSetsResponse\x12=\n\x0c\x66\x65\x61ture_sets\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.FeatureSet\" \n\x10GetEntityRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"%\n\x13ListEntitiesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\"M\n\x14ListEntitiesResponse\x12\x35\n\x08\x65ntities\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.Entity\";\n\x1a\x42\x61tchGetFeatureSetsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"\\\n\x1b\x42\x61tchGetFeatureSetsResponse\x12=\n\x0c\x66\x65\x61ture_sets\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.FeatureSet\"\'\n\x17\x44\x65leteFeatureSetRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"@\n\x19ValidateFeatureSetRequest\x12\x13\n\x0b\x65nvironment\x18\x01 \x01(\t\x12\x0e\n\x06schema\x18\x02 \x01(\t\"0\n GetFeatureSetDatasetStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"+\n\x1bGetModelDatasetStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"2\n\"GetModelVersionDatasetStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"5\n%GetBatchPredictionDatasetStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\")\n\x19GetPredictionStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\")\n\x19GetFeatureSetStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"$\n\x14GetModelStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\")\n\x19\x43omputeColumnStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"E\n\x0eGetDataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"O\n\x0fGetDataResponse\x12\x14\n\x0crecord_count\x18\x03 \x01(\x05\x12\x17\n\x0fnext_page_token\x18\x04 \x01(\t\x12\r\n\x05value\x18\x05 \x03(\t\"\x7f\n\x0fGetUsageRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"[\n\x10GetUsageResponse\x12G\n\x0f\x63redits_per_day\x18\x01 \x03(\x0b\x32..continual.rpc.management.v1.CreditUsageRecord\"#\n\x11ListPricesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\"H\n\x12ListPricesResponse\x12\x32\n\x06prices\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Price\"L\n\x13ListInvoicesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"g\n\x14ListInvoicesResponse\x12\x36\n\x08invoices\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Invoice\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x1f\n\x0fGetPriceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\'\n\x17GetPaymentMethodRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\")\n\x19GetUpcomingInvoiceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"&\n\x16GetSubscriptionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"(\n\x18GetBillingAccountRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x94\x01\n\x1bUpdateBillingAccountRequest\x12\x44\n\x0f\x62illing_account\x18\x01 \x01(\x0b\x32+.continual.rpc.management.v1.BillingAccount\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"\x8d\x01\n\x19UpdateSubscriptionRequest\x12?\n\x0csubscription\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.Subscription\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"H\n#CreateFeatureSetCSVSignedURLRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tfile_name\x18\x02 \x01(\t\"O\n$CreateFeatureSetCSVSignedURLResponse\x12\x12\n\nsigned_url\x18\x01 \x01(\t\x12\x13\n\x0b\x62ucket_path\x18\x02 \x01(\t\"*\n\x1aPromoteModelVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xfb\x01\n\x15ListPromotionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\x12\x0e\n\x06\x61\x63tive\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\"m\n\x16ListPromotionsResponse\x12:\n\npromotions\x18\x01 \x03(\x0b\x32&.continual.rpc.management.v1.Promotion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"#\n\x13GetPromotionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\":\n\x19\x42\x61tchGetPromotionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"X\n\x1a\x42\x61tchGetPromotionsResponse\x12:\n\npromotions\x18\x01 \x03(\x0b\x32&.continual.rpc.management.v1.Promotion\"\xe7\x01\n\x11ListEventsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x02 \x03(\t\x12\x16\n\x0eignore_filters\x18\x03 \x03(\t\x12.\n\nstart_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\npage_token\x18\x06 \x01(\t\x12\x11\n\tpage_size\x18\x07 \x01(\x05\x12\x14\n\x0c\x61ll_projects\x18\x08 \x01(\x08\"a\n\x12ListEventsResponse\x12\x32\n\x06\x65vents\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Event\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x1f\n\x0fGetEventRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"6\n\x15\x42\x61tchGetEventsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"L\n\x16\x42\x61tchGetEventsResponse\x12\x32\n\x06\x65vents\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Event\"q\n\x18GetInferredSchemaRequest\x12\x0f\n\x07project\x18\x01 \x01(\t\x12\x0c\n\x04\x66ile\x18\x02 \x01(\t\x12\r\n\x05query\x18\x04 \x01(\t\x12\r\n\x05table\x18\x05 \x01(\t\x12\x18\n\x10sample_row_count\x18\x06 \x01(\x05\"}\n\x19GetInferredSchemaResponse\x12:\n\x07\x63olumns\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.ColumnConfig\x12\x14\n\x0crecord_count\x18\x02 \x01(\x05\x12\x0e\n\x06values\x18\x03 \x03(\t\"\x1f\n\x0fGetModelRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"6\n\x15\x42\x61tchGetModelsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"L\n\x16\x42\x61tchGetModelsResponse\x12\x32\n\x06models\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Model\"\x8d\x02\n\x11ListModelsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x14\n\x0cshow_deleted\x18\x04 \x01(\x08\x12\x0e\n\x06latest\x18\x05 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x06 \x03(\t\x12\x16\n\x0eignore_filters\x18\x07 \x03(\t\x12.\n\nstart_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\n \x01(\x08\"a\n\x12ListModelsResponse\x12\x32\n\x06models\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Model\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\"\n\x12\x44\x65leteModelRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\";\n\x14ValidateModelRequest\x12\x13\n\x0b\x65nvironment\x18\x01 \x01(\t\x12\x0e\n\x06schema\x18\x02 \x01(\t\"&\n\x16GetModelVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"+\n\x1bGetModelVersionQueryRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"Q\n\x1cGetModelVersionQueryResponse\x12\x31\n\x05query\x18\x01 \x01(\x0b\x32\".continual.rpc.management.v1.Query\"=\n\x1c\x42\x61tchGetModelVersionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"b\n\x1d\x42\x61tchGetModelVersionsResponse\x12\x41\n\x0emodel_versions\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.ModelVersion\"\xfe\x01\n\x18ListModelVersionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\"w\n\x19ListModelVersionsResponse\x12\x41\n\x0emodel_versions\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.ModelVersion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\xc1\x01\n\x11TrainModelRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\x0bsystem_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x36\n\x12visible_start_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x35\n\x11visible_stop_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"%\n\x15\x43\x61ncelTrainingRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"/\n\x19GetTrainingMetricsRequest\x12\x12\n\nexperiment\x18\x01 \x01(\t\"d\n\x1aGetTrainingMetricsResponse\x12\x46\n\x10training_metrics\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.TrainingMetrics\"R\n\x1aListTrainingMetricsRequest\x12\r\n\x05model\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\"~\n\x1bListTrainingMetricsResponse\x12\x46\n\x10training_metrics\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.TrainingMetrics\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"$\n\x14GetExperimentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\";\n\x1a\x42\x61tchGetExperimentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"[\n\x1b\x42\x61tchGetExperimentsResponse\x12<\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.Experiment\"\xec\x01\n\x16ListExperimentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\x08 \x01(\x08\"p\n\x17ListExperimentsResponse\x12<\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.Experiment\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x82\x02\n\x1c\x43reateBatchPredictionRequest\x12\r\n\x05model\x18\x01 \x01(\t\x12J\n\x0bsource_type\x18\x02 \x01(\x0e\x32\x35.continual.rpc.management.v1.BatchPredictionStoreType\x12\x0e\n\x06source\x18\x03 \x01(\t\x12H\n\tdest_type\x18\x04 \x01(\x0e\x32\x35.continual.rpc.management.v1.BatchPredictionStoreType\x12\x13\n\x0bincremental\x18\x05 \x01(\x08\x12\x18\n\x10prediction_model\x18\x06 \x01(\t\"\x81\x02\n\x1bListBatchPredictionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\"\x80\x01\n\x1cListBatchPredictionsResponse\x12G\n\x11\x62\x61tch_predictions\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.BatchPrediction\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"&\n\x16\x42\x61tchPredictionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"@\n\x1f\x42\x61tchGetBatchPredictionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"k\n BatchGetBatchPredictionsResponse\x12G\n\x11\x62\x61tch_predictions\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.BatchPrediction\"\x8d\x01\n\x15GetServingDataRequest\x12\x31\n\x05query\x18\x01 \x01(\x0b\x32\".continual.rpc.management.v1.Query\x12\x11\n\toverrides\x18\x02 \x01(\t\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\'\n\x16GetServingDataResponse\x12\r\n\x05value\x18\x01 \x03(\t\"K\n\x16GetTrainingDataRequest\x12\x31\n\x05query\x18\x01 \x01(\x0b\x32\".continual.rpc.management.v1.Query\"(\n\x17GetTrainingDataResponse\x12\r\n\x05value\x18\x01 \x03(\t\"\xf6\x01\n\x1a\x42rowseDataWarehouseRequest\x12W\n\x0bobject_type\x18\x01 \x01(\x0e\x32\x42.continual.rpc.management.v1.BrowseDataWarehouseRequest.ObjectType\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x03 \x01(\t\x12\x0e\n\x06schema\x18\x04 \x01(\t\x12\r\n\x05table\x18\x05 \x01(\t\"=\n\nObjectType\x12\x0c\n\x08\x44\x41TABASE\x10\x00\x12\n\n\x06SCHEMA\x10\x01\x12\t\n\x05TABLE\x10\x02\x12\n\n\x06\x43OLUMN\x10\x03\"\x87\x02\n\x1b\x42rowseDataWarehouseResponse\x12W\n\x0bobject_type\x18\x01 \x01(\x0e\x32\x42.continual.rpc.management.v1.BrowseDataWarehouseRequest.ObjectType\x12\r\n\x05names\x18\x02 \x03(\t\x12R\n\x05types\x18\x03 \x03(\x0b\x32\x43.continual.rpc.management.v1.BrowseDataWarehouseResponse.TypesEntry\x1a,\n\nTypesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"E\n\x0bSeedRequest\x12\x0f\n\x07project\x18\x01 \x01(\t\x12\x11\n\tfile_path\x18\x02 \x01(\t\x12\x12\n\ntable_name\x18\x03 \x01(\t\"7\n\x0cSeedResponse\x12\x13\n\x0bschema_name\x18\x01 \x01(\t\x12\x12\n\ntable_name\x18\x02 \x01(\t\"e\n\x0fYamlDiffRequest\x12\r\n\x05right\x18\x01 \x01(\t\x12\x12\n\nright_text\x18\x02 \x01(\t\x12\x0c\n\x04left\x18\x03 \x01(\t\x12\x11\n\tleft_text\x18\x04 \x01(\t\x12\x0e\n\x06parent\x18\x05 \x01(\t\" \n\x10YamlDiffResponse\x12\x0c\n\x04text\x18\x01 \x01(\t\"\xa1\x01\n\x14GetModelQueryRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x31\n\x05model\x18\x02 \x01(\x0b\x32\".continual.rpc.management.v1.Model\x12H\n\x17\x61\x64\x64itional_feature_sets\x18\x03 \x03(\x0b\x32\'.continual.rpc.management.v1.FeatureSet\"+\n\x15GetModelQueryResponse\x12\x12\n\nquery_json\x18\x01 \x01(\t\"\xae\x01\n\x15GetSchemaGraphRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12H\n\x17\x61\x64\x64itional_feature_sets\x18\x02 \x03(\x0b\x32\'.continual.rpc.management.v1.FeatureSet\x12=\n\x11\x61\x64\x64itional_models\x18\x03 \x03(\x0b\x32\".continual.rpc.management.v1.Model\",\n\x16GetSchemaGraphResponse\x12\x12\n\ngraph_json\x18\x01 \x01(\t\"\xaa\x01\n\x18\x43reateEnvironmentRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x16\n\x0e\x65nvironment_id\x18\x02 \x01(\t\x12\x0e\n\x06source\x18\x03 \x01(\t\x12:\n\ndata_store\x18\x04 \x01(\x0b\x32&.continual.rpc.management.v1.DataStore\x12\x1a\n\x12scheduling_enabled\x18\x05 \x01(\x08\"\x8a\x01\n\x18UpdateEnvironmentRequest\x12=\n\x0b\x65nvironment\x18\x01 \x01(\x0b\x32(.continual.rpc.management.v1.Environment\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"%\n\x15GetEnvironmentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\",\n\x1b\x42\x61tchGetEnvironmentsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"^\n\x1c\x42\x61tchGetEnvironmentsResponse\x12>\n\x0c\x65nvironments\x18\x01 \x03(\x0b\x32(.continual.rpc.management.v1.Environment\"?\n\x18\x44\x65leteEnvironmentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rdelete_schema\x18\x02 \x01(\x08\"\xd7\x01\n\x17ListEnvironmentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"s\n\x18ListEnvironmentsResponse\x12>\n\x0c\x65nvironments\x18\x01 \x03(\x0b\x32(.continual.rpc.management.v1.Environment\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"/\n\x1fTestEnvironmentDataStoreRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"J\n TestEnvironmentDataStoreResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"\x19\n\x17GetSystemMessageRequest\"+\n\x18GetSystemMessageResponse\x12\x0f\n\x07message\x18\x01 \x01(\t2\xbf\xa4\x01\n\rManagementAPI\x12u\n\x07GetUser\x12+.continual.rpc.management.v1.GetUserRequest\x1a!.continual.rpc.management.v1.User\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/{name=users/*}\x12}\n\tListUsers\x12-.continual.rpc.management.v1.ListUsersRequest\x1a..continual.rpc.management.v1.ListUsersResponse\"\x11\x82\xd3\xe4\x93\x02\x0b\x12\t/v1/users\x12\x92\x01\n\rBatchGetUsers\x12\x31.continual.rpc.management.v1.BatchGetUsersRequest\x1a\x32.continual.rpc.management.v1.BatchGetUsersResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/users:batchGet\x12\x86\x01\n\nUpdateUser\x12..continual.rpc.management.v1.UpdateUserRequest\x1a!.continual.rpc.management.v1.User\"%\x82\xd3\xe4\x93\x02\x1f\x32\x17/v1/{user.name=users/*}:\x04user\x12p\n\nDeleteUser\x12..continual.rpc.management.v1.DeleteUserRequest\x1a\x16.google.protobuf.Empty\"\x1a\x82\xd3\xe4\x93\x02\x14*\x12/v1/{name=users/*}\x12Z\n\tGetViewer\x12\x16.google.protobuf.Empty\x1a!.continual.rpc.management.v1.User\"\x12\x82\xd3\xe4\x93\x02\x0c\x12\n/v1/viewer\x12\x83\x01\n\x08Register\x12,.continual.rpc.management.v1.RegisterRequest\x1a-.continual.rpc.management.v1.RegisterResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\"\x12/v1/users:register\x12w\n\x05Login\x12).continual.rpc.management.v1.LoginRequest\x1a*.continual.rpc.management.v1.LoginResponse\"\x17\x82\xd3\xe4\x93\x02\x11\"\x0f/v1/users:login\x12R\n\x06Logout\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x18\x82\xd3\xe4\x93\x02\x12\"\x10/v1/users:logout\x12\x93\x01\n\x15SendEmailVerification\x12\x39.continual.rpc.management.v1.SendEmailVerificationRequest\x1a\x16.google.protobuf.Empty\"\'\x82\xd3\xe4\x93\x02!\"\x1f/v1/users:sendEmailVerification\x12\x8f\x01\n\x0bVerifyEmail\x12/.continual.rpc.management.v1.VerifyEmailRequest\x1a\x30.continual.rpc.management.v1.VerifyEmailResponse\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x15/v1/users:verifyEmail\x12\x87\x01\n\x11SendPasswordReset\x12\x35.continual.rpc.management.v1.SendPasswordResetRequest\x1a\x16.google.protobuf.Empty\"#\x82\xd3\xe4\x93\x02\x1d\"\x1b/v1/users:sendPasswordReset\x12\x98\x01\n\rResetPassword\x12\x31.continual.rpc.management.v1.ResetPasswordRequest\x1a\x32.continual.rpc.management.v1.ResetPasswordResponse\" \x82\xd3\xe4\x93\x02\x1a\"\x18/v1/:users:resetPassword\x12\x89\x01\n\nGetSession\x12..continual.rpc.management.v1.GetSessionRequest\x1a$.continual.rpc.management.v1.Session\"%\x82\xd3\xe4\x93\x02\x1f\x12\x1d/v1/{name=users/*/sessions/*}\x12\xac\x01\n\x0cListSessions\x12\x30.continual.rpc.management.v1.ListSessionsRequest\x1a\x31.continual.rpc.management.v1.ListSessionsResponse\"7\x82\xd3\xe4\x93\x02\x1f\x12\x1d/v1/{parent=users/*}/sessions\xca\x45\x08sessions\xd2\x45\x04User\x12\x81\x01\n\rDeleteSession\x12\x31.continual.rpc.management.v1.DeleteSessionRequest\x1a\x16.google.protobuf.Empty\"%\x82\xd3\xe4\x93\x02\x1f*\x1d/v1/{name=users/*/sessions/*}\x12\x89\x01\n\x11\x44\x65leteAllSessions\x12\x35.continual.rpc.management.v1.DeleteAllSessionsRequest\x1a\x16.google.protobuf.Empty\"%\x82\xd3\xe4\x93\x02\x1f*\x1d/v1/{parent=users/*}/sessions\x12\x95\x01\n\x0fGetOrganization\x12\x33.continual.rpc.management.v1.GetOrganizationRequest\x1a).continual.rpc.management.v1.Organization\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/v1/{name=organizations/*}\x12\xb2\x01\n\x15\x42\x61tchGetOrganizations\x12\x39.continual.rpc.management.v1.BatchGetOrganizationsRequest\x1a:.continual.rpc.management.v1.BatchGetOrganizationsResponse\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/v1/organizations:batchGet\x12\x9d\x01\n\x11ListOrganizations\x12\x35.continual.rpc.management.v1.ListOrganizationsRequest\x1a\x36.continual.rpc.management.v1.ListOrganizationsResponse\"\x19\x82\xd3\xe4\x93\x02\x13\x12\x11/v1/organizations\x12\xa0\x01\n\x12\x43reateOrganization\x12\x36.continual.rpc.management.v1.CreateOrganizationRequest\x1a).continual.rpc.management.v1.Organization\"\'\x82\xd3\xe4\x93\x02!\"\x11/v1/organizations:\x0corganization\x12\xb6\x01\n\x12UpdateOrganization\x12\x36.continual.rpc.management.v1.UpdateOrganizationRequest\x1a).continual.rpc.management.v1.Organization\"=\x82\xd3\xe4\x93\x02\x37\x32\'/v1/{organization.name=organizations/*}:\x0corganization\x12\x88\x01\n\x12\x44\x65leteOrganization\x12\x36.continual.rpc.management.v1.DeleteOrganizationRequest\x1a\x16.google.protobuf.Empty\"\"\x82\xd3\xe4\x93\x02\x1c*\x1a/v1/{name=organizations/*}\x12\xcf\x01\n\rGetOnboarding\x12\x31.continual.rpc.management.v1.GetOnboardingRequest\x1a\'.continual.rpc.management.v1.Onboarding\"b\x82\xd3\xe4\x93\x02\'\x12%/v1/{name=organizations/*/onboarding}\xd2\x45\x0cOrganization\xca\x45\nonboarding\xda\x45\x16name:name:$/onboarding\x12\xb7\x01\n\x10UpdateOnboarding\x12\x34.continual.rpc.management.v1.UpdateOnboardingRequest\x1a\'.continual.rpc.management.v1.Onboarding\"D\x82\xd3\xe4\x93\x02>20/v1/{onboarding.name=organizations/*/onboarding}:\nonboarding\x12}\n\tListRoles\x12-.continual.rpc.management.v1.ListRolesRequest\x1a..continual.rpc.management.v1.ListRolesResponse\"\x11\x82\xd3\xe4\x93\x02\x0b\x12\t/v1/roles\x12u\n\x07GetRole\x12+.continual.rpc.management.v1.GetRoleRequest\x1a!.continual.rpc.management.v1.Role\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/{name=roles/*}\x12\x92\x01\n\rBatchGetRoles\x12\x31.continual.rpc.management.v1.BatchGetRolesRequest\x1a\x32.continual.rpc.management.v1.BatchGetRolesResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/roles:batchGet\x12\xae\x01\n\x12\x43reateAccessPolicy\x12\x36.continual.rpc.management.v1.CreateAccessPolicyRequest\x1a).continual.rpc.management.v1.AccessPolicy\"5\x82\xd3\xe4\x93\x02/\"\x1e/v1/{parent=**}/accessPolicies:\raccess_policy\x12\x99\x01\n\x0fGetAccessPolicy\x12\x33.continual.rpc.management.v1.GetAccessPolicyRequest\x1a).continual.rpc.management.v1.AccessPolicy\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=**/accessPolicies/*}\x12\x94\x02\n\x12ListAccessPolicies\x12\x36.continual.rpc.management.v1.ListAccessPoliciesRequest\x1a\x37.continual.rpc.management.v1.ListAccessPoliciesResponse\"\x8c\x01\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=**}/accessPolicies\xd2\x45\x44Organization,Project,FeatureSet,Model,ModelVersion,Experiment,Change\xca\x45\x0e\x61\x63\x63\x65ssPolicies\xda\x45\x0bname:parent\x12\x8c\x01\n\x12\x44\x65leteAccessPolicy\x12\x36.continual.rpc.management.v1.DeleteAccessPolicyRequest\x1a\x16.google.protobuf.Empty\"&\x82\xd3\xe4\x93\x02 *\x1e/v1/{name=**/accessPolicies/*}\x12\xbc\x01\n\x12UpdateAccessPolicy\x12\x36.continual.rpc.management.v1.UpdateAccessPolicyRequest\x1a).continual.rpc.management.v1.AccessPolicy\"C\x82\xd3\xe4\x93\x02=2,/v1/{access_policy.name=**/accessPolicies/*}:\raccess_policy\x12\x90\x02\n\x0fTestPermissions\x12\x33.continual.rpc.management.v1.TestPermissionsRequest\x1a\x34.continual.rpc.management.v1.TestPermissionsResponse\"\x91\x01\x82\xd3\xe4\x93\x02&\"!/v1/{resource=**}:testPermissions:\x01*\xd2\x45\x44Organization,Project,FeatureSet,Model,ModelVersion,Experiment,Change\xca\x45\x0bpermissions\xda\x45\rname:resource\x12\x86\x01\n\x0c\x43reateInvite\x12\x30.continual.rpc.management.v1.CreateInviteRequest\x1a#.continual.rpc.management.v1.Invite\"\x1f\x82\xd3\xe4\x93\x02\x19\"\x17/v1/{parent=**}/invites\x12\x80\x01\n\tGetInvite\x12-.continual.rpc.management.v1.GetInviteRequest\x1a#.continual.rpc.management.v1.Invite\"\x1f\x82\xd3\xe4\x93\x02\x19\x12\x17/v1/{name=**/invites/*}\x12\x91\x01\n\x0bListInvites\x12/.continual.rpc.management.v1.ListInvitesRequest\x1a\x30.continual.rpc.management.v1.ListInvitesResponse\"\x1f\x82\xd3\xe4\x93\x02\x19\x12\x17/v1/{parent=**}/invites\x12\x95\x01\n\x0cUpdateInvite\x12\x30.continual.rpc.management.v1.UpdateInviteRequest\x1a#.continual.rpc.management.v1.Invite\".\x82\xd3\xe4\x93\x02(2\x1e/v1/{invite.name=**/invites/*}:\x06invite\x12\x8d\x01\n\x0cResendInvite\x12\x30.continual.rpc.management.v1.ResendInviteRequest\x1a#.continual.rpc.management.v1.Invite\"&\x82\xd3\xe4\x93\x02 \"\x1e/v1/{name=**/invites/*}:resend\x12y\n\x0c\x44\x65leteInvite\x12\x30.continual.rpc.management.v1.DeleteInviteRequest\x1a\x16.google.protobuf.Empty\"\x1f\x82\xd3\xe4\x93\x02\x19*\x17/v1/{name=**/invites/*}\x12v\n\x0c\x41\x63\x63\x65ptInvite\x12\x30.continual.rpc.management.v1.AcceptInviteRequest\x1a\x16.google.protobuf.Empty\"\x1c\x82\xd3\xe4\x93\x02\x16\"\x14/v1/confirm/{code=*}\x12\x81\x01\n\nGetProject\x12..continual.rpc.management.v1.GetProjectRequest\x1a$.continual.rpc.management.v1.Project\"\x1d\x82\xd3\xe4\x93\x02\x17\x12\x15/v1/{name=projects/*}\x12\x9e\x01\n\x10\x42\x61tchGetProjects\x12\x34.continual.rpc.management.v1.BatchGetProjectsRequest\x1a\x35.continual.rpc.management.v1.BatchGetProjectsResponse\"\x1d\x82\xd3\xe4\x93\x02\x17\x12\x15/v1/projects:batchGet\x12\xda\x01\n\x0cListProjects\x12\x30.continual.rpc.management.v1.ListProjectsRequest\x1a\x31.continual.rpc.management.v1.ListProjectsResponse\"e\x82\xd3\xe4\x93\x02\x37\x12\x0c/v1/projectsZ\'\x12%/v1/{parent=organizations/*}/projects\xd2\x45\x0cOrganization\xca\x45\x08projects\xda\x45\x0bname:parent\x12\x87\x01\n\rCreateProject\x12\x31.continual.rpc.management.v1.CreateProjectRequest\x1a$.continual.rpc.management.v1.Project\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x0c/v1/projects:\x07project\x12\x98\x01\n\rUpdateProject\x12\x31.continual.rpc.management.v1.UpdateProjectRequest\x1a$.continual.rpc.management.v1.Project\".\x82\xd3\xe4\x93\x02(2\x1d/v1/{project.name=projects/*}:\x07project\x12y\n\rDeleteProject\x12\x31.continual.rpc.management.v1.DeleteProjectRequest\x1a\x16.google.protobuf.Empty\"\x1d\x82\xd3\xe4\x93\x02\x17*\x15/v1/{name=projects/*}\x12\xb8\x01\n\x14TestProjectDataStore\x12\x38.continual.rpc.management.v1.TestProjectDataStoreRequest\x1a\x39.continual.rpc.management.v1.TestProjectDataStoreResponse\"+\x82\xd3\xe4\x93\x02%\"#/v1/{name=projects/*}:testDataStore\x12\x95\x01\n\rTestDataStore\x12\x31.continual.rpc.management.v1.TestDataStoreRequest\x1a\x32.continual.rpc.management.v1.TestDataStoreResponse\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x15/v1/app:testDataStore\x12\xb5\x01\n\x0eGetSchemaGraph\x12\x32.continual.rpc.management.v1.GetSchemaGraphRequest\x1a\x33.continual.rpc.management.v1.GetSchemaGraphResponse\":\xd2\x45\x13Project,Environment\xca\x45\x0bschemaGraph\xda\x45\x13name:name;name:name\x12\x98\x01\n\rGetFeatureSet\x12\x31.continual.rpc.management.v1.GetFeatureSetRequest\x1a\'.continual.rpc.management.v1.FeatureSet\"+\x82\xd3\xe4\x93\x02%\x12#/v1/{name=projects/*/featureSets/*}\x12\xbe\x01\n\x13\x42\x61tchGetFeatureSets\x12\x37.continual.rpc.management.v1.BatchGetFeatureSetsRequest\x1a\x38.continual.rpc.management.v1.BatchGetFeatureSetsResponse\"4\x82\xd3\xe4\x93\x02.\x12,/v1/{parent=projects/*}/featureSets:batchGet\x12\xe7\x01\n\x0fListFeatureSets\x12\x33.continual.rpc.management.v1.ListFeatureSetsRequest\x1a\x34.continual.rpc.management.v1.ListFeatureSetsResponse\"i\x82\xd3\xe4\x93\x02%\x12#/v1/{parent=projects/*}/featureSets\xd2\x45\x13Project,Environment\xca\x45\x0b\x66\x65\x61tureSets\xda\x45\x17name:parent;name:parent\x12\x8d\x01\n\x10\x44\x65leteFeatureSet\x12\x34.continual.rpc.management.v1.DeleteFeatureSetRequest\x1a\x16.google.protobuf.Empty\"+\x82\xd3\xe4\x93\x02%*#/v1/{name=projects/*/featureSets/*}\x12\x64\n\x12ValidateFeatureSet\x12\x36.continual.rpc.management.v1.ValidateFeatureSetRequest\x1a\x16.google.protobuf.Empty\x12\x89\x01\n\tGetEntity\x12-.continual.rpc.management.v1.GetEntityRequest\x1a#.continual.rpc.management.v1.Entity\"(\x82\xd3\xe4\x93\x02\"\x12 /v1/{name=projects/*/entities/*}\x12\xd8\x01\n\x0cListEntities\x12\x30.continual.rpc.management.v1.ListEntitiesRequest\x1a\x31.continual.rpc.management.v1.ListEntitiesResponse\"c\x82\xd3\xe4\x93\x02\"\x12 /v1/{parent=projects/*}:entities\xd2\x45\x13Project,Environment\xca\x45\x08\x65ntities\xda\x45\x17name:parent;name:parent\x12\x8d\x01\n\x14GetEntityFeatureSets\x12\x38.continual.rpc.management.v1.GetEntityFeatureSetsRequest\x1a\x39.continual.rpc.management.v1.GetEntityFeatureSetsResponse\"\x00\x12\xe7\x01\n\x19GetFeatureSetDatasetStats\x12=.continual.rpc.management.v1.GetFeatureSetDatasetStatsRequest\x1a).continual.rpc.management.v1.DatasetStats\"`\x82\xd3\xe4\x93\x02\x32\x12\x30/v1/{name=projects/*/featureSets/*/datasetstats}\xd2\x45\nFeatureSet\xca\x45\x0c\x64\x61tasetStats\xda\x45\tname:name\x12\xd3\x01\n\x14GetModelDatasetStats\x12\x38.continual.rpc.management.v1.GetModelDatasetStatsRequest\x1a).continual.rpc.management.v1.DatasetStats\"V\x82\xd3\xe4\x93\x02-\x12+/v1/{name=projects/*/models/*/datasetstats}\xd2\x45\x05Model\xca\x45\x0c\x64\x61tasetStats\xda\x45\tname:name\x12\xf3\x01\n\x1bGetModelVersionDatasetStats\x12?.continual.rpc.management.v1.GetModelVersionDatasetStatsRequest\x1a).continual.rpc.management.v1.DatasetStats\"h\x82\xd3\xe4\x93\x02\x38\x12\x36/v1/{name=projects/*/models/*/versions/*/datasetstats}\xd2\x45\x0cModelVersion\xca\x45\x0c\x64\x61tasetStats\xda\x45\tname:name\x12\x84\x02\n\x1eGetBatchPredictionDatasetStats\x12\x42.continual.rpc.management.v1.GetBatchPredictionDatasetStatsRequest\x1a).continual.rpc.management.v1.DatasetStats\"s\x82\xd3\xe4\x93\x02@\x12>/v1/{name=projects/*/models/*/batchPredictions/*/datasetstats}\xd2\x45\x0f\x42\x61tchPrediction\xca\x45\x0c\x64\x61tasetStats\xda\x45\tname:name\x12\x81\x02\n\x12GetPredictionStats\x12\x36.continual.rpc.management.v1.GetPredictionStatsRequest\x1a).continual.rpc.management.v1.DatasetStats\"\x87\x01\x82\xd3\xe4\x93\x02\x43\x12\x41/v1/{name=projects/*/models/*/batchPredictions/*/predictionstats}\xd2\x45\x0f\x42\x61tchPrediction\xca\x45\x0fpredictionStats\xda\x45\x17name:name:$/predictions\x12\xcd\x01\n\x12GetFeatureSetStats\x12\x36.continual.rpc.management.v1.GetFeatureSetStatsRequest\x1a+.continual.rpc.management.v1.ColumnSetStats\"R\x82\xd3\xe4\x93\x02+\x12)/v1/{name=projects/*/featureSets/*/stats}\xd2\x45\nFeatureSet\xca\x45\x05stats\xda\x45\tname:name\x12\xb9\x01\n\rGetModelStats\x12\x31.continual.rpc.management.v1.GetModelStatsRequest\x1a+.continual.rpc.management.v1.ColumnSetStats\"H\x82\xd3\xe4\x93\x02&\x12$/v1/{name=projects/*/models/*/stats}\xd2\x45\x05Model\xca\x45\x05stats\xda\x45\tname:name\x12\x95\x01\n\x12\x43omputeColumnStats\x12\x36.continual.rpc.management.v1.ComputeColumnStatsRequest\x1a\x16.google.protobuf.Empty\"/\x82\xd3\xe4\x93\x02)\"\'/v1/{name=projects/*/*/*}/stats:compute\x12\x64\n\x07GetData\x12+.continual.rpc.management.v1.GetDataRequest\x1a,.continual.rpc.management.v1.GetDataResponse\x12\xce\x01\n\x1c\x43reateFeatureSetCSVSignedURL\x12@.continual.rpc.management.v1.CreateFeatureSetCSVSignedURLRequest\x1a\x41.continual.rpc.management.v1.CreateFeatureSetCSVSignedURLResponse\")\x82\xd3\xe4\x93\x02#\"!/v1/{parent=projects/*}:uploadCsv\x12\x8d\x01\n\x0c\x43reateChange\x12\x30.continual.rpc.management.v1.CreateChangeRequest\x1a#.continual.rpc.management.v1.Change\"&\x82\xd3\xe4\x93\x02 \"\x1e/v1/{parent=projects/*}:change\x12W\n\x05ReRun\x12).continual.rpc.management.v1.ReRunRequest\x1a#.continual.rpc.management.v1.Change\x12\x87\x01\n\tGetChange\x12-.continual.rpc.management.v1.GetChangeRequest\x1a#.continual.rpc.management.v1.Change\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=projects/*}/change/*\x12\xd3\x01\n\x0bListChanges\x12/.continual.rpc.management.v1.ListChangesRequest\x1a\x30.continual.rpc.management.v1.ListChangesResponse\"a\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=projects/*}/change\xca\x45\x05plans\xd2\x45\x13Project,Environment\xe0\x45\x00\xda\x45\x17name:parent;name:parent\x12\x87\x01\n\x0c\x43\x61ncelChange\x12\x30.continual.rpc.management.v1.CancelChangeRequest\x1a\x16.google.protobuf.Empty\"-\x82\xd3\xe4\x93\x02\'\"%/v1/{name=projects/*/change/*}:cancel\x12\xb1\x01\n\x13PromoteModelVersion\x12\x37.continual.rpc.management.v1.PromoteModelVersionRequest\x1a&.continual.rpc.management.v1.Promotion\"9\x82\xd3\xe4\x93\x02\x33\x12\x31/v1/{name=projects/*/models/*/versions/*}:promote\x12\x9c\x02\n\x0eListPromotions\x12\x32.continual.rpc.management.v1.ListPromotionsRequest\x1a\x33.continual.rpc.management.v1.ListPromotionsResponse\"\xa0\x01\x82\xd3\xe4\x93\x02g\x12+/v1/{parent=projects/*/models/*}/promotionsZ8\x12\x36/v1/{parent=projects/*/models/*/versions/*}/promotions\xca\x45\npromotions\xd2\x45\x12ModelVersion,Model\xe0\x45\x00\xda\x45\x0bname:parent\x12\x9d\x01\n\x0cGetPromotion\x12\x30.continual.rpc.management.v1.GetPromotionRequest\x1a&.continual.rpc.management.v1.Promotion\"3\x82\xd3\xe4\x93\x02-\x12+/v1/{name=projects/*/models/*/promotions/*}\x12\xc3\x01\n\x12\x42\x61tchGetPromotions\x12\x36.continual.rpc.management.v1.BatchGetPromotionsRequest\x1a\x37.continual.rpc.management.v1.BatchGetPromotionsResponse\"<\x82\xd3\xe4\x93\x02\x36\x12\x34/v1/{parent=projects/*/models/*}/promotions:batchGet\x12\x95\x01\n\nListEvents\x12..continual.rpc.management.v1.ListEventsRequest\x1a/.continual.rpc.management.v1.ListEventsResponse\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=projects/*}:events\x12\x84\x01\n\x08GetEvent\x12,.continual.rpc.management.v1.GetEventRequest\x1a\".continual.rpc.management.v1.Event\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=projects/*/events/*}\x12\xaa\x01\n\x0e\x42\x61tchGetEvents\x12\x32.continual.rpc.management.v1.BatchGetEventsRequest\x1a\x33.continual.rpc.management.v1.BatchGetEventsResponse\"/\x82\xd3\xe4\x93\x02)\x12\'/v1/{parent=projects/*}/events:batchGet\x12\xd9\x01\n\x0fGetSubscription\x12\x33.continual.rpc.management.v1.GetSubscriptionRequest\x1a).continual.rpc.management.v1.Subscription\"f\x82\xd3\xe4\x93\x02\'\x12%/v1/{name=organizations/subscription}\xd2\x45\x0cOrganization\xca\x45\x0csubscription\xda\x45\x18name:name:$/subscription\x12\xe7\x01\n\x11GetBillingAccount\x12\x35.continual.rpc.management.v1.GetBillingAccountRequest\x1a+.continual.rpc.management.v1.BillingAccount\"n\x82\xd3\xe4\x93\x02+\x12)/v1/{name=organizations/*/billingAccount}\xd2\x45\x0cOrganization\xca\x45\x0e\x62illingAccount\xda\x45\x1aname:name:$/billingAccount\x12\xb8\x01\n\x08GetUsage\x12,.continual.rpc.management.v1.GetUsageRequest\x1a-.continual.rpc.management.v1.GetUsageResponse\"O\x82\xd3\xe4\x93\x02$\"\"/v1/{parent=organizations/*}:usage\xd2\x45\x0cOrganization\xca\x45\x05usage\xda\x45\x0bname:parent\x12\xca\x01\n\x0cListInvoices\x12\x30.continual.rpc.management.v1.ListInvoicesRequest\x1a\x31.continual.rpc.management.v1.ListInvoicesResponse\"U\x82\xd3\xe4\x93\x02\'\x12%/v1/{parent=organizations/*}/invoices\xd2\x45\x0cOrganization\xca\x45\x08invoices\xda\x45\x0bname:parent\x12\xc4\x01\n\x14UpdateBillingAccount\x12\x38.continual.rpc.management.v1.UpdateBillingAccountRequest\x1a+.continual.rpc.management.v1.BillingAccount\"E\x82\xd3\xe4\x93\x02?2,/v1/{billing_account.name=**/billingAccount}:\x0f\x62illing_account\x12\xb6\x01\n\x12UpdateSubscription\x12\x36.continual.rpc.management.v1.UpdateSubscriptionRequest\x1a).continual.rpc.management.v1.Subscription\"=\x82\xd3\xe4\x93\x02\x37\x32\'/v1/{subscription.name=**/subscription}:\x0csubscription\x12\xc0\x01\n\nListPrices\x12..continual.rpc.management.v1.ListPricesRequest\x1a/.continual.rpc.management.v1.ListPricesResponse\"Q\x82\xd3\xe4\x93\x02%\x12#/v1/{parent=organizations/*}/prices\xd2\x45\x0cOrganization\xca\x45\x06prices\xda\x45\x0bname:parent\x12\x89\x01\n\x08GetPrice\x12,.continual.rpc.management.v1.GetPriceRequest\x1a\".continual.rpc.management.v1.Price\"+\x82\xd3\xe4\x93\x02%\x12#/v1/{name=organizations/*/prices/*}\x12\xa9\x01\n\x10GetPaymentMethod\x12\x34.continual.rpc.management.v1.GetPaymentMethodRequest\x1a*.continual.rpc.management.v1.PaymentMethod\"3\x82\xd3\xe4\x93\x02-\x12+/v1/{name=organizations/*/paymentMethods/*}\x12\xe5\x01\n\x12GetUpcomingInvoice\x12\x36.continual.rpc.management.v1.GetUpcomingInvoiceRequest\x1a$.continual.rpc.management.v1.Invoice\"q\x82\xd3\xe4\x93\x02,\x12*/v1/{name=organizations/*/upcomingInvoice}\xd2\x45\x0cOrganization\xca\x45\x0fupcomingInvoice\xda\x45\x1bname:name:$/upcomingInvoice\x12\x82\x01\n\x11GetInferredSchema\x12\x35.continual.rpc.management.v1.GetInferredSchemaRequest\x1a\x36.continual.rpc.management.v1.GetInferredSchemaResponse\x12y\n\x0eGetServingData\x12\x32.continual.rpc.management.v1.GetServingDataRequest\x1a\x33.continual.rpc.management.v1.GetServingDataResponse\x12|\n\x0fGetTrainingData\x12\x33.continual.rpc.management.v1.GetTrainingDataRequest\x1a\x34.continual.rpc.management.v1.GetTrainingDataResponse\x12\x88\x01\n\x13\x42rowseDataWarehouse\x12\x37.continual.rpc.management.v1.BrowseDataWarehouseRequest\x1a\x38.continual.rpc.management.v1.BrowseDataWarehouseResponse\x12[\n\x04Seed\x12(.continual.rpc.management.v1.SeedRequest\x1a).continual.rpc.management.v1.SeedResponse\x12\x84\x01\n\x08GetModel\x12,.continual.rpc.management.v1.GetModelRequest\x1a\".continual.rpc.management.v1.Model\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=projects/*/models/*}\x12\xaa\x01\n\x0e\x42\x61tchGetModels\x12\x32.continual.rpc.management.v1.BatchGetModelsRequest\x1a\x33.continual.rpc.management.v1.BatchGetModelsResponse\"/\x82\xd3\xe4\x93\x02)\x12\'/v1/{parent=projects/*}/models:batchGet\x12\xce\x01\n\nListModels\x12..continual.rpc.management.v1.ListModelsRequest\x1a/.continual.rpc.management.v1.ListModelsResponse\"_\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=projects/*}/models\xd2\x45\x13Project,Environment\xca\x45\x06models\xda\x45\x17name:parent;name:parent\x12~\n\x0b\x44\x65leteModel\x12/.continual.rpc.management.v1.DeleteModelRequest\x1a\x16.google.protobuf.Empty\"&\x82\xd3\xe4\x93\x02 *\x1e/v1/{name=projects/*/models/*}\x12\\\n\rValidateModel\x12\x31.continual.rpc.management.v1.ValidateModelRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x99\x01\n\rGetModelQuery\x12\x31.continual.rpc.management.v1.GetModelQueryRequest\x1a\x32.continual.rpc.management.v1.GetModelQueryResponse\"!\xd2\x45\x05Model\xca\x45\nmodelQuery\xda\x45\tname:name\x12\xa4\x01\n\x0fGetModelVersion\x12\x33.continual.rpc.management.v1.GetModelVersionRequest\x1a).continual.rpc.management.v1.ModelVersion\"1\x82\xd3\xe4\x93\x02+\x12)/v1/{name=projects/*/models/*/versions/*}\x12\xc7\x01\n\x14GetModelVersionQuery\x12\x38.continual.rpc.management.v1.GetModelVersionQueryRequest\x1a\x39.continual.rpc.management.v1.GetModelVersionQueryResponse\":\x82\xd3\xe4\x93\x02\x34\x12\x32/v1/{name=projects/*/models/*/versions/*}:getQuery\x12\xca\x01\n\x15\x42\x61tchGetModelVersions\x12\x39.continual.rpc.management.v1.BatchGetModelVersionsRequest\x1a:.continual.rpc.management.v1.BatchGetModelVersionsResponse\":\x82\xd3\xe4\x93\x02\x34\x12\x32/v1/{parent=projects/*/models/*}/versions:batchGet\x12\xa1\x02\n\x11ListModelVersions\x12\x35.continual.rpc.management.v1.ListModelVersionsRequest\x1a\x36.continual.rpc.management.v1.ListModelVersionsResponse\"\x9c\x01\x82\xd3\xe4\x93\x02+\x12)/v1/{parent=projects/*/models/*}/versions\xca\x45\rmodelVersions\xd2\x45\x19Model,Project,Environment\xe0\x45\x00\xda\x45\x39name:parent;name:parent:$/models/-;name:parent:$/models/-\x12\xa0\x01\n\nTrainModel\x12..continual.rpc.management.v1.TrainModelRequest\x1a).continual.rpc.management.v1.ModelVersion\"7\x82\xd3\xe4\x93\x02\x31\"//v1/{name=projects/*/models/*/versions/*}:train\x12\x9e\x01\n\x0e\x43\x61ncelTraining\x12\x32.continual.rpc.management.v1.CancelTrainingRequest\x1a\x16.google.protobuf.Empty\"@\x82\xd3\xe4\x93\x02:\"8/v1/{name=projects/*/models/*/versions/*}:cancelTraining\x12\xad\x01\n\x0c\x42\x61tchPredict\x12\x39.continual.rpc.management.v1.CreateBatchPredictionRequest\x1a,.continual.rpc.management.v1.BatchPrediction\"4\x82\xd3\xe4\x93\x02.\",/v1/{model=projects/*/models/*}:batchPredict\x12\xb2\x01\n\x12GetBatchPrediction\x12\x33.continual.rpc.management.v1.BatchPredictionRequest\x1a,.continual.rpc.management.v1.BatchPrediction\"9\x82\xd3\xe4\x93\x02\x33\x12\x31/v1/{name=projects/*/models/*/batchPredictions/*}\x12\xdb\x01\n\x18\x42\x61tchGetBatchPredictions\x12<.continual.rpc.management.v1.BatchGetBatchPredictionsRequest\x1a=.continual.rpc.management.v1.BatchGetBatchPredictionsResponse\"B\x82\xd3\xe4\x93\x02<\x12:/v1/{parent=projects/*/models/*}/batchPredictions:batchGet\x12\x9f\x02\n\x14ListBatchPredictions\x12\x38.continual.rpc.management.v1.ListBatchPredictionsRequest\x1a\x39.continual.rpc.management.v1.ListBatchPredictionsResponse\"\x91\x01\x82\xd3\xe4\x93\x02\x33\x12\x31/v1/{parent=projects/*/models/*}/batchPredictions\xca\x45\x10\x62\x61tchPredictions\xd2\x45\x19Project,Model,Environment\xe0\x45\x00\xda\x45#name:parent;name:parent;name:parent\x12\xbd\x01\n\x15\x43\x61ncelBatchPrediction\x12\x33.continual.rpc.management.v1.BatchPredictionRequest\x1a\x16.google.protobuf.Empty\"W\x82\xd3\xe4\x93\x02Q\"O/v1/{name=projects/*/models/*/versions/*/batchPredictions/*}:cancelBatchPredict\x12\xd1\x01\n\x12GetTrainingMetrics\x12\x36.continual.rpc.management.v1.GetTrainingMetricsRequest\x1a\x37.continual.rpc.management.v1.GetTrainingMetricsResponse\"J\x82\xd3\xe4\x93\x02\x44\x12\x42/v1/{experiment=projects/*/models/*/versions/*}:getTrainingMetrics\x12\xd0\x01\n\x13ListTrainingMetrics\x12\x37.continual.rpc.management.v1.ListTrainingMetricsRequest\x1a\x38.continual.rpc.management.v1.ListTrainingMetricsResponse\"F\x82\xd3\xe4\x93\x02@\x12>/v1/{model=projects/*/models/*/versions/*}:listTrainingMetrics\x12\xac\x01\n\rGetExperiment\x12\x31.continual.rpc.management.v1.GetExperimentRequest\x1a\'.continual.rpc.management.v1.Experiment\"?\x82\xd3\xe4\x93\x02\x39\x12\x37/v1/{name=projects/*/models/*/versions/*/experiments/*}\x12\xd2\x01\n\x13\x42\x61tchGetExperiments\x12\x37.continual.rpc.management.v1.BatchGetExperimentsRequest\x1a\x38.continual.rpc.management.v1.BatchGetExperimentsResponse\"H\x82\xd3\xe4\x93\x02\x42\x12@/v1/{parent=projects/*/models/*/versions/*}/experiments:batchGet\x12\xeb\x01\n\x0fListExperiments\x12\x33.continual.rpc.management.v1.ListExperimentsRequest\x1a\x34.continual.rpc.management.v1.ListExperimentsResponse\"m\x82\xd3\xe4\x93\x02\x39\x12\x37/v1/{parent=projects/*/models/*/versions/*}/experiments\xca\x45\x0b\x65xperiments\xd2\x45\x0cModelVersion\xe0\x45\x00\xda\x45\x0bname:parent\x12l\n\x0bGetYamlDiff\x12,.continual.rpc.management.v1.YamlDiffRequest\x1a-.continual.rpc.management.v1.YamlDiffResponse\"\x00\x12p\n\x0eGetEnvironment\x12\x32.continual.rpc.management.v1.GetEnvironmentRequest\x1a(.continual.rpc.management.v1.Environment\"\x00\x12\x8d\x01\n\x14\x42\x61tchGetEnvironments\x12\x38.continual.rpc.management.v1.BatchGetEnvironmentsRequest\x1a\x39.continual.rpc.management.v1.BatchGetEnvironmentsResponse\"\x00\x12\xa8\x01\n\x10ListEnvironments\x12\x34.continual.rpc.management.v1.ListEnvironmentsRequest\x1a\x35.continual.rpc.management.v1.ListEnvironmentsResponse\"\'\xd2\x45\x07Project\xca\x45\x0c\x65nvironments\xda\x45\x0bname:parent\x12v\n\x11\x43reateEnvironment\x12\x35.continual.rpc.management.v1.CreateEnvironmentRequest\x1a(.continual.rpc.management.v1.Environment\"\x00\x12v\n\x11UpdateEnvironment\x12\x35.continual.rpc.management.v1.UpdateEnvironmentRequest\x1a(.continual.rpc.management.v1.Environment\"\x00\x12\x64\n\x11\x44\x65leteEnvironment\x12\x35.continual.rpc.management.v1.DeleteEnvironmentRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x99\x01\n\x18TestEnvironmentDataStore\x12<.continual.rpc.management.v1.TestEnvironmentDataStoreRequest\x1a=.continual.rpc.management.v1.TestEnvironmentDataStoreResponse\"\x00\x12\x9a\x01\n\x10GetSystemMessage\x12\x34.continual.rpc.management.v1.GetSystemMessageRequest\x1a\x35.continual.rpc.management.v1.GetSystemMessageResponse\"\x19\x82\xd3\xe4\x93\x02\x13\"\x11/v1/systemMessageBKZIgithub.com/continualiq/continual/continual/rpc/management/v1;managementv1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_api_dot_annotations__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,google_dot_protobuf_dot_field__mask__pb2.DESCRIPTOR,continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.DESCRIPTOR,continual_dot_rpc_dot_graphql_dot_options__pb2.DESCRIPTOR,])



_BROWSEDATAWAREHOUSEREQUEST_OBJECTTYPE = _descriptor.EnumDescriptor(
  name='ObjectType',
  full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest.ObjectType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DATABASE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SCHEMA', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TABLE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=13719,
  serialized_end=13780,
)
_sym_db.RegisterEnumDescriptor(_BROWSEDATAWAREHOUSEREQUEST_OBJECTTYPE)


_UPDATEUSERREQUEST = _descriptor.Descriptor(
  name='UpdateUserRequest',
  full_name='continual.rpc.management.v1.UpdateUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.UpdateUserRequest.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateUserRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=292,
  serialized_end=409,
)


_GETUSERREQUEST = _descriptor.Descriptor(
  name='GetUserRequest',
  full_name='continual.rpc.management.v1.GetUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetUserRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=411,
  serialized_end=441,
)


_BATCHGETUSERSREQUEST = _descriptor.Descriptor(
  name='BatchGetUsersRequest',
  full_name='continual.rpc.management.v1.BatchGetUsersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetUsersRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=443,
  serialized_end=480,
)


_BATCHGETUSERSRESPONSE = _descriptor.Descriptor(
  name='BatchGetUsersResponse',
  full_name='continual.rpc.management.v1.BatchGetUsersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='users', full_name='continual.rpc.management.v1.BatchGetUsersResponse.users', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=482,
  serialized_end=555,
)


_DELETEUSERREQUEST = _descriptor.Descriptor(
  name='DeleteUserRequest',
  full_name='continual.rpc.management.v1.DeleteUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteUserRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=557,
  serialized_end=590,
)


_LISTUSERSREQUEST = _descriptor.Descriptor(
  name='ListUsersRequest',
  full_name='continual.rpc.management.v1.ListUsersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListUsersRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListUsersRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=592,
  serialized_end=649,
)


_LISTUSERSRESPONSE = _descriptor.Descriptor(
  name='ListUsersResponse',
  full_name='continual.rpc.management.v1.ListUsersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='users', full_name='continual.rpc.management.v1.ListUsersResponse.users', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListUsersResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=651,
  serialized_end=745,
)


_REGISTERREQUEST = _descriptor.Descriptor(
  name='RegisterRequest',
  full_name='continual.rpc.management.v1.RegisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='full_name', full_name='continual.rpc.management.v1.RegisterRequest.full_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.RegisterRequest.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.RegisterRequest.password', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='continual.rpc.management.v1.RegisterRequest.first_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_name', full_name='continual.rpc.management.v1.RegisterRequest.last_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=747,
  serialized_end=855,
)


_REGISTERRESPONSE = _descriptor.Descriptor(
  name='RegisterResponse',
  full_name='continual.rpc.management.v1.RegisterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.RegisterResponse.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_token', full_name='continual.rpc.management.v1.RegisterResponse.auth_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=857,
  serialized_end=944,
)


_LOGINREQUEST = _descriptor.Descriptor(
  name='LoginRequest',
  full_name='continual.rpc.management.v1.LoginRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.LoginRequest.email', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.LoginRequest.password', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=946,
  serialized_end=993,
)


_LOGINRESPONSE = _descriptor.Descriptor(
  name='LoginResponse',
  full_name='continual.rpc.management.v1.LoginResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.LoginResponse.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_token', full_name='continual.rpc.management.v1.LoginResponse.auth_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=995,
  serialized_end=1079,
)


_SENDEMAILVERIFICATIONREQUEST = _descriptor.Descriptor(
  name='SendEmailVerificationRequest',
  full_name='continual.rpc.management.v1.SendEmailVerificationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.SendEmailVerificationRequest.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1081,
  serialized_end=1125,
)


_VERIFYEMAILREQUEST = _descriptor.Descriptor(
  name='VerifyEmailRequest',
  full_name='continual.rpc.management.v1.VerifyEmailRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.VerifyEmailRequest.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.VerifyEmailRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1127,
  serialized_end=1175,
)


_VERIFYEMAILRESPONSE = _descriptor.Descriptor(
  name='VerifyEmailResponse',
  full_name='continual.rpc.management.v1.VerifyEmailResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.VerifyEmailResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1177,
  serialized_end=1215,
)


_SENDPASSWORDRESETREQUEST = _descriptor.Descriptor(
  name='SendPasswordResetRequest',
  full_name='continual.rpc.management.v1.SendPasswordResetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.SendPasswordResetRequest.email', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1217,
  serialized_end=1258,
)


_GETPASSWORDRESETREQUEST = _descriptor.Descriptor(
  name='GetPasswordResetRequest',
  full_name='continual.rpc.management.v1.GetPasswordResetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.GetPasswordResetRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1260,
  serialized_end=1299,
)


_RESETPASSWORDREQUEST = _descriptor.Descriptor(
  name='ResetPasswordRequest',
  full_name='continual.rpc.management.v1.ResetPasswordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.ResetPasswordRequest.password', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.ResetPasswordRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1301,
  serialized_end=1355,
)


_RESETPASSWORDRESPONSE = _descriptor.Descriptor(
  name='ResetPasswordResponse',
  full_name='continual.rpc.management.v1.ResetPasswordResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.ResetPasswordResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1357,
  serialized_end=1397,
)


_GETSESSIONREQUEST = _descriptor.Descriptor(
  name='GetSessionRequest',
  full_name='continual.rpc.management.v1.GetSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetSessionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1399,
  serialized_end=1432,
)


_LISTSESSIONSREQUEST = _descriptor.Descriptor(
  name='ListSessionsRequest',
  full_name='continual.rpc.management.v1.ListSessionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListSessionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListSessionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListSessionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1434,
  serialized_end=1510,
)


_LISTSESSIONSRESPONSE = _descriptor.Descriptor(
  name='ListSessionsResponse',
  full_name='continual.rpc.management.v1.ListSessionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sessions', full_name='continual.rpc.management.v1.ListSessionsResponse.sessions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListSessionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1512,
  serialized_end=1615,
)


_DELETESESSIONREQUEST = _descriptor.Descriptor(
  name='DeleteSessionRequest',
  full_name='continual.rpc.management.v1.DeleteSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteSessionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1617,
  serialized_end=1653,
)


_DELETEALLSESSIONSREQUEST = _descriptor.Descriptor(
  name='DeleteAllSessionsRequest',
  full_name='continual.rpc.management.v1.DeleteAllSessionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.DeleteAllSessionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1655,
  serialized_end=1697,
)


_LISTROLESREQUEST = _descriptor.Descriptor(
  name='ListRolesRequest',
  full_name='continual.rpc.management.v1.ListRolesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListRolesRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListRolesRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1699,
  serialized_end=1756,
)


_LISTROLESRESPONSE = _descriptor.Descriptor(
  name='ListRolesResponse',
  full_name='continual.rpc.management.v1.ListRolesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='roles', full_name='continual.rpc.management.v1.ListRolesResponse.roles', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListRolesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1758,
  serialized_end=1852,
)


_GETROLEREQUEST = _descriptor.Descriptor(
  name='GetRoleRequest',
  full_name='continual.rpc.management.v1.GetRoleRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetRoleRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1854,
  serialized_end=1884,
)


_BATCHGETROLESREQUEST = _descriptor.Descriptor(
  name='BatchGetRolesRequest',
  full_name='continual.rpc.management.v1.BatchGetRolesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetRolesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetRolesRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1886,
  serialized_end=1939,
)


_BATCHGETROLESRESPONSE = _descriptor.Descriptor(
  name='BatchGetRolesResponse',
  full_name='continual.rpc.management.v1.BatchGetRolesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='roles', full_name='continual.rpc.management.v1.BatchGetRolesResponse.roles', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1941,
  serialized_end=2014,
)


_LISTACCESSPOLICIESREQUEST = _descriptor.Descriptor(
  name='ListAccessPoliciesRequest',
  full_name='continual.rpc.management.v1.ListAccessPoliciesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2016,
  serialized_end=2098,
)


_LISTACCESSPOLICIESRESPONSE = _descriptor.Descriptor(
  name='ListAccessPoliciesResponse',
  full_name='continual.rpc.management.v1.ListAccessPoliciesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_policies', full_name='continual.rpc.management.v1.ListAccessPoliciesResponse.access_policies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListAccessPoliciesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2100,
  serialized_end=2221,
)


_CREATEACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='CreateAccessPolicyRequest',
  full_name='continual.rpc.management.v1.CreateAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateAccessPolicyRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='access_policy', full_name='continual.rpc.management.v1.CreateAccessPolicyRequest.access_policy', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2223,
  serialized_end=2332,
)


_GETACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='GetAccessPolicyRequest',
  full_name='continual.rpc.management.v1.GetAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetAccessPolicyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2334,
  serialized_end=2372,
)


_UPDATEACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='UpdateAccessPolicyRequest',
  full_name='continual.rpc.management.v1.UpdateAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_policy', full_name='continual.rpc.management.v1.UpdateAccessPolicyRequest.access_policy', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateAccessPolicyRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2375,
  serialized_end=2517,
)


_DELETEACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='DeleteAccessPolicyRequest',
  full_name='continual.rpc.management.v1.DeleteAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteAccessPolicyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2519,
  serialized_end=2560,
)


_CREATEINVITEREQUEST = _descriptor.Descriptor(
  name='CreateInviteRequest',
  full_name='continual.rpc.management.v1.CreateInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateInviteRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite', full_name='continual.rpc.management.v1.CreateInviteRequest.invite', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2562,
  serialized_end=2652,
)


_GETINVITEREQUEST = _descriptor.Descriptor(
  name='GetInviteRequest',
  full_name='continual.rpc.management.v1.GetInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetInviteRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2654,
  serialized_end=2686,
)


_LISTINVITESREQUEST = _descriptor.Descriptor(
  name='ListInvitesRequest',
  full_name='continual.rpc.management.v1.ListInvitesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListInvitesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListInvitesRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListInvitesRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2688,
  serialized_end=2763,
)


_LISTINVITESRESPONSE = _descriptor.Descriptor(
  name='ListInvitesResponse',
  full_name='continual.rpc.management.v1.ListInvitesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invites', full_name='continual.rpc.management.v1.ListInvitesResponse.invites', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListInvitesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2765,
  serialized_end=2865,
)


_UPDATEINVITEREQUEST = _descriptor.Descriptor(
  name='UpdateInviteRequest',
  full_name='continual.rpc.management.v1.UpdateInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invite', full_name='continual.rpc.management.v1.UpdateInviteRequest.invite', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateInviteRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2867,
  serialized_end=2990,
)


_DELETEINVITEREQUEST = _descriptor.Descriptor(
  name='DeleteInviteRequest',
  full_name='continual.rpc.management.v1.DeleteInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteInviteRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2992,
  serialized_end=3027,
)


_ACCEPTINVITEREQUEST = _descriptor.Descriptor(
  name='AcceptInviteRequest',
  full_name='continual.rpc.management.v1.AcceptInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.AcceptInviteRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3029,
  serialized_end=3064,
)


_RESENDINVITEREQUEST = _descriptor.Descriptor(
  name='ResendInviteRequest',
  full_name='continual.rpc.management.v1.ResendInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ResendInviteRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3066,
  serialized_end=3101,
)


_TESTPERMISSIONSREQUEST = _descriptor.Descriptor(
  name='TestPermissionsRequest',
  full_name='continual.rpc.management.v1.TestPermissionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.TestPermissionsRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='permissions', full_name='continual.rpc.management.v1.TestPermissionsRequest.permissions', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3103,
  serialized_end=3166,
)


_TESTPERMISSIONSRESPONSE = _descriptor.Descriptor(
  name='TestPermissionsResponse',
  full_name='continual.rpc.management.v1.TestPermissionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='permissions', full_name='continual.rpc.management.v1.TestPermissionsResponse.permissions', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3168,
  serialized_end=3214,
)


_CREATEORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='CreateOrganizationRequest',
  full_name='continual.rpc.management.v1.CreateOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.CreateOrganizationRequest.organization', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3216,
  serialized_end=3308,
)


_UPDATEORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='UpdateOrganizationRequest',
  full_name='continual.rpc.management.v1.UpdateOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.UpdateOrganizationRequest.organization', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateOrganizationRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3311,
  serialized_end=3452,
)


_GETORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='GetOrganizationRequest',
  full_name='continual.rpc.management.v1.GetOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetOrganizationRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3454,
  serialized_end=3492,
)


_BATCHGETORGANIZATIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetOrganizationsRequest',
  full_name='continual.rpc.management.v1.BatchGetOrganizationsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetOrganizationsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3494,
  serialized_end=3539,
)


_BATCHGETORGANIZATIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetOrganizationsResponse',
  full_name='continual.rpc.management.v1.BatchGetOrganizationsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organizations', full_name='continual.rpc.management.v1.BatchGetOrganizationsResponse.organizations', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3541,
  serialized_end=3638,
)


_DELETEORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='DeleteOrganizationRequest',
  full_name='continual.rpc.management.v1.DeleteOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteOrganizationRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3640,
  serialized_end=3681,
)


_LISTORGANIZATIONSREQUEST = _descriptor.Descriptor(
  name='ListOrganizationsRequest',
  full_name='continual.rpc.management.v1.ListOrganizationsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListOrganizationsRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListOrganizationsRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3683,
  serialized_end=3748,
)


_LISTORGANIZATIONSRESPONSE = _descriptor.Descriptor(
  name='ListOrganizationsResponse',
  full_name='continual.rpc.management.v1.ListOrganizationsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organizations', full_name='continual.rpc.management.v1.ListOrganizationsResponse.organizations', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListOrganizationsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3750,
  serialized_end=3868,
)


_GETONBOARDINGREQUEST = _descriptor.Descriptor(
  name='GetOnboardingRequest',
  full_name='continual.rpc.management.v1.GetOnboardingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetOnboardingRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3870,
  serialized_end=3906,
)


_UPDATEONBOARDINGREQUEST = _descriptor.Descriptor(
  name='UpdateOnboardingRequest',
  full_name='continual.rpc.management.v1.UpdateOnboardingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='onboarding', full_name='continual.rpc.management.v1.UpdateOnboardingRequest.onboarding', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateOnboardingRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3909,
  serialized_end=4044,
)


_CREATEPROJECTREQUEST = _descriptor.Descriptor(
  name='CreateProjectRequest',
  full_name='continual.rpc.management.v1.CreateProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.CreateProjectRequest.project', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='continual.rpc.management.v1.CreateProjectRequest.project_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4046,
  serialized_end=4143,
)


_UPDATEPROJECTREQUEST = _descriptor.Descriptor(
  name='UpdateProjectRequest',
  full_name='continual.rpc.management.v1.UpdateProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.UpdateProjectRequest.project', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateProjectRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4145,
  serialized_end=4271,
)


_GETPROJECTREQUEST = _descriptor.Descriptor(
  name='GetProjectRequest',
  full_name='continual.rpc.management.v1.GetProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetProjectRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4273,
  serialized_end=4306,
)


_BATCHGETPROJECTSREQUEST = _descriptor.Descriptor(
  name='BatchGetProjectsRequest',
  full_name='continual.rpc.management.v1.BatchGetProjectsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetProjectsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4308,
  serialized_end=4348,
)


_BATCHGETPROJECTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetProjectsResponse',
  full_name='continual.rpc.management.v1.BatchGetProjectsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='projects', full_name='continual.rpc.management.v1.BatchGetProjectsResponse.projects', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4350,
  serialized_end=4432,
)


_DELETEPROJECTREQUEST = _descriptor.Descriptor(
  name='DeleteProjectRequest',
  full_name='continual.rpc.management.v1.DeleteProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteProjectRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delete_schema', full_name='continual.rpc.management.v1.DeleteProjectRequest.delete_schema', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4434,
  serialized_end=4493,
)


_LISTPROJECTSREQUEST = _descriptor.Descriptor(
  name='ListProjectsRequest',
  full_name='continual.rpc.management.v1.ListProjectsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListProjectsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListProjectsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListProjectsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListProjectsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListProjectsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListProjectsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListProjectsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4496,
  serialized_end=4707,
)


_LISTPROJECTSRESPONSE = _descriptor.Descriptor(
  name='ListProjectsResponse',
  full_name='continual.rpc.management.v1.ListProjectsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='projects', full_name='continual.rpc.management.v1.ListProjectsResponse.projects', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListProjectsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4709,
  serialized_end=4812,
)


_TESTPROJECTDATASTOREREQUEST = _descriptor.Descriptor(
  name='TestProjectDataStoreRequest',
  full_name='continual.rpc.management.v1.TestProjectDataStoreRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.TestProjectDataStoreRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4814,
  serialized_end=4857,
)


_TESTPROJECTDATASTORERESPONSE = _descriptor.Descriptor(
  name='TestProjectDataStoreResponse',
  full_name='continual.rpc.management.v1.TestProjectDataStoreResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.TestProjectDataStoreResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.TestProjectDataStoreResponse.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4859,
  serialized_end=4929,
)


_TESTDATASTOREREQUEST = _descriptor.Descriptor(
  name='TestDataStoreRequest',
  full_name='continual.rpc.management.v1.TestDataStoreRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_store', full_name='continual.rpc.management.v1.TestDataStoreRequest.data_store', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4931,
  serialized_end=5013,
)


_TESTDATASTORERESPONSE = _descriptor.Descriptor(
  name='TestDataStoreResponse',
  full_name='continual.rpc.management.v1.TestDataStoreResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.TestDataStoreResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.TestDataStoreResponse.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5015,
  serialized_end=5078,
)


_GETFEATURESETREQUEST = _descriptor.Descriptor(
  name='GetFeatureSetRequest',
  full_name='continual.rpc.management.v1.GetFeatureSetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetFeatureSetRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5080,
  serialized_end=5116,
)


_CREATECHANGEREQUEST = _descriptor.Descriptor(
  name='CreateChangeRequest',
  full_name='continual.rpc.management.v1.CreateChangeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateChangeRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schemas', full_name='continual.rpc.management.v1.CreateChangeRequest.schemas', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plan_only', full_name='continual.rpc.management.v1.CreateChangeRequest.plan_only', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='purge', full_name='continual.rpc.management.v1.CreateChangeRequest.purge', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_all', full_name='continual.rpc.management.v1.CreateChangeRequest.update_all', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='train_none', full_name='continual.rpc.management.v1.CreateChangeRequest.train_none', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='train_all', full_name='continual.rpc.management.v1.CreateChangeRequest.train_all', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='profile_none', full_name='continual.rpc.management.v1.CreateChangeRequest.profile_none', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.CreateChangeRequest.message', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trigger', full_name='continual.rpc.management.v1.CreateChangeRequest.trigger', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.CreateChangeRequest.resource', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5119,
  serialized_end=5340,
)


_RERUNREQUEST = _descriptor.Descriptor(
  name='ReRunRequest',
  full_name='continual.rpc.management.v1.ReRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ReRunRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.ReRunRequest.message', index=1,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trigger', full_name='continual.rpc.management.v1.ReRunRequest.trigger', index=2,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5342,
  serialized_end=5404,
)


_GETCHANGEREQUEST = _descriptor.Descriptor(
  name='GetChangeRequest',
  full_name='continual.rpc.management.v1.GetChangeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetChangeRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5406,
  serialized_end=5438,
)


_GETCHANGERESPONSE = _descriptor.Descriptor(
  name='GetChangeResponse',
  full_name='continual.rpc.management.v1.GetChangeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetChangeResponse.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plan', full_name='continual.rpc.management.v1.GetChangeResponse.plan', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5440,
  serialized_end=5526,
)


_LISTCHANGESREQUEST = _descriptor.Descriptor(
  name='ListChangesRequest',
  full_name='continual.rpc.management.v1.ListChangesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListChangesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListChangesRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListChangesRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListChangesRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListChangesRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListChangesRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListChangesRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListChangesRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListChangesRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5529,
  serialized_end=5777,
)


_LISTCHANGESRESPONSE = _descriptor.Descriptor(
  name='ListChangesResponse',
  full_name='continual.rpc.management.v1.ListChangesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='plans', full_name='continual.rpc.management.v1.ListChangesResponse.plans', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListChangesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5779,
  serialized_end=5877,
)


_CANCELCHANGEREQUEST = _descriptor.Descriptor(
  name='CancelChangeRequest',
  full_name='continual.rpc.management.v1.CancelChangeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.CancelChangeRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5879,
  serialized_end=5914,
)


_LISTFEATURESETSREQUEST = _descriptor.Descriptor(
  name='ListFeatureSetsRequest',
  full_name='continual.rpc.management.v1.ListFeatureSetsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListFeatureSetsRequest.all_projects', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5917,
  serialized_end=6153,
)


_LISTFEATURESETSRESPONSE = _descriptor.Descriptor(
  name='ListFeatureSetsResponse',
  full_name='continual.rpc.management.v1.ListFeatureSetsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_sets', full_name='continual.rpc.management.v1.ListFeatureSetsResponse.feature_sets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListFeatureSetsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6155,
  serialized_end=6268,
)


_GETENTITYFEATURESETSREQUEST = _descriptor.Descriptor(
  name='GetEntityFeatureSetsRequest',
  full_name='continual.rpc.management.v1.GetEntityFeatureSetsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project_name', full_name='continual.rpc.management.v1.GetEntityFeatureSetsRequest.project_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity_id', full_name='continual.rpc.management.v1.GetEntityFeatureSetsRequest.entity_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6270,
  serialized_end=6340,
)


_GETENTITYFEATURESETSRESPONSE = _descriptor.Descriptor(
  name='GetEntityFeatureSetsResponse',
  full_name='continual.rpc.management.v1.GetEntityFeatureSetsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_sets', full_name='continual.rpc.management.v1.GetEntityFeatureSetsResponse.feature_sets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6342,
  serialized_end=6435,
)


_GETENTITYREQUEST = _descriptor.Descriptor(
  name='GetEntityRequest',
  full_name='continual.rpc.management.v1.GetEntityRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetEntityRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6437,
  serialized_end=6469,
)


_LISTENTITIESREQUEST = _descriptor.Descriptor(
  name='ListEntitiesRequest',
  full_name='continual.rpc.management.v1.ListEntitiesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListEntitiesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6471,
  serialized_end=6508,
)


_LISTENTITIESRESPONSE = _descriptor.Descriptor(
  name='ListEntitiesResponse',
  full_name='continual.rpc.management.v1.ListEntitiesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entities', full_name='continual.rpc.management.v1.ListEntitiesResponse.entities', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6510,
  serialized_end=6587,
)


_BATCHGETFEATURESETSREQUEST = _descriptor.Descriptor(
  name='BatchGetFeatureSetsRequest',
  full_name='continual.rpc.management.v1.BatchGetFeatureSetsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetFeatureSetsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetFeatureSetsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6589,
  serialized_end=6648,
)


_BATCHGETFEATURESETSRESPONSE = _descriptor.Descriptor(
  name='BatchGetFeatureSetsResponse',
  full_name='continual.rpc.management.v1.BatchGetFeatureSetsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_sets', full_name='continual.rpc.management.v1.BatchGetFeatureSetsResponse.feature_sets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6650,
  serialized_end=6742,
)


_DELETEFEATURESETREQUEST = _descriptor.Descriptor(
  name='DeleteFeatureSetRequest',
  full_name='continual.rpc.management.v1.DeleteFeatureSetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteFeatureSetRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6744,
  serialized_end=6783,
)


_VALIDATEFEATURESETREQUEST = _descriptor.Descriptor(
  name='ValidateFeatureSetRequest',
  full_name='continual.rpc.management.v1.ValidateFeatureSetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environment', full_name='continual.rpc.management.v1.ValidateFeatureSetRequest.environment', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='continual.rpc.management.v1.ValidateFeatureSetRequest.schema', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6785,
  serialized_end=6849,
)


_GETFEATURESETDATASETSTATSREQUEST = _descriptor.Descriptor(
  name='GetFeatureSetDatasetStatsRequest',
  full_name='continual.rpc.management.v1.GetFeatureSetDatasetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetFeatureSetDatasetStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6851,
  serialized_end=6899,
)


_GETMODELDATASETSTATSREQUEST = _descriptor.Descriptor(
  name='GetModelDatasetStatsRequest',
  full_name='continual.rpc.management.v1.GetModelDatasetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelDatasetStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6901,
  serialized_end=6944,
)


_GETMODELVERSIONDATASETSTATSREQUEST = _descriptor.Descriptor(
  name='GetModelVersionDatasetStatsRequest',
  full_name='continual.rpc.management.v1.GetModelVersionDatasetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelVersionDatasetStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6946,
  serialized_end=6996,
)


_GETBATCHPREDICTIONDATASETSTATSREQUEST = _descriptor.Descriptor(
  name='GetBatchPredictionDatasetStatsRequest',
  full_name='continual.rpc.management.v1.GetBatchPredictionDatasetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetBatchPredictionDatasetStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6998,
  serialized_end=7051,
)


_GETPREDICTIONSTATSREQUEST = _descriptor.Descriptor(
  name='GetPredictionStatsRequest',
  full_name='continual.rpc.management.v1.GetPredictionStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPredictionStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7053,
  serialized_end=7094,
)


_GETFEATURESETSTATSREQUEST = _descriptor.Descriptor(
  name='GetFeatureSetStatsRequest',
  full_name='continual.rpc.management.v1.GetFeatureSetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetFeatureSetStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7096,
  serialized_end=7137,
)


_GETMODELSTATSREQUEST = _descriptor.Descriptor(
  name='GetModelStatsRequest',
  full_name='continual.rpc.management.v1.GetModelStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7139,
  serialized_end=7175,
)


_COMPUTECOLUMNSTATSREQUEST = _descriptor.Descriptor(
  name='ComputeColumnStatsRequest',
  full_name='continual.rpc.management.v1.ComputeColumnStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ComputeColumnStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7177,
  serialized_end=7218,
)


_GETDATAREQUEST = _descriptor.Descriptor(
  name='GetDataRequest',
  full_name='continual.rpc.management.v1.GetDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetDataRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.GetDataRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.GetDataRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7220,
  serialized_end=7289,
)


_GETDATARESPONSE = _descriptor.Descriptor(
  name='GetDataResponse',
  full_name='continual.rpc.management.v1.GetDataResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='record_count', full_name='continual.rpc.management.v1.GetDataResponse.record_count', index=0,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.GetDataResponse.next_page_token', index=1,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.GetDataResponse.value', index=2,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7291,
  serialized_end=7370,
)


_GETUSAGEREQUEST = _descriptor.Descriptor(
  name='GetUsageRequest',
  full_name='continual.rpc.management.v1.GetUsageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.GetUsageRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.GetUsageRequest.start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.GetUsageRequest.end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7372,
  serialized_end=7499,
)


_GETUSAGERESPONSE = _descriptor.Descriptor(
  name='GetUsageResponse',
  full_name='continual.rpc.management.v1.GetUsageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='credits_per_day', full_name='continual.rpc.management.v1.GetUsageResponse.credits_per_day', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7501,
  serialized_end=7592,
)


_LISTPRICESREQUEST = _descriptor.Descriptor(
  name='ListPricesRequest',
  full_name='continual.rpc.management.v1.ListPricesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListPricesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7594,
  serialized_end=7629,
)


_LISTPRICESRESPONSE = _descriptor.Descriptor(
  name='ListPricesResponse',
  full_name='continual.rpc.management.v1.ListPricesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='prices', full_name='continual.rpc.management.v1.ListPricesResponse.prices', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7631,
  serialized_end=7703,
)


_LISTINVOICESREQUEST = _descriptor.Descriptor(
  name='ListInvoicesRequest',
  full_name='continual.rpc.management.v1.ListInvoicesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListInvoicesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListInvoicesRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListInvoicesRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7705,
  serialized_end=7781,
)


_LISTINVOICESRESPONSE = _descriptor.Descriptor(
  name='ListInvoicesResponse',
  full_name='continual.rpc.management.v1.ListInvoicesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invoices', full_name='continual.rpc.management.v1.ListInvoicesResponse.invoices', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListInvoicesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7783,
  serialized_end=7886,
)


_GETPRICEREQUEST = _descriptor.Descriptor(
  name='GetPriceRequest',
  full_name='continual.rpc.management.v1.GetPriceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPriceRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7888,
  serialized_end=7919,
)


_GETPAYMENTMETHODREQUEST = _descriptor.Descriptor(
  name='GetPaymentMethodRequest',
  full_name='continual.rpc.management.v1.GetPaymentMethodRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPaymentMethodRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7921,
  serialized_end=7960,
)


_GETUPCOMINGINVOICEREQUEST = _descriptor.Descriptor(
  name='GetUpcomingInvoiceRequest',
  full_name='continual.rpc.management.v1.GetUpcomingInvoiceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetUpcomingInvoiceRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7962,
  serialized_end=8003,
)


_GETSUBSCRIPTIONREQUEST = _descriptor.Descriptor(
  name='GetSubscriptionRequest',
  full_name='continual.rpc.management.v1.GetSubscriptionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetSubscriptionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8005,
  serialized_end=8043,
)


_GETBILLINGACCOUNTREQUEST = _descriptor.Descriptor(
  name='GetBillingAccountRequest',
  full_name='continual.rpc.management.v1.GetBillingAccountRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetBillingAccountRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8045,
  serialized_end=8085,
)


_UPDATEBILLINGACCOUNTREQUEST = _descriptor.Descriptor(
  name='UpdateBillingAccountRequest',
  full_name='continual.rpc.management.v1.UpdateBillingAccountRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='billing_account', full_name='continual.rpc.management.v1.UpdateBillingAccountRequest.billing_account', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateBillingAccountRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8088,
  serialized_end=8236,
)


_UPDATESUBSCRIPTIONREQUEST = _descriptor.Descriptor(
  name='UpdateSubscriptionRequest',
  full_name='continual.rpc.management.v1.UpdateSubscriptionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='subscription', full_name='continual.rpc.management.v1.UpdateSubscriptionRequest.subscription', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateSubscriptionRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8239,
  serialized_end=8380,
)


_CREATEFEATURESETCSVSIGNEDURLREQUEST = _descriptor.Descriptor(
  name='CreateFeatureSetCSVSignedURLRequest',
  full_name='continual.rpc.management.v1.CreateFeatureSetCSVSignedURLRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateFeatureSetCSVSignedURLRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='continual.rpc.management.v1.CreateFeatureSetCSVSignedURLRequest.file_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8382,
  serialized_end=8454,
)


_CREATEFEATURESETCSVSIGNEDURLRESPONSE = _descriptor.Descriptor(
  name='CreateFeatureSetCSVSignedURLResponse',
  full_name='continual.rpc.management.v1.CreateFeatureSetCSVSignedURLResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signed_url', full_name='continual.rpc.management.v1.CreateFeatureSetCSVSignedURLResponse.signed_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bucket_path', full_name='continual.rpc.management.v1.CreateFeatureSetCSVSignedURLResponse.bucket_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8456,
  serialized_end=8535,
)


_PROMOTEMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='PromoteModelVersionRequest',
  full_name='continual.rpc.management.v1.PromoteModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.PromoteModelVersionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8537,
  serialized_end=8579,
)


_LISTPROMOTIONSREQUEST = _descriptor.Descriptor(
  name='ListPromotionsRequest',
  full_name='continual.rpc.management.v1.ListPromotionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListPromotionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListPromotionsRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListPromotionsRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active', full_name='continual.rpc.management.v1.ListPromotionsRequest.active', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListPromotionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListPromotionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListPromotionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListPromotionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListPromotionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8582,
  serialized_end=8833,
)


_LISTPROMOTIONSRESPONSE = _descriptor.Descriptor(
  name='ListPromotionsResponse',
  full_name='continual.rpc.management.v1.ListPromotionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='promotions', full_name='continual.rpc.management.v1.ListPromotionsResponse.promotions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListPromotionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8835,
  serialized_end=8944,
)


_GETPROMOTIONREQUEST = _descriptor.Descriptor(
  name='GetPromotionRequest',
  full_name='continual.rpc.management.v1.GetPromotionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPromotionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8946,
  serialized_end=8981,
)


_BATCHGETPROMOTIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetPromotionsRequest',
  full_name='continual.rpc.management.v1.BatchGetPromotionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetPromotionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetPromotionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8983,
  serialized_end=9041,
)


_BATCHGETPROMOTIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetPromotionsResponse',
  full_name='continual.rpc.management.v1.BatchGetPromotionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='promotions', full_name='continual.rpc.management.v1.BatchGetPromotionsResponse.promotions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9043,
  serialized_end=9131,
)


_LISTEVENTSREQUEST = _descriptor.Descriptor(
  name='ListEventsRequest',
  full_name='continual.rpc.management.v1.ListEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListEventsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListEventsRequest.filters', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListEventsRequest.ignore_filters', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListEventsRequest.start_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListEventsRequest.end_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListEventsRequest.page_token', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListEventsRequest.page_size', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListEventsRequest.all_projects', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9134,
  serialized_end=9365,
)


_LISTEVENTSRESPONSE = _descriptor.Descriptor(
  name='ListEventsResponse',
  full_name='continual.rpc.management.v1.ListEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='continual.rpc.management.v1.ListEventsResponse.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListEventsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9367,
  serialized_end=9464,
)


_GETEVENTREQUEST = _descriptor.Descriptor(
  name='GetEventRequest',
  full_name='continual.rpc.management.v1.GetEventRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetEventRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9466,
  serialized_end=9497,
)


_BATCHGETEVENTSREQUEST = _descriptor.Descriptor(
  name='BatchGetEventsRequest',
  full_name='continual.rpc.management.v1.BatchGetEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetEventsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetEventsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9499,
  serialized_end=9553,
)


_BATCHGETEVENTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetEventsResponse',
  full_name='continual.rpc.management.v1.BatchGetEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='continual.rpc.management.v1.BatchGetEventsResponse.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9555,
  serialized_end=9631,
)


_GETINFERREDSCHEMAREQUEST = _descriptor.Descriptor(
  name='GetInferredSchemaRequest',
  full_name='continual.rpc.management.v1.GetInferredSchemaRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.GetInferredSchemaRequest.project', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file', full_name='continual.rpc.management.v1.GetInferredSchemaRequest.file', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.GetInferredSchemaRequest.query', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.GetInferredSchemaRequest.table', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_row_count', full_name='continual.rpc.management.v1.GetInferredSchemaRequest.sample_row_count', index=4,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9633,
  serialized_end=9746,
)


_GETINFERREDSCHEMARESPONSE = _descriptor.Descriptor(
  name='GetInferredSchemaResponse',
  full_name='continual.rpc.management.v1.GetInferredSchemaResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='columns', full_name='continual.rpc.management.v1.GetInferredSchemaResponse.columns', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='record_count', full_name='continual.rpc.management.v1.GetInferredSchemaResponse.record_count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='continual.rpc.management.v1.GetInferredSchemaResponse.values', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9748,
  serialized_end=9873,
)


_GETMODELREQUEST = _descriptor.Descriptor(
  name='GetModelRequest',
  full_name='continual.rpc.management.v1.GetModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9875,
  serialized_end=9906,
)


_BATCHGETMODELSREQUEST = _descriptor.Descriptor(
  name='BatchGetModelsRequest',
  full_name='continual.rpc.management.v1.BatchGetModelsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetModelsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetModelsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9908,
  serialized_end=9962,
)


_BATCHGETMODELSRESPONSE = _descriptor.Descriptor(
  name='BatchGetModelsResponse',
  full_name='continual.rpc.management.v1.BatchGetModelsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='models', full_name='continual.rpc.management.v1.BatchGetModelsResponse.models', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9964,
  serialized_end=10040,
)


_LISTMODELSREQUEST = _descriptor.Descriptor(
  name='ListModelsRequest',
  full_name='continual.rpc.management.v1.ListModelsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListModelsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListModelsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListModelsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='show_deleted', full_name='continual.rpc.management.v1.ListModelsRequest.show_deleted', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListModelsRequest.latest', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListModelsRequest.filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListModelsRequest.ignore_filters', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListModelsRequest.start_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListModelsRequest.end_time', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListModelsRequest.all_projects', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10043,
  serialized_end=10312,
)


_LISTMODELSRESPONSE = _descriptor.Descriptor(
  name='ListModelsResponse',
  full_name='continual.rpc.management.v1.ListModelsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='models', full_name='continual.rpc.management.v1.ListModelsResponse.models', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListModelsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10314,
  serialized_end=10411,
)


_DELETEMODELREQUEST = _descriptor.Descriptor(
  name='DeleteModelRequest',
  full_name='continual.rpc.management.v1.DeleteModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteModelRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10413,
  serialized_end=10447,
)


_VALIDATEMODELREQUEST = _descriptor.Descriptor(
  name='ValidateModelRequest',
  full_name='continual.rpc.management.v1.ValidateModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environment', full_name='continual.rpc.management.v1.ValidateModelRequest.environment', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='continual.rpc.management.v1.ValidateModelRequest.schema', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10449,
  serialized_end=10508,
)


_GETMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='GetModelVersionRequest',
  full_name='continual.rpc.management.v1.GetModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelVersionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10510,
  serialized_end=10548,
)


_GETMODELVERSIONQUERYREQUEST = _descriptor.Descriptor(
  name='GetModelVersionQueryRequest',
  full_name='continual.rpc.management.v1.GetModelVersionQueryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelVersionQueryRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10550,
  serialized_end=10593,
)


_GETMODELVERSIONQUERYRESPONSE = _descriptor.Descriptor(
  name='GetModelVersionQueryResponse',
  full_name='continual.rpc.management.v1.GetModelVersionQueryResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.GetModelVersionQueryResponse.query', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10595,
  serialized_end=10676,
)


_BATCHGETMODELVERSIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetModelVersionsRequest',
  full_name='continual.rpc.management.v1.BatchGetModelVersionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetModelVersionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetModelVersionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10678,
  serialized_end=10739,
)


_BATCHGETMODELVERSIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetModelVersionsResponse',
  full_name='continual.rpc.management.v1.BatchGetModelVersionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_versions', full_name='continual.rpc.management.v1.BatchGetModelVersionsResponse.model_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10741,
  serialized_end=10839,
)


_LISTMODELVERSIONSREQUEST = _descriptor.Descriptor(
  name='ListModelVersionsRequest',
  full_name='continual.rpc.management.v1.ListModelVersionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListModelVersionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListModelVersionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListModelVersionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListModelVersionsRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListModelVersionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListModelVersionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListModelVersionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListModelVersionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListModelVersionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10842,
  serialized_end=11096,
)


_LISTMODELVERSIONSRESPONSE = _descriptor.Descriptor(
  name='ListModelVersionsResponse',
  full_name='continual.rpc.management.v1.ListModelVersionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_versions', full_name='continual.rpc.management.v1.ListModelVersionsResponse.model_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListModelVersionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11098,
  serialized_end=11217,
)


_TRAINMODELREQUEST = _descriptor.Descriptor(
  name='TrainModelRequest',
  full_name='continual.rpc.management.v1.TrainModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.TrainModelRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='system_time', full_name='continual.rpc.management.v1.TrainModelRequest.system_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visible_start_time', full_name='continual.rpc.management.v1.TrainModelRequest.visible_start_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visible_stop_time', full_name='continual.rpc.management.v1.TrainModelRequest.visible_stop_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11220,
  serialized_end=11413,
)


_CANCELTRAININGREQUEST = _descriptor.Descriptor(
  name='CancelTrainingRequest',
  full_name='continual.rpc.management.v1.CancelTrainingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.CancelTrainingRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11415,
  serialized_end=11452,
)


_GETTRAININGMETRICSREQUEST = _descriptor.Descriptor(
  name='GetTrainingMetricsRequest',
  full_name='continual.rpc.management.v1.GetTrainingMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment', full_name='continual.rpc.management.v1.GetTrainingMetricsRequest.experiment', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11454,
  serialized_end=11501,
)


_GETTRAININGMETRICSRESPONSE = _descriptor.Descriptor(
  name='GetTrainingMetricsResponse',
  full_name='continual.rpc.management.v1.GetTrainingMetricsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='training_metrics', full_name='continual.rpc.management.v1.GetTrainingMetricsResponse.training_metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11503,
  serialized_end=11603,
)


_LISTTRAININGMETRICSREQUEST = _descriptor.Descriptor(
  name='ListTrainingMetricsRequest',
  full_name='continual.rpc.management.v1.ListTrainingMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.ListTrainingMetricsRequest.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListTrainingMetricsRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListTrainingMetricsRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11605,
  serialized_end=11687,
)


_LISTTRAININGMETRICSRESPONSE = _descriptor.Descriptor(
  name='ListTrainingMetricsResponse',
  full_name='continual.rpc.management.v1.ListTrainingMetricsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='training_metrics', full_name='continual.rpc.management.v1.ListTrainingMetricsResponse.training_metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListTrainingMetricsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11689,
  serialized_end=11815,
)


_GETEXPERIMENTREQUEST = _descriptor.Descriptor(
  name='GetExperimentRequest',
  full_name='continual.rpc.management.v1.GetExperimentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetExperimentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11817,
  serialized_end=11853,
)


_BATCHGETEXPERIMENTSREQUEST = _descriptor.Descriptor(
  name='BatchGetExperimentsRequest',
  full_name='continual.rpc.management.v1.BatchGetExperimentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetExperimentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetExperimentsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11855,
  serialized_end=11914,
)


_BATCHGETEXPERIMENTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetExperimentsResponse',
  full_name='continual.rpc.management.v1.BatchGetExperimentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiments', full_name='continual.rpc.management.v1.BatchGetExperimentsResponse.experiments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11916,
  serialized_end=12007,
)


_LISTEXPERIMENTSREQUEST = _descriptor.Descriptor(
  name='ListExperimentsRequest',
  full_name='continual.rpc.management.v1.ListExperimentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListExperimentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListExperimentsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListExperimentsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListExperimentsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListExperimentsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListExperimentsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListExperimentsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListExperimentsRequest.all_projects', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12010,
  serialized_end=12246,
)


_LISTEXPERIMENTSRESPONSE = _descriptor.Descriptor(
  name='ListExperimentsResponse',
  full_name='continual.rpc.management.v1.ListExperimentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiments', full_name='continual.rpc.management.v1.ListExperimentsResponse.experiments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListExperimentsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12248,
  serialized_end=12360,
)


_CREATEBATCHPREDICTIONREQUEST = _descriptor.Descriptor(
  name='CreateBatchPredictionRequest',
  full_name='continual.rpc.management.v1.CreateBatchPredictionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_type', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.source_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.source', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dest_type', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.dest_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='incremental', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.incremental', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_model', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.prediction_model', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12363,
  serialized_end=12621,
)


_LISTBATCHPREDICTIONSREQUEST = _descriptor.Descriptor(
  name='ListBatchPredictionsRequest',
  full_name='continual.rpc.management.v1.ListBatchPredictionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12624,
  serialized_end=12881,
)


_LISTBATCHPREDICTIONSRESPONSE = _descriptor.Descriptor(
  name='ListBatchPredictionsResponse',
  full_name='continual.rpc.management.v1.ListBatchPredictionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_predictions', full_name='continual.rpc.management.v1.ListBatchPredictionsResponse.batch_predictions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListBatchPredictionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12884,
  serialized_end=13012,
)


_BATCHPREDICTIONREQUEST = _descriptor.Descriptor(
  name='BatchPredictionRequest',
  full_name='continual.rpc.management.v1.BatchPredictionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BatchPredictionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13014,
  serialized_end=13052,
)


_BATCHGETBATCHPREDICTIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetBatchPredictionsRequest',
  full_name='continual.rpc.management.v1.BatchGetBatchPredictionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetBatchPredictionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetBatchPredictionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13054,
  serialized_end=13118,
)


_BATCHGETBATCHPREDICTIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetBatchPredictionsResponse',
  full_name='continual.rpc.management.v1.BatchGetBatchPredictionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_predictions', full_name='continual.rpc.management.v1.BatchGetBatchPredictionsResponse.batch_predictions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13120,
  serialized_end=13227,
)


_GETSERVINGDATAREQUEST = _descriptor.Descriptor(
  name='GetServingDataRequest',
  full_name='continual.rpc.management.v1.GetServingDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.GetServingDataRequest.query', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='overrides', full_name='continual.rpc.management.v1.GetServingDataRequest.overrides', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_time', full_name='continual.rpc.management.v1.GetServingDataRequest.event_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13230,
  serialized_end=13371,
)


_GETSERVINGDATARESPONSE = _descriptor.Descriptor(
  name='GetServingDataResponse',
  full_name='continual.rpc.management.v1.GetServingDataResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.GetServingDataResponse.value', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13373,
  serialized_end=13412,
)


_GETTRAININGDATAREQUEST = _descriptor.Descriptor(
  name='GetTrainingDataRequest',
  full_name='continual.rpc.management.v1.GetTrainingDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='continual.rpc.management.v1.GetTrainingDataRequest.query', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13414,
  serialized_end=13489,
)


_GETTRAININGDATARESPONSE = _descriptor.Descriptor(
  name='GetTrainingDataResponse',
  full_name='continual.rpc.management.v1.GetTrainingDataResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.GetTrainingDataResponse.value', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13491,
  serialized_end=13531,
)


_BROWSEDATAWAREHOUSEREQUEST = _descriptor.Descriptor(
  name='BrowseDataWarehouseRequest',
  full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='object_type', full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest.object_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest.project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest.database', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest.schema', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.BrowseDataWarehouseRequest.table', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BROWSEDATAWAREHOUSEREQUEST_OBJECTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13534,
  serialized_end=13780,
)


_BROWSEDATAWAREHOUSERESPONSE_TYPESENTRY = _descriptor.Descriptor(
  name='TypesEntry',
  full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse.TypesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse.TypesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse.TypesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14002,
  serialized_end=14046,
)

_BROWSEDATAWAREHOUSERESPONSE = _descriptor.Descriptor(
  name='BrowseDataWarehouseResponse',
  full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='object_type', full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse.object_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='types', full_name='continual.rpc.management.v1.BrowseDataWarehouseResponse.types', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_BROWSEDATAWAREHOUSERESPONSE_TYPESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13783,
  serialized_end=14046,
)


_SEEDREQUEST = _descriptor.Descriptor(
  name='SeedRequest',
  full_name='continual.rpc.management.v1.SeedRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.SeedRequest.project', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_path', full_name='continual.rpc.management.v1.SeedRequest.file_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table_name', full_name='continual.rpc.management.v1.SeedRequest.table_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14048,
  serialized_end=14117,
)


_SEEDRESPONSE = _descriptor.Descriptor(
  name='SeedResponse',
  full_name='continual.rpc.management.v1.SeedResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema_name', full_name='continual.rpc.management.v1.SeedResponse.schema_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table_name', full_name='continual.rpc.management.v1.SeedResponse.table_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14119,
  serialized_end=14174,
)


_YAMLDIFFREQUEST = _descriptor.Descriptor(
  name='YamlDiffRequest',
  full_name='continual.rpc.management.v1.YamlDiffRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='right', full_name='continual.rpc.management.v1.YamlDiffRequest.right', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='right_text', full_name='continual.rpc.management.v1.YamlDiffRequest.right_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='left', full_name='continual.rpc.management.v1.YamlDiffRequest.left', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='left_text', full_name='continual.rpc.management.v1.YamlDiffRequest.left_text', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.YamlDiffRequest.parent', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14176,
  serialized_end=14277,
)


_YAMLDIFFRESPONSE = _descriptor.Descriptor(
  name='YamlDiffResponse',
  full_name='continual.rpc.management.v1.YamlDiffResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='continual.rpc.management.v1.YamlDiffResponse.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14279,
  serialized_end=14311,
)


_GETMODELQUERYREQUEST = _descriptor.Descriptor(
  name='GetModelQueryRequest',
  full_name='continual.rpc.management.v1.GetModelQueryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelQueryRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.GetModelQueryRequest.model', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='additional_feature_sets', full_name='continual.rpc.management.v1.GetModelQueryRequest.additional_feature_sets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14314,
  serialized_end=14475,
)


_GETMODELQUERYRESPONSE = _descriptor.Descriptor(
  name='GetModelQueryResponse',
  full_name='continual.rpc.management.v1.GetModelQueryResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query_json', full_name='continual.rpc.management.v1.GetModelQueryResponse.query_json', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14477,
  serialized_end=14520,
)


_GETSCHEMAGRAPHREQUEST = _descriptor.Descriptor(
  name='GetSchemaGraphRequest',
  full_name='continual.rpc.management.v1.GetSchemaGraphRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetSchemaGraphRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='additional_feature_sets', full_name='continual.rpc.management.v1.GetSchemaGraphRequest.additional_feature_sets', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='additional_models', full_name='continual.rpc.management.v1.GetSchemaGraphRequest.additional_models', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14523,
  serialized_end=14697,
)


_GETSCHEMAGRAPHRESPONSE = _descriptor.Descriptor(
  name='GetSchemaGraphResponse',
  full_name='continual.rpc.management.v1.GetSchemaGraphResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='graph_json', full_name='continual.rpc.management.v1.GetSchemaGraphResponse.graph_json', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14699,
  serialized_end=14743,
)


_CREATEENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='CreateEnvironmentRequest',
  full_name='continual.rpc.management.v1.CreateEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='environment_id', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.environment_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.source', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_store', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.data_store', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheduling_enabled', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.scheduling_enabled', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14746,
  serialized_end=14916,
)


_UPDATEENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='UpdateEnvironmentRequest',
  full_name='continual.rpc.management.v1.UpdateEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environment', full_name='continual.rpc.management.v1.UpdateEnvironmentRequest.environment', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateEnvironmentRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14919,
  serialized_end=15057,
)


_GETENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='GetEnvironmentRequest',
  full_name='continual.rpc.management.v1.GetEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetEnvironmentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15059,
  serialized_end=15096,
)


_BATCHGETENVIRONMENTSREQUEST = _descriptor.Descriptor(
  name='BatchGetEnvironmentsRequest',
  full_name='continual.rpc.management.v1.BatchGetEnvironmentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetEnvironmentsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15098,
  serialized_end=15142,
)


_BATCHGETENVIRONMENTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetEnvironmentsResponse',
  full_name='continual.rpc.management.v1.BatchGetEnvironmentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environments', full_name='continual.rpc.management.v1.BatchGetEnvironmentsResponse.environments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15144,
  serialized_end=15238,
)


_DELETEENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='DeleteEnvironmentRequest',
  full_name='continual.rpc.management.v1.DeleteEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteEnvironmentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delete_schema', full_name='continual.rpc.management.v1.DeleteEnvironmentRequest.delete_schema', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15240,
  serialized_end=15303,
)


_LISTENVIRONMENTSREQUEST = _descriptor.Descriptor(
  name='ListEnvironmentsRequest',
  full_name='continual.rpc.management.v1.ListEnvironmentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15306,
  serialized_end=15521,
)


_LISTENVIRONMENTSRESPONSE = _descriptor.Descriptor(
  name='ListEnvironmentsResponse',
  full_name='continual.rpc.management.v1.ListEnvironmentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environments', full_name='continual.rpc.management.v1.ListEnvironmentsResponse.environments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListEnvironmentsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15523,
  serialized_end=15638,
)


_TESTENVIRONMENTDATASTOREREQUEST = _descriptor.Descriptor(
  name='TestEnvironmentDataStoreRequest',
  full_name='continual.rpc.management.v1.TestEnvironmentDataStoreRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.TestEnvironmentDataStoreRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15640,
  serialized_end=15687,
)


_TESTENVIRONMENTDATASTORERESPONSE = _descriptor.Descriptor(
  name='TestEnvironmentDataStoreResponse',
  full_name='continual.rpc.management.v1.TestEnvironmentDataStoreResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.TestEnvironmentDataStoreResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.TestEnvironmentDataStoreResponse.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15689,
  serialized_end=15763,
)


_GETSYSTEMMESSAGEREQUEST = _descriptor.Descriptor(
  name='GetSystemMessageRequest',
  full_name='continual.rpc.management.v1.GetSystemMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15765,
  serialized_end=15790,
)


_GETSYSTEMMESSAGERESPONSE = _descriptor.Descriptor(
  name='GetSystemMessageResponse',
  full_name='continual.rpc.management.v1.GetSystemMessageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.GetSystemMessageResponse.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15792,
  serialized_end=15835,
)

_UPDATEUSERREQUEST.fields_by_name['user'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_UPDATEUSERREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETUSERSRESPONSE.fields_by_name['users'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_LISTUSERSRESPONSE.fields_by_name['users'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_REGISTERRESPONSE.fields_by_name['user'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_LOGINRESPONSE.fields_by_name['user'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_LISTSESSIONSRESPONSE.fields_by_name['sessions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SESSION
_LISTROLESRESPONSE.fields_by_name['roles'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ROLE
_BATCHGETROLESRESPONSE.fields_by_name['roles'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ROLE
_LISTACCESSPOLICIESRESPONSE.fields_by_name['access_policies'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY
_CREATEACCESSPOLICYREQUEST.fields_by_name['access_policy'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY
_UPDATEACCESSPOLICYREQUEST.fields_by_name['access_policy'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY
_UPDATEACCESSPOLICYREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEINVITEREQUEST.fields_by_name['invite'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE
_LISTINVITESRESPONSE.fields_by_name['invites'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE
_UPDATEINVITEREQUEST.fields_by_name['invite'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE
_UPDATEINVITEREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEORGANIZATIONREQUEST.fields_by_name['organization'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_UPDATEORGANIZATIONREQUEST.fields_by_name['organization'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_UPDATEORGANIZATIONREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETORGANIZATIONSRESPONSE.fields_by_name['organizations'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_LISTORGANIZATIONSRESPONSE.fields_by_name['organizations'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_UPDATEONBOARDINGREQUEST.fields_by_name['onboarding'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ONBOARDING
_UPDATEONBOARDINGREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEPROJECTREQUEST.fields_by_name['project'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_UPDATEPROJECTREQUEST.fields_by_name['project'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_UPDATEPROJECTREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETPROJECTSRESPONSE.fields_by_name['projects'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_LISTPROJECTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROJECTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROJECTSRESPONSE.fields_by_name['projects'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_TESTDATASTOREREQUEST.fields_by_name['data_store'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASTORE
_GETCHANGERESPONSE.fields_by_name['plan'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PLANSTEP
_LISTCHANGESREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTCHANGESREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTCHANGESRESPONSE.fields_by_name['plans'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._CHANGE
_LISTFEATURESETSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTFEATURESETSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTFEATURESETSRESPONSE.fields_by_name['feature_sets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._FEATURESET
_GETENTITYFEATURESETSRESPONSE.fields_by_name['feature_sets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._FEATURESET
_LISTENTITIESRESPONSE.fields_by_name['entities'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENTITY
_BATCHGETFEATURESETSRESPONSE.fields_by_name['feature_sets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._FEATURESET
_GETUSAGEREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GETUSAGEREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GETUSAGERESPONSE.fields_by_name['credits_per_day'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._CREDITUSAGERECORD
_LISTPRICESRESPONSE.fields_by_name['prices'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PRICE
_LISTINVOICESRESPONSE.fields_by_name['invoices'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVOICE
_UPDATEBILLINGACCOUNTREQUEST.fields_by_name['billing_account'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BILLINGACCOUNT
_UPDATEBILLINGACCOUNTREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_UPDATESUBSCRIPTIONREQUEST.fields_by_name['subscription'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SUBSCRIPTION
_UPDATESUBSCRIPTIONREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_LISTPROMOTIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROMOTIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROMOTIONSRESPONSE.fields_by_name['promotions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION
_BATCHGETPROMOTIONSRESPONSE.fields_by_name['promotions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION
_LISTEVENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEVENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEVENTSRESPONSE.fields_by_name['events'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT
_BATCHGETEVENTSRESPONSE.fields_by_name['events'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT
_GETINFERREDSCHEMARESPONSE.fields_by_name['columns'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._COLUMNCONFIG
_BATCHGETMODELSRESPONSE.fields_by_name['models'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL
_LISTMODELSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELSRESPONSE.fields_by_name['models'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL
_GETMODELVERSIONQUERYRESPONSE.fields_by_name['query'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._QUERY
_BATCHGETMODELVERSIONSRESPONSE.fields_by_name['model_versions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION
_LISTMODELVERSIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELVERSIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELVERSIONSRESPONSE.fields_by_name['model_versions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION
_TRAINMODELREQUEST.fields_by_name['system_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TRAINMODELREQUEST.fields_by_name['visible_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TRAINMODELREQUEST.fields_by_name['visible_stop_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GETTRAININGMETRICSRESPONSE.fields_by_name['training_metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TRAININGMETRICS
_LISTTRAININGMETRICSRESPONSE.fields_by_name['training_metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TRAININGMETRICS
_BATCHGETEXPERIMENTSRESPONSE.fields_by_name['experiments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT
_LISTEXPERIMENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEXPERIMENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEXPERIMENTSRESPONSE.fields_by_name['experiments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT
_CREATEBATCHPREDICTIONREQUEST.fields_by_name['source_type'].enum_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTIONSTORETYPE
_CREATEBATCHPREDICTIONREQUEST.fields_by_name['dest_type'].enum_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTIONSTORETYPE
_LISTBATCHPREDICTIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTBATCHPREDICTIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTBATCHPREDICTIONSRESPONSE.fields_by_name['batch_predictions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION
_BATCHGETBATCHPREDICTIONSRESPONSE.fields_by_name['batch_predictions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION
_GETSERVINGDATAREQUEST.fields_by_name['query'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._QUERY
_GETSERVINGDATAREQUEST.fields_by_name['event_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GETTRAININGDATAREQUEST.fields_by_name['query'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._QUERY
_BROWSEDATAWAREHOUSEREQUEST.fields_by_name['object_type'].enum_type = _BROWSEDATAWAREHOUSEREQUEST_OBJECTTYPE
_BROWSEDATAWAREHOUSEREQUEST_OBJECTTYPE.containing_type = _BROWSEDATAWAREHOUSEREQUEST
_BROWSEDATAWAREHOUSERESPONSE_TYPESENTRY.containing_type = _BROWSEDATAWAREHOUSERESPONSE
_BROWSEDATAWAREHOUSERESPONSE.fields_by_name['object_type'].enum_type = _BROWSEDATAWAREHOUSEREQUEST_OBJECTTYPE
_BROWSEDATAWAREHOUSERESPONSE.fields_by_name['types'].message_type = _BROWSEDATAWAREHOUSERESPONSE_TYPESENTRY
_GETMODELQUERYREQUEST.fields_by_name['model'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL
_GETMODELQUERYREQUEST.fields_by_name['additional_feature_sets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._FEATURESET
_GETSCHEMAGRAPHREQUEST.fields_by_name['additional_feature_sets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._FEATURESET
_GETSCHEMAGRAPHREQUEST.fields_by_name['additional_models'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL
_CREATEENVIRONMENTREQUEST.fields_by_name['data_store'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASTORE
_UPDATEENVIRONMENTREQUEST.fields_by_name['environment'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT
_UPDATEENVIRONMENTREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETENVIRONMENTSRESPONSE.fields_by_name['environments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT
_LISTENVIRONMENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTENVIRONMENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTENVIRONMENTSRESPONSE.fields_by_name['environments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT
DESCRIPTOR.message_types_by_name['UpdateUserRequest'] = _UPDATEUSERREQUEST
DESCRIPTOR.message_types_by_name['GetUserRequest'] = _GETUSERREQUEST
DESCRIPTOR.message_types_by_name['BatchGetUsersRequest'] = _BATCHGETUSERSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetUsersResponse'] = _BATCHGETUSERSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteUserRequest'] = _DELETEUSERREQUEST
DESCRIPTOR.message_types_by_name['ListUsersRequest'] = _LISTUSERSREQUEST
DESCRIPTOR.message_types_by_name['ListUsersResponse'] = _LISTUSERSRESPONSE
DESCRIPTOR.message_types_by_name['RegisterRequest'] = _REGISTERREQUEST
DESCRIPTOR.message_types_by_name['RegisterResponse'] = _REGISTERRESPONSE
DESCRIPTOR.message_types_by_name['LoginRequest'] = _LOGINREQUEST
DESCRIPTOR.message_types_by_name['LoginResponse'] = _LOGINRESPONSE
DESCRIPTOR.message_types_by_name['SendEmailVerificationRequest'] = _SENDEMAILVERIFICATIONREQUEST
DESCRIPTOR.message_types_by_name['VerifyEmailRequest'] = _VERIFYEMAILREQUEST
DESCRIPTOR.message_types_by_name['VerifyEmailResponse'] = _VERIFYEMAILRESPONSE
DESCRIPTOR.message_types_by_name['SendPasswordResetRequest'] = _SENDPASSWORDRESETREQUEST
DESCRIPTOR.message_types_by_name['GetPasswordResetRequest'] = _GETPASSWORDRESETREQUEST
DESCRIPTOR.message_types_by_name['ResetPasswordRequest'] = _RESETPASSWORDREQUEST
DESCRIPTOR.message_types_by_name['ResetPasswordResponse'] = _RESETPASSWORDRESPONSE
DESCRIPTOR.message_types_by_name['GetSessionRequest'] = _GETSESSIONREQUEST
DESCRIPTOR.message_types_by_name['ListSessionsRequest'] = _LISTSESSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListSessionsResponse'] = _LISTSESSIONSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteSessionRequest'] = _DELETESESSIONREQUEST
DESCRIPTOR.message_types_by_name['DeleteAllSessionsRequest'] = _DELETEALLSESSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListRolesRequest'] = _LISTROLESREQUEST
DESCRIPTOR.message_types_by_name['ListRolesResponse'] = _LISTROLESRESPONSE
DESCRIPTOR.message_types_by_name['GetRoleRequest'] = _GETROLEREQUEST
DESCRIPTOR.message_types_by_name['BatchGetRolesRequest'] = _BATCHGETROLESREQUEST
DESCRIPTOR.message_types_by_name['BatchGetRolesResponse'] = _BATCHGETROLESRESPONSE
DESCRIPTOR.message_types_by_name['ListAccessPoliciesRequest'] = _LISTACCESSPOLICIESREQUEST
DESCRIPTOR.message_types_by_name['ListAccessPoliciesResponse'] = _LISTACCESSPOLICIESRESPONSE
DESCRIPTOR.message_types_by_name['CreateAccessPolicyRequest'] = _CREATEACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['GetAccessPolicyRequest'] = _GETACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['UpdateAccessPolicyRequest'] = _UPDATEACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['DeleteAccessPolicyRequest'] = _DELETEACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['CreateInviteRequest'] = _CREATEINVITEREQUEST
DESCRIPTOR.message_types_by_name['GetInviteRequest'] = _GETINVITEREQUEST
DESCRIPTOR.message_types_by_name['ListInvitesRequest'] = _LISTINVITESREQUEST
DESCRIPTOR.message_types_by_name['ListInvitesResponse'] = _LISTINVITESRESPONSE
DESCRIPTOR.message_types_by_name['UpdateInviteRequest'] = _UPDATEINVITEREQUEST
DESCRIPTOR.message_types_by_name['DeleteInviteRequest'] = _DELETEINVITEREQUEST
DESCRIPTOR.message_types_by_name['AcceptInviteRequest'] = _ACCEPTINVITEREQUEST
DESCRIPTOR.message_types_by_name['ResendInviteRequest'] = _RESENDINVITEREQUEST
DESCRIPTOR.message_types_by_name['TestPermissionsRequest'] = _TESTPERMISSIONSREQUEST
DESCRIPTOR.message_types_by_name['TestPermissionsResponse'] = _TESTPERMISSIONSRESPONSE
DESCRIPTOR.message_types_by_name['CreateOrganizationRequest'] = _CREATEORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['UpdateOrganizationRequest'] = _UPDATEORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['GetOrganizationRequest'] = _GETORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetOrganizationsRequest'] = _BATCHGETORGANIZATIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetOrganizationsResponse'] = _BATCHGETORGANIZATIONSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteOrganizationRequest'] = _DELETEORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['ListOrganizationsRequest'] = _LISTORGANIZATIONSREQUEST
DESCRIPTOR.message_types_by_name['ListOrganizationsResponse'] = _LISTORGANIZATIONSRESPONSE
DESCRIPTOR.message_types_by_name['GetOnboardingRequest'] = _GETONBOARDINGREQUEST
DESCRIPTOR.message_types_by_name['UpdateOnboardingRequest'] = _UPDATEONBOARDINGREQUEST
DESCRIPTOR.message_types_by_name['CreateProjectRequest'] = _CREATEPROJECTREQUEST
DESCRIPTOR.message_types_by_name['UpdateProjectRequest'] = _UPDATEPROJECTREQUEST
DESCRIPTOR.message_types_by_name['GetProjectRequest'] = _GETPROJECTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetProjectsRequest'] = _BATCHGETPROJECTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetProjectsResponse'] = _BATCHGETPROJECTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteProjectRequest'] = _DELETEPROJECTREQUEST
DESCRIPTOR.message_types_by_name['ListProjectsRequest'] = _LISTPROJECTSREQUEST
DESCRIPTOR.message_types_by_name['ListProjectsResponse'] = _LISTPROJECTSRESPONSE
DESCRIPTOR.message_types_by_name['TestProjectDataStoreRequest'] = _TESTPROJECTDATASTOREREQUEST
DESCRIPTOR.message_types_by_name['TestProjectDataStoreResponse'] = _TESTPROJECTDATASTORERESPONSE
DESCRIPTOR.message_types_by_name['TestDataStoreRequest'] = _TESTDATASTOREREQUEST
DESCRIPTOR.message_types_by_name['TestDataStoreResponse'] = _TESTDATASTORERESPONSE
DESCRIPTOR.message_types_by_name['GetFeatureSetRequest'] = _GETFEATURESETREQUEST
DESCRIPTOR.message_types_by_name['CreateChangeRequest'] = _CREATECHANGEREQUEST
DESCRIPTOR.message_types_by_name['ReRunRequest'] = _RERUNREQUEST
DESCRIPTOR.message_types_by_name['GetChangeRequest'] = _GETCHANGEREQUEST
DESCRIPTOR.message_types_by_name['GetChangeResponse'] = _GETCHANGERESPONSE
DESCRIPTOR.message_types_by_name['ListChangesRequest'] = _LISTCHANGESREQUEST
DESCRIPTOR.message_types_by_name['ListChangesResponse'] = _LISTCHANGESRESPONSE
DESCRIPTOR.message_types_by_name['CancelChangeRequest'] = _CANCELCHANGEREQUEST
DESCRIPTOR.message_types_by_name['ListFeatureSetsRequest'] = _LISTFEATURESETSREQUEST
DESCRIPTOR.message_types_by_name['ListFeatureSetsResponse'] = _LISTFEATURESETSRESPONSE
DESCRIPTOR.message_types_by_name['GetEntityFeatureSetsRequest'] = _GETENTITYFEATURESETSREQUEST
DESCRIPTOR.message_types_by_name['GetEntityFeatureSetsResponse'] = _GETENTITYFEATURESETSRESPONSE
DESCRIPTOR.message_types_by_name['GetEntityRequest'] = _GETENTITYREQUEST
DESCRIPTOR.message_types_by_name['ListEntitiesRequest'] = _LISTENTITIESREQUEST
DESCRIPTOR.message_types_by_name['ListEntitiesResponse'] = _LISTENTITIESRESPONSE
DESCRIPTOR.message_types_by_name['BatchGetFeatureSetsRequest'] = _BATCHGETFEATURESETSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetFeatureSetsResponse'] = _BATCHGETFEATURESETSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteFeatureSetRequest'] = _DELETEFEATURESETREQUEST
DESCRIPTOR.message_types_by_name['ValidateFeatureSetRequest'] = _VALIDATEFEATURESETREQUEST
DESCRIPTOR.message_types_by_name['GetFeatureSetDatasetStatsRequest'] = _GETFEATURESETDATASETSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetModelDatasetStatsRequest'] = _GETMODELDATASETSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetModelVersionDatasetStatsRequest'] = _GETMODELVERSIONDATASETSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetBatchPredictionDatasetStatsRequest'] = _GETBATCHPREDICTIONDATASETSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetPredictionStatsRequest'] = _GETPREDICTIONSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetFeatureSetStatsRequest'] = _GETFEATURESETSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetModelStatsRequest'] = _GETMODELSTATSREQUEST
DESCRIPTOR.message_types_by_name['ComputeColumnStatsRequest'] = _COMPUTECOLUMNSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetDataRequest'] = _GETDATAREQUEST
DESCRIPTOR.message_types_by_name['GetDataResponse'] = _GETDATARESPONSE
DESCRIPTOR.message_types_by_name['GetUsageRequest'] = _GETUSAGEREQUEST
DESCRIPTOR.message_types_by_name['GetUsageResponse'] = _GETUSAGERESPONSE
DESCRIPTOR.message_types_by_name['ListPricesRequest'] = _LISTPRICESREQUEST
DESCRIPTOR.message_types_by_name['ListPricesResponse'] = _LISTPRICESRESPONSE
DESCRIPTOR.message_types_by_name['ListInvoicesRequest'] = _LISTINVOICESREQUEST
DESCRIPTOR.message_types_by_name['ListInvoicesResponse'] = _LISTINVOICESRESPONSE
DESCRIPTOR.message_types_by_name['GetPriceRequest'] = _GETPRICEREQUEST
DESCRIPTOR.message_types_by_name['GetPaymentMethodRequest'] = _GETPAYMENTMETHODREQUEST
DESCRIPTOR.message_types_by_name['GetUpcomingInvoiceRequest'] = _GETUPCOMINGINVOICEREQUEST
DESCRIPTOR.message_types_by_name['GetSubscriptionRequest'] = _GETSUBSCRIPTIONREQUEST
DESCRIPTOR.message_types_by_name['GetBillingAccountRequest'] = _GETBILLINGACCOUNTREQUEST
DESCRIPTOR.message_types_by_name['UpdateBillingAccountRequest'] = _UPDATEBILLINGACCOUNTREQUEST
DESCRIPTOR.message_types_by_name['UpdateSubscriptionRequest'] = _UPDATESUBSCRIPTIONREQUEST
DESCRIPTOR.message_types_by_name['CreateFeatureSetCSVSignedURLRequest'] = _CREATEFEATURESETCSVSIGNEDURLREQUEST
DESCRIPTOR.message_types_by_name['CreateFeatureSetCSVSignedURLResponse'] = _CREATEFEATURESETCSVSIGNEDURLRESPONSE
DESCRIPTOR.message_types_by_name['PromoteModelVersionRequest'] = _PROMOTEMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['ListPromotionsRequest'] = _LISTPROMOTIONSREQUEST
DESCRIPTOR.message_types_by_name['ListPromotionsResponse'] = _LISTPROMOTIONSRESPONSE
DESCRIPTOR.message_types_by_name['GetPromotionRequest'] = _GETPROMOTIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetPromotionsRequest'] = _BATCHGETPROMOTIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetPromotionsResponse'] = _BATCHGETPROMOTIONSRESPONSE
DESCRIPTOR.message_types_by_name['ListEventsRequest'] = _LISTEVENTSREQUEST
DESCRIPTOR.message_types_by_name['ListEventsResponse'] = _LISTEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['GetEventRequest'] = _GETEVENTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEventsRequest'] = _BATCHGETEVENTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEventsResponse'] = _BATCHGETEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['GetInferredSchemaRequest'] = _GETINFERREDSCHEMAREQUEST
DESCRIPTOR.message_types_by_name['GetInferredSchemaResponse'] = _GETINFERREDSCHEMARESPONSE
DESCRIPTOR.message_types_by_name['GetModelRequest'] = _GETMODELREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelsRequest'] = _BATCHGETMODELSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelsResponse'] = _BATCHGETMODELSRESPONSE
DESCRIPTOR.message_types_by_name['ListModelsRequest'] = _LISTMODELSREQUEST
DESCRIPTOR.message_types_by_name['ListModelsResponse'] = _LISTMODELSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteModelRequest'] = _DELETEMODELREQUEST
DESCRIPTOR.message_types_by_name['ValidateModelRequest'] = _VALIDATEMODELREQUEST
DESCRIPTOR.message_types_by_name['GetModelVersionRequest'] = _GETMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['GetModelVersionQueryRequest'] = _GETMODELVERSIONQUERYREQUEST
DESCRIPTOR.message_types_by_name['GetModelVersionQueryResponse'] = _GETMODELVERSIONQUERYRESPONSE
DESCRIPTOR.message_types_by_name['BatchGetModelVersionsRequest'] = _BATCHGETMODELVERSIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelVersionsResponse'] = _BATCHGETMODELVERSIONSRESPONSE
DESCRIPTOR.message_types_by_name['ListModelVersionsRequest'] = _LISTMODELVERSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListModelVersionsResponse'] = _LISTMODELVERSIONSRESPONSE
DESCRIPTOR.message_types_by_name['TrainModelRequest'] = _TRAINMODELREQUEST
DESCRIPTOR.message_types_by_name['CancelTrainingRequest'] = _CANCELTRAININGREQUEST
DESCRIPTOR.message_types_by_name['GetTrainingMetricsRequest'] = _GETTRAININGMETRICSREQUEST
DESCRIPTOR.message_types_by_name['GetTrainingMetricsResponse'] = _GETTRAININGMETRICSRESPONSE
DESCRIPTOR.message_types_by_name['ListTrainingMetricsRequest'] = _LISTTRAININGMETRICSREQUEST
DESCRIPTOR.message_types_by_name['ListTrainingMetricsResponse'] = _LISTTRAININGMETRICSRESPONSE
DESCRIPTOR.message_types_by_name['GetExperimentRequest'] = _GETEXPERIMENTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetExperimentsRequest'] = _BATCHGETEXPERIMENTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetExperimentsResponse'] = _BATCHGETEXPERIMENTSRESPONSE
DESCRIPTOR.message_types_by_name['ListExperimentsRequest'] = _LISTEXPERIMENTSREQUEST
DESCRIPTOR.message_types_by_name['ListExperimentsResponse'] = _LISTEXPERIMENTSRESPONSE
DESCRIPTOR.message_types_by_name['CreateBatchPredictionRequest'] = _CREATEBATCHPREDICTIONREQUEST
DESCRIPTOR.message_types_by_name['ListBatchPredictionsRequest'] = _LISTBATCHPREDICTIONSREQUEST
DESCRIPTOR.message_types_by_name['ListBatchPredictionsResponse'] = _LISTBATCHPREDICTIONSRESPONSE
DESCRIPTOR.message_types_by_name['BatchPredictionRequest'] = _BATCHPREDICTIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetBatchPredictionsRequest'] = _BATCHGETBATCHPREDICTIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetBatchPredictionsResponse'] = _BATCHGETBATCHPREDICTIONSRESPONSE
DESCRIPTOR.message_types_by_name['GetServingDataRequest'] = _GETSERVINGDATAREQUEST
DESCRIPTOR.message_types_by_name['GetServingDataResponse'] = _GETSERVINGDATARESPONSE
DESCRIPTOR.message_types_by_name['GetTrainingDataRequest'] = _GETTRAININGDATAREQUEST
DESCRIPTOR.message_types_by_name['GetTrainingDataResponse'] = _GETTRAININGDATARESPONSE
DESCRIPTOR.message_types_by_name['BrowseDataWarehouseRequest'] = _BROWSEDATAWAREHOUSEREQUEST
DESCRIPTOR.message_types_by_name['BrowseDataWarehouseResponse'] = _BROWSEDATAWAREHOUSERESPONSE
DESCRIPTOR.message_types_by_name['SeedRequest'] = _SEEDREQUEST
DESCRIPTOR.message_types_by_name['SeedResponse'] = _SEEDRESPONSE
DESCRIPTOR.message_types_by_name['YamlDiffRequest'] = _YAMLDIFFREQUEST
DESCRIPTOR.message_types_by_name['YamlDiffResponse'] = _YAMLDIFFRESPONSE
DESCRIPTOR.message_types_by_name['GetModelQueryRequest'] = _GETMODELQUERYREQUEST
DESCRIPTOR.message_types_by_name['GetModelQueryResponse'] = _GETMODELQUERYRESPONSE
DESCRIPTOR.message_types_by_name['GetSchemaGraphRequest'] = _GETSCHEMAGRAPHREQUEST
DESCRIPTOR.message_types_by_name['GetSchemaGraphResponse'] = _GETSCHEMAGRAPHRESPONSE
DESCRIPTOR.message_types_by_name['CreateEnvironmentRequest'] = _CREATEENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['UpdateEnvironmentRequest'] = _UPDATEENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['GetEnvironmentRequest'] = _GETENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEnvironmentsRequest'] = _BATCHGETENVIRONMENTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEnvironmentsResponse'] = _BATCHGETENVIRONMENTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteEnvironmentRequest'] = _DELETEENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['ListEnvironmentsRequest'] = _LISTENVIRONMENTSREQUEST
DESCRIPTOR.message_types_by_name['ListEnvironmentsResponse'] = _LISTENVIRONMENTSRESPONSE
DESCRIPTOR.message_types_by_name['TestEnvironmentDataStoreRequest'] = _TESTENVIRONMENTDATASTOREREQUEST
DESCRIPTOR.message_types_by_name['TestEnvironmentDataStoreResponse'] = _TESTENVIRONMENTDATASTORERESPONSE
DESCRIPTOR.message_types_by_name['GetSystemMessageRequest'] = _GETSYSTEMMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['GetSystemMessageResponse'] = _GETSYSTEMMESSAGERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UpdateUserRequest = _reflection.GeneratedProtocolMessageType('UpdateUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEUSERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateUserRequest)
  })
_sym_db.RegisterMessage(UpdateUserRequest)

GetUserRequest = _reflection.GeneratedProtocolMessageType('GetUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUserRequest)
  })
_sym_db.RegisterMessage(GetUserRequest)

BatchGetUsersRequest = _reflection.GeneratedProtocolMessageType('BatchGetUsersRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETUSERSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetUsersRequest)
  })
_sym_db.RegisterMessage(BatchGetUsersRequest)

BatchGetUsersResponse = _reflection.GeneratedProtocolMessageType('BatchGetUsersResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETUSERSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetUsersResponse)
  })
_sym_db.RegisterMessage(BatchGetUsersResponse)

DeleteUserRequest = _reflection.GeneratedProtocolMessageType('DeleteUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEUSERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteUserRequest)
  })
_sym_db.RegisterMessage(DeleteUserRequest)

ListUsersRequest = _reflection.GeneratedProtocolMessageType('ListUsersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUsersRequest)
  })
_sym_db.RegisterMessage(ListUsersRequest)

ListUsersResponse = _reflection.GeneratedProtocolMessageType('ListUsersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUsersResponse)
  })
_sym_db.RegisterMessage(ListUsersResponse)

RegisterRequest = _reflection.GeneratedProtocolMessageType('RegisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RegisterRequest)
  })
_sym_db.RegisterMessage(RegisterRequest)

RegisterResponse = _reflection.GeneratedProtocolMessageType('RegisterResponse', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RegisterResponse)
  })
_sym_db.RegisterMessage(RegisterResponse)

LoginRequest = _reflection.GeneratedProtocolMessageType('LoginRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGINREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.LoginRequest)
  })
_sym_db.RegisterMessage(LoginRequest)

LoginResponse = _reflection.GeneratedProtocolMessageType('LoginResponse', (_message.Message,), {
  'DESCRIPTOR' : _LOGINRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.LoginResponse)
  })
_sym_db.RegisterMessage(LoginResponse)

SendEmailVerificationRequest = _reflection.GeneratedProtocolMessageType('SendEmailVerificationRequest', (_message.Message,), {
  'DESCRIPTOR' : _SENDEMAILVERIFICATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SendEmailVerificationRequest)
  })
_sym_db.RegisterMessage(SendEmailVerificationRequest)

VerifyEmailRequest = _reflection.GeneratedProtocolMessageType('VerifyEmailRequest', (_message.Message,), {
  'DESCRIPTOR' : _VERIFYEMAILREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.VerifyEmailRequest)
  })
_sym_db.RegisterMessage(VerifyEmailRequest)

VerifyEmailResponse = _reflection.GeneratedProtocolMessageType('VerifyEmailResponse', (_message.Message,), {
  'DESCRIPTOR' : _VERIFYEMAILRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.VerifyEmailResponse)
  })
_sym_db.RegisterMessage(VerifyEmailResponse)

SendPasswordResetRequest = _reflection.GeneratedProtocolMessageType('SendPasswordResetRequest', (_message.Message,), {
  'DESCRIPTOR' : _SENDPASSWORDRESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SendPasswordResetRequest)
  })
_sym_db.RegisterMessage(SendPasswordResetRequest)

GetPasswordResetRequest = _reflection.GeneratedProtocolMessageType('GetPasswordResetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPASSWORDRESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPasswordResetRequest)
  })
_sym_db.RegisterMessage(GetPasswordResetRequest)

ResetPasswordRequest = _reflection.GeneratedProtocolMessageType('ResetPasswordRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESETPASSWORDREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResetPasswordRequest)
  })
_sym_db.RegisterMessage(ResetPasswordRequest)

ResetPasswordResponse = _reflection.GeneratedProtocolMessageType('ResetPasswordResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESETPASSWORDRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResetPasswordResponse)
  })
_sym_db.RegisterMessage(ResetPasswordResponse)

GetSessionRequest = _reflection.GeneratedProtocolMessageType('GetSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSESSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSessionRequest)
  })
_sym_db.RegisterMessage(GetSessionRequest)

ListSessionsRequest = _reflection.GeneratedProtocolMessageType('ListSessionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTSESSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListSessionsRequest)
  })
_sym_db.RegisterMessage(ListSessionsRequest)

ListSessionsResponse = _reflection.GeneratedProtocolMessageType('ListSessionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTSESSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListSessionsResponse)
  })
_sym_db.RegisterMessage(ListSessionsResponse)

DeleteSessionRequest = _reflection.GeneratedProtocolMessageType('DeleteSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETESESSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteSessionRequest)
  })
_sym_db.RegisterMessage(DeleteSessionRequest)

DeleteAllSessionsRequest = _reflection.GeneratedProtocolMessageType('DeleteAllSessionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEALLSESSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteAllSessionsRequest)
  })
_sym_db.RegisterMessage(DeleteAllSessionsRequest)

ListRolesRequest = _reflection.GeneratedProtocolMessageType('ListRolesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTROLESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListRolesRequest)
  })
_sym_db.RegisterMessage(ListRolesRequest)

ListRolesResponse = _reflection.GeneratedProtocolMessageType('ListRolesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTROLESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListRolesResponse)
  })
_sym_db.RegisterMessage(ListRolesResponse)

GetRoleRequest = _reflection.GeneratedProtocolMessageType('GetRoleRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETROLEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetRoleRequest)
  })
_sym_db.RegisterMessage(GetRoleRequest)

BatchGetRolesRequest = _reflection.GeneratedProtocolMessageType('BatchGetRolesRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETROLESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetRolesRequest)
  })
_sym_db.RegisterMessage(BatchGetRolesRequest)

BatchGetRolesResponse = _reflection.GeneratedProtocolMessageType('BatchGetRolesResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETROLESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetRolesResponse)
  })
_sym_db.RegisterMessage(BatchGetRolesResponse)

ListAccessPoliciesRequest = _reflection.GeneratedProtocolMessageType('ListAccessPoliciesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTACCESSPOLICIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListAccessPoliciesRequest)
  })
_sym_db.RegisterMessage(ListAccessPoliciesRequest)

ListAccessPoliciesResponse = _reflection.GeneratedProtocolMessageType('ListAccessPoliciesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTACCESSPOLICIESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListAccessPoliciesResponse)
  })
_sym_db.RegisterMessage(ListAccessPoliciesResponse)

CreateAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('CreateAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateAccessPolicyRequest)
  })
_sym_db.RegisterMessage(CreateAccessPolicyRequest)

GetAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('GetAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetAccessPolicyRequest)
  })
_sym_db.RegisterMessage(GetAccessPolicyRequest)

UpdateAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('UpdateAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateAccessPolicyRequest)
  })
_sym_db.RegisterMessage(UpdateAccessPolicyRequest)

DeleteAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('DeleteAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteAccessPolicyRequest)
  })
_sym_db.RegisterMessage(DeleteAccessPolicyRequest)

CreateInviteRequest = _reflection.GeneratedProtocolMessageType('CreateInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateInviteRequest)
  })
_sym_db.RegisterMessage(CreateInviteRequest)

GetInviteRequest = _reflection.GeneratedProtocolMessageType('GetInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetInviteRequest)
  })
_sym_db.RegisterMessage(GetInviteRequest)

ListInvitesRequest = _reflection.GeneratedProtocolMessageType('ListInvitesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVITESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvitesRequest)
  })
_sym_db.RegisterMessage(ListInvitesRequest)

ListInvitesResponse = _reflection.GeneratedProtocolMessageType('ListInvitesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVITESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvitesResponse)
  })
_sym_db.RegisterMessage(ListInvitesResponse)

UpdateInviteRequest = _reflection.GeneratedProtocolMessageType('UpdateInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateInviteRequest)
  })
_sym_db.RegisterMessage(UpdateInviteRequest)

DeleteInviteRequest = _reflection.GeneratedProtocolMessageType('DeleteInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteInviteRequest)
  })
_sym_db.RegisterMessage(DeleteInviteRequest)

AcceptInviteRequest = _reflection.GeneratedProtocolMessageType('AcceptInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _ACCEPTINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.AcceptInviteRequest)
  })
_sym_db.RegisterMessage(AcceptInviteRequest)

ResendInviteRequest = _reflection.GeneratedProtocolMessageType('ResendInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESENDINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResendInviteRequest)
  })
_sym_db.RegisterMessage(ResendInviteRequest)

TestPermissionsRequest = _reflection.GeneratedProtocolMessageType('TestPermissionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTPERMISSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestPermissionsRequest)
  })
_sym_db.RegisterMessage(TestPermissionsRequest)

TestPermissionsResponse = _reflection.GeneratedProtocolMessageType('TestPermissionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _TESTPERMISSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestPermissionsResponse)
  })
_sym_db.RegisterMessage(TestPermissionsResponse)

CreateOrganizationRequest = _reflection.GeneratedProtocolMessageType('CreateOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateOrganizationRequest)
  })
_sym_db.RegisterMessage(CreateOrganizationRequest)

UpdateOrganizationRequest = _reflection.GeneratedProtocolMessageType('UpdateOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateOrganizationRequest)
  })
_sym_db.RegisterMessage(UpdateOrganizationRequest)

GetOrganizationRequest = _reflection.GeneratedProtocolMessageType('GetOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOrganizationRequest)
  })
_sym_db.RegisterMessage(GetOrganizationRequest)

BatchGetOrganizationsRequest = _reflection.GeneratedProtocolMessageType('BatchGetOrganizationsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETORGANIZATIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetOrganizationsRequest)
  })
_sym_db.RegisterMessage(BatchGetOrganizationsRequest)

BatchGetOrganizationsResponse = _reflection.GeneratedProtocolMessageType('BatchGetOrganizationsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETORGANIZATIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetOrganizationsResponse)
  })
_sym_db.RegisterMessage(BatchGetOrganizationsResponse)

DeleteOrganizationRequest = _reflection.GeneratedProtocolMessageType('DeleteOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteOrganizationRequest)
  })
_sym_db.RegisterMessage(DeleteOrganizationRequest)

ListOrganizationsRequest = _reflection.GeneratedProtocolMessageType('ListOrganizationsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTORGANIZATIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListOrganizationsRequest)
  })
_sym_db.RegisterMessage(ListOrganizationsRequest)

ListOrganizationsResponse = _reflection.GeneratedProtocolMessageType('ListOrganizationsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTORGANIZATIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListOrganizationsResponse)
  })
_sym_db.RegisterMessage(ListOrganizationsResponse)

GetOnboardingRequest = _reflection.GeneratedProtocolMessageType('GetOnboardingRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETONBOARDINGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOnboardingRequest)
  })
_sym_db.RegisterMessage(GetOnboardingRequest)

UpdateOnboardingRequest = _reflection.GeneratedProtocolMessageType('UpdateOnboardingRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEONBOARDINGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateOnboardingRequest)
  })
_sym_db.RegisterMessage(UpdateOnboardingRequest)

CreateProjectRequest = _reflection.GeneratedProtocolMessageType('CreateProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateProjectRequest)
  })
_sym_db.RegisterMessage(CreateProjectRequest)

UpdateProjectRequest = _reflection.GeneratedProtocolMessageType('UpdateProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateProjectRequest)
  })
_sym_db.RegisterMessage(UpdateProjectRequest)

GetProjectRequest = _reflection.GeneratedProtocolMessageType('GetProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetProjectRequest)
  })
_sym_db.RegisterMessage(GetProjectRequest)

BatchGetProjectsRequest = _reflection.GeneratedProtocolMessageType('BatchGetProjectsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROJECTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetProjectsRequest)
  })
_sym_db.RegisterMessage(BatchGetProjectsRequest)

BatchGetProjectsResponse = _reflection.GeneratedProtocolMessageType('BatchGetProjectsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROJECTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetProjectsResponse)
  })
_sym_db.RegisterMessage(BatchGetProjectsResponse)

DeleteProjectRequest = _reflection.GeneratedProtocolMessageType('DeleteProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteProjectRequest)
  })
_sym_db.RegisterMessage(DeleteProjectRequest)

ListProjectsRequest = _reflection.GeneratedProtocolMessageType('ListProjectsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROJECTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListProjectsRequest)
  })
_sym_db.RegisterMessage(ListProjectsRequest)

ListProjectsResponse = _reflection.GeneratedProtocolMessageType('ListProjectsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROJECTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListProjectsResponse)
  })
_sym_db.RegisterMessage(ListProjectsResponse)

TestProjectDataStoreRequest = _reflection.GeneratedProtocolMessageType('TestProjectDataStoreRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTPROJECTDATASTOREREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestProjectDataStoreRequest)
  })
_sym_db.RegisterMessage(TestProjectDataStoreRequest)

TestProjectDataStoreResponse = _reflection.GeneratedProtocolMessageType('TestProjectDataStoreResponse', (_message.Message,), {
  'DESCRIPTOR' : _TESTPROJECTDATASTORERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestProjectDataStoreResponse)
  })
_sym_db.RegisterMessage(TestProjectDataStoreResponse)

TestDataStoreRequest = _reflection.GeneratedProtocolMessageType('TestDataStoreRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTDATASTOREREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestDataStoreRequest)
  })
_sym_db.RegisterMessage(TestDataStoreRequest)

TestDataStoreResponse = _reflection.GeneratedProtocolMessageType('TestDataStoreResponse', (_message.Message,), {
  'DESCRIPTOR' : _TESTDATASTORERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestDataStoreResponse)
  })
_sym_db.RegisterMessage(TestDataStoreResponse)

GetFeatureSetRequest = _reflection.GeneratedProtocolMessageType('GetFeatureSetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFEATURESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetFeatureSetRequest)
  })
_sym_db.RegisterMessage(GetFeatureSetRequest)

CreateChangeRequest = _reflection.GeneratedProtocolMessageType('CreateChangeRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATECHANGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateChangeRequest)
  })
_sym_db.RegisterMessage(CreateChangeRequest)

ReRunRequest = _reflection.GeneratedProtocolMessageType('ReRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _RERUNREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ReRunRequest)
  })
_sym_db.RegisterMessage(ReRunRequest)

GetChangeRequest = _reflection.GeneratedProtocolMessageType('GetChangeRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCHANGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetChangeRequest)
  })
_sym_db.RegisterMessage(GetChangeRequest)

GetChangeResponse = _reflection.GeneratedProtocolMessageType('GetChangeResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETCHANGERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetChangeResponse)
  })
_sym_db.RegisterMessage(GetChangeResponse)

ListChangesRequest = _reflection.GeneratedProtocolMessageType('ListChangesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTCHANGESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListChangesRequest)
  })
_sym_db.RegisterMessage(ListChangesRequest)

ListChangesResponse = _reflection.GeneratedProtocolMessageType('ListChangesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTCHANGESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListChangesResponse)
  })
_sym_db.RegisterMessage(ListChangesResponse)

CancelChangeRequest = _reflection.GeneratedProtocolMessageType('CancelChangeRequest', (_message.Message,), {
  'DESCRIPTOR' : _CANCELCHANGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CancelChangeRequest)
  })
_sym_db.RegisterMessage(CancelChangeRequest)

ListFeatureSetsRequest = _reflection.GeneratedProtocolMessageType('ListFeatureSetsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTFEATURESETSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListFeatureSetsRequest)
  })
_sym_db.RegisterMessage(ListFeatureSetsRequest)

ListFeatureSetsResponse = _reflection.GeneratedProtocolMessageType('ListFeatureSetsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTFEATURESETSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListFeatureSetsResponse)
  })
_sym_db.RegisterMessage(ListFeatureSetsResponse)

GetEntityFeatureSetsRequest = _reflection.GeneratedProtocolMessageType('GetEntityFeatureSetsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETENTITYFEATURESETSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEntityFeatureSetsRequest)
  })
_sym_db.RegisterMessage(GetEntityFeatureSetsRequest)

GetEntityFeatureSetsResponse = _reflection.GeneratedProtocolMessageType('GetEntityFeatureSetsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETENTITYFEATURESETSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEntityFeatureSetsResponse)
  })
_sym_db.RegisterMessage(GetEntityFeatureSetsResponse)

GetEntityRequest = _reflection.GeneratedProtocolMessageType('GetEntityRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETENTITYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEntityRequest)
  })
_sym_db.RegisterMessage(GetEntityRequest)

ListEntitiesRequest = _reflection.GeneratedProtocolMessageType('ListEntitiesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTENTITIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEntitiesRequest)
  })
_sym_db.RegisterMessage(ListEntitiesRequest)

ListEntitiesResponse = _reflection.GeneratedProtocolMessageType('ListEntitiesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTENTITIESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEntitiesResponse)
  })
_sym_db.RegisterMessage(ListEntitiesResponse)

BatchGetFeatureSetsRequest = _reflection.GeneratedProtocolMessageType('BatchGetFeatureSetsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETFEATURESETSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetFeatureSetsRequest)
  })
_sym_db.RegisterMessage(BatchGetFeatureSetsRequest)

BatchGetFeatureSetsResponse = _reflection.GeneratedProtocolMessageType('BatchGetFeatureSetsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETFEATURESETSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetFeatureSetsResponse)
  })
_sym_db.RegisterMessage(BatchGetFeatureSetsResponse)

DeleteFeatureSetRequest = _reflection.GeneratedProtocolMessageType('DeleteFeatureSetRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEFEATURESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteFeatureSetRequest)
  })
_sym_db.RegisterMessage(DeleteFeatureSetRequest)

ValidateFeatureSetRequest = _reflection.GeneratedProtocolMessageType('ValidateFeatureSetRequest', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATEFEATURESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ValidateFeatureSetRequest)
  })
_sym_db.RegisterMessage(ValidateFeatureSetRequest)

GetFeatureSetDatasetStatsRequest = _reflection.GeneratedProtocolMessageType('GetFeatureSetDatasetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFEATURESETDATASETSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetFeatureSetDatasetStatsRequest)
  })
_sym_db.RegisterMessage(GetFeatureSetDatasetStatsRequest)

GetModelDatasetStatsRequest = _reflection.GeneratedProtocolMessageType('GetModelDatasetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELDATASETSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelDatasetStatsRequest)
  })
_sym_db.RegisterMessage(GetModelDatasetStatsRequest)

GetModelVersionDatasetStatsRequest = _reflection.GeneratedProtocolMessageType('GetModelVersionDatasetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELVERSIONDATASETSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelVersionDatasetStatsRequest)
  })
_sym_db.RegisterMessage(GetModelVersionDatasetStatsRequest)

GetBatchPredictionDatasetStatsRequest = _reflection.GeneratedProtocolMessageType('GetBatchPredictionDatasetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETBATCHPREDICTIONDATASETSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetBatchPredictionDatasetStatsRequest)
  })
_sym_db.RegisterMessage(GetBatchPredictionDatasetStatsRequest)

GetPredictionStatsRequest = _reflection.GeneratedProtocolMessageType('GetPredictionStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPREDICTIONSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPredictionStatsRequest)
  })
_sym_db.RegisterMessage(GetPredictionStatsRequest)

GetFeatureSetStatsRequest = _reflection.GeneratedProtocolMessageType('GetFeatureSetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFEATURESETSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetFeatureSetStatsRequest)
  })
_sym_db.RegisterMessage(GetFeatureSetStatsRequest)

GetModelStatsRequest = _reflection.GeneratedProtocolMessageType('GetModelStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelStatsRequest)
  })
_sym_db.RegisterMessage(GetModelStatsRequest)

ComputeColumnStatsRequest = _reflection.GeneratedProtocolMessageType('ComputeColumnStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _COMPUTECOLUMNSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ComputeColumnStatsRequest)
  })
_sym_db.RegisterMessage(ComputeColumnStatsRequest)

GetDataRequest = _reflection.GeneratedProtocolMessageType('GetDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetDataRequest)
  })
_sym_db.RegisterMessage(GetDataRequest)

GetDataResponse = _reflection.GeneratedProtocolMessageType('GetDataResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETDATARESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetDataResponse)
  })
_sym_db.RegisterMessage(GetDataResponse)

GetUsageRequest = _reflection.GeneratedProtocolMessageType('GetUsageRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSAGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUsageRequest)
  })
_sym_db.RegisterMessage(GetUsageRequest)

GetUsageResponse = _reflection.GeneratedProtocolMessageType('GetUsageResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETUSAGERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUsageResponse)
  })
_sym_db.RegisterMessage(GetUsageResponse)

ListPricesRequest = _reflection.GeneratedProtocolMessageType('ListPricesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPRICESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPricesRequest)
  })
_sym_db.RegisterMessage(ListPricesRequest)

ListPricesResponse = _reflection.GeneratedProtocolMessageType('ListPricesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPRICESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPricesResponse)
  })
_sym_db.RegisterMessage(ListPricesResponse)

ListInvoicesRequest = _reflection.GeneratedProtocolMessageType('ListInvoicesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVOICESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvoicesRequest)
  })
_sym_db.RegisterMessage(ListInvoicesRequest)

ListInvoicesResponse = _reflection.GeneratedProtocolMessageType('ListInvoicesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVOICESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvoicesResponse)
  })
_sym_db.RegisterMessage(ListInvoicesResponse)

GetPriceRequest = _reflection.GeneratedProtocolMessageType('GetPriceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPRICEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPriceRequest)
  })
_sym_db.RegisterMessage(GetPriceRequest)

GetPaymentMethodRequest = _reflection.GeneratedProtocolMessageType('GetPaymentMethodRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPAYMENTMETHODREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPaymentMethodRequest)
  })
_sym_db.RegisterMessage(GetPaymentMethodRequest)

GetUpcomingInvoiceRequest = _reflection.GeneratedProtocolMessageType('GetUpcomingInvoiceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUPCOMINGINVOICEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUpcomingInvoiceRequest)
  })
_sym_db.RegisterMessage(GetUpcomingInvoiceRequest)

GetSubscriptionRequest = _reflection.GeneratedProtocolMessageType('GetSubscriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSUBSCRIPTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSubscriptionRequest)
  })
_sym_db.RegisterMessage(GetSubscriptionRequest)

GetBillingAccountRequest = _reflection.GeneratedProtocolMessageType('GetBillingAccountRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETBILLINGACCOUNTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetBillingAccountRequest)
  })
_sym_db.RegisterMessage(GetBillingAccountRequest)

UpdateBillingAccountRequest = _reflection.GeneratedProtocolMessageType('UpdateBillingAccountRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEBILLINGACCOUNTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateBillingAccountRequest)
  })
_sym_db.RegisterMessage(UpdateBillingAccountRequest)

UpdateSubscriptionRequest = _reflection.GeneratedProtocolMessageType('UpdateSubscriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESUBSCRIPTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateSubscriptionRequest)
  })
_sym_db.RegisterMessage(UpdateSubscriptionRequest)

CreateFeatureSetCSVSignedURLRequest = _reflection.GeneratedProtocolMessageType('CreateFeatureSetCSVSignedURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEFEATURESETCSVSIGNEDURLREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateFeatureSetCSVSignedURLRequest)
  })
_sym_db.RegisterMessage(CreateFeatureSetCSVSignedURLRequest)

CreateFeatureSetCSVSignedURLResponse = _reflection.GeneratedProtocolMessageType('CreateFeatureSetCSVSignedURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEFEATURESETCSVSIGNEDURLRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateFeatureSetCSVSignedURLResponse)
  })
_sym_db.RegisterMessage(CreateFeatureSetCSVSignedURLResponse)

PromoteModelVersionRequest = _reflection.GeneratedProtocolMessageType('PromoteModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _PROMOTEMODELVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PromoteModelVersionRequest)
  })
_sym_db.RegisterMessage(PromoteModelVersionRequest)

ListPromotionsRequest = _reflection.GeneratedProtocolMessageType('ListPromotionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROMOTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPromotionsRequest)
  })
_sym_db.RegisterMessage(ListPromotionsRequest)

ListPromotionsResponse = _reflection.GeneratedProtocolMessageType('ListPromotionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROMOTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPromotionsResponse)
  })
_sym_db.RegisterMessage(ListPromotionsResponse)

GetPromotionRequest = _reflection.GeneratedProtocolMessageType('GetPromotionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPROMOTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPromotionRequest)
  })
_sym_db.RegisterMessage(GetPromotionRequest)

BatchGetPromotionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetPromotionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROMOTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetPromotionsRequest)
  })
_sym_db.RegisterMessage(BatchGetPromotionsRequest)

BatchGetPromotionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetPromotionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROMOTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetPromotionsResponse)
  })
_sym_db.RegisterMessage(BatchGetPromotionsResponse)

ListEventsRequest = _reflection.GeneratedProtocolMessageType('ListEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTEVENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEventsRequest)
  })
_sym_db.RegisterMessage(ListEventsRequest)

ListEventsResponse = _reflection.GeneratedProtocolMessageType('ListEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTEVENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEventsResponse)
  })
_sym_db.RegisterMessage(ListEventsResponse)

GetEventRequest = _reflection.GeneratedProtocolMessageType('GetEventRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEVENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEventRequest)
  })
_sym_db.RegisterMessage(GetEventRequest)

BatchGetEventsRequest = _reflection.GeneratedProtocolMessageType('BatchGetEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEVENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEventsRequest)
  })
_sym_db.RegisterMessage(BatchGetEventsRequest)

BatchGetEventsResponse = _reflection.GeneratedProtocolMessageType('BatchGetEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEVENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEventsResponse)
  })
_sym_db.RegisterMessage(BatchGetEventsResponse)

GetInferredSchemaRequest = _reflection.GeneratedProtocolMessageType('GetInferredSchemaRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINFERREDSCHEMAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetInferredSchemaRequest)
  })
_sym_db.RegisterMessage(GetInferredSchemaRequest)

GetInferredSchemaResponse = _reflection.GeneratedProtocolMessageType('GetInferredSchemaResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETINFERREDSCHEMARESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetInferredSchemaResponse)
  })
_sym_db.RegisterMessage(GetInferredSchemaResponse)

GetModelRequest = _reflection.GeneratedProtocolMessageType('GetModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelRequest)
  })
_sym_db.RegisterMessage(GetModelRequest)

BatchGetModelsRequest = _reflection.GeneratedProtocolMessageType('BatchGetModelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelsRequest)
  })
_sym_db.RegisterMessage(BatchGetModelsRequest)

BatchGetModelsResponse = _reflection.GeneratedProtocolMessageType('BatchGetModelsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelsResponse)
  })
_sym_db.RegisterMessage(BatchGetModelsResponse)

ListModelsRequest = _reflection.GeneratedProtocolMessageType('ListModelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelsRequest)
  })
_sym_db.RegisterMessage(ListModelsRequest)

ListModelsResponse = _reflection.GeneratedProtocolMessageType('ListModelsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelsResponse)
  })
_sym_db.RegisterMessage(ListModelsResponse)

DeleteModelRequest = _reflection.GeneratedProtocolMessageType('DeleteModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteModelRequest)
  })
_sym_db.RegisterMessage(DeleteModelRequest)

ValidateModelRequest = _reflection.GeneratedProtocolMessageType('ValidateModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATEMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ValidateModelRequest)
  })
_sym_db.RegisterMessage(ValidateModelRequest)

GetModelVersionRequest = _reflection.GeneratedProtocolMessageType('GetModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelVersionRequest)
  })
_sym_db.RegisterMessage(GetModelVersionRequest)

GetModelVersionQueryRequest = _reflection.GeneratedProtocolMessageType('GetModelVersionQueryRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELVERSIONQUERYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelVersionQueryRequest)
  })
_sym_db.RegisterMessage(GetModelVersionQueryRequest)

GetModelVersionQueryResponse = _reflection.GeneratedProtocolMessageType('GetModelVersionQueryResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELVERSIONQUERYRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelVersionQueryResponse)
  })
_sym_db.RegisterMessage(GetModelVersionQueryResponse)

BatchGetModelVersionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetModelVersionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELVERSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelVersionsRequest)
  })
_sym_db.RegisterMessage(BatchGetModelVersionsRequest)

BatchGetModelVersionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetModelVersionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELVERSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelVersionsResponse)
  })
_sym_db.RegisterMessage(BatchGetModelVersionsResponse)

ListModelVersionsRequest = _reflection.GeneratedProtocolMessageType('ListModelVersionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELVERSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelVersionsRequest)
  })
_sym_db.RegisterMessage(ListModelVersionsRequest)

ListModelVersionsResponse = _reflection.GeneratedProtocolMessageType('ListModelVersionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELVERSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelVersionsResponse)
  })
_sym_db.RegisterMessage(ListModelVersionsResponse)

TrainModelRequest = _reflection.GeneratedProtocolMessageType('TrainModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRAINMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TrainModelRequest)
  })
_sym_db.RegisterMessage(TrainModelRequest)

CancelTrainingRequest = _reflection.GeneratedProtocolMessageType('CancelTrainingRequest', (_message.Message,), {
  'DESCRIPTOR' : _CANCELTRAININGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CancelTrainingRequest)
  })
_sym_db.RegisterMessage(CancelTrainingRequest)

GetTrainingMetricsRequest = _reflection.GeneratedProtocolMessageType('GetTrainingMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAININGMETRICSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetTrainingMetricsRequest)
  })
_sym_db.RegisterMessage(GetTrainingMetricsRequest)

GetTrainingMetricsResponse = _reflection.GeneratedProtocolMessageType('GetTrainingMetricsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAININGMETRICSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetTrainingMetricsResponse)
  })
_sym_db.RegisterMessage(GetTrainingMetricsResponse)

ListTrainingMetricsRequest = _reflection.GeneratedProtocolMessageType('ListTrainingMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTTRAININGMETRICSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListTrainingMetricsRequest)
  })
_sym_db.RegisterMessage(ListTrainingMetricsRequest)

ListTrainingMetricsResponse = _reflection.GeneratedProtocolMessageType('ListTrainingMetricsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTTRAININGMETRICSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListTrainingMetricsResponse)
  })
_sym_db.RegisterMessage(ListTrainingMetricsResponse)

GetExperimentRequest = _reflection.GeneratedProtocolMessageType('GetExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetExperimentRequest)
  })
_sym_db.RegisterMessage(GetExperimentRequest)

BatchGetExperimentsRequest = _reflection.GeneratedProtocolMessageType('BatchGetExperimentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEXPERIMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetExperimentsRequest)
  })
_sym_db.RegisterMessage(BatchGetExperimentsRequest)

BatchGetExperimentsResponse = _reflection.GeneratedProtocolMessageType('BatchGetExperimentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEXPERIMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetExperimentsResponse)
  })
_sym_db.RegisterMessage(BatchGetExperimentsResponse)

ListExperimentsRequest = _reflection.GeneratedProtocolMessageType('ListExperimentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTEXPERIMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListExperimentsRequest)
  })
_sym_db.RegisterMessage(ListExperimentsRequest)

ListExperimentsResponse = _reflection.GeneratedProtocolMessageType('ListExperimentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTEXPERIMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListExperimentsResponse)
  })
_sym_db.RegisterMessage(ListExperimentsResponse)

CreateBatchPredictionRequest = _reflection.GeneratedProtocolMessageType('CreateBatchPredictionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEBATCHPREDICTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateBatchPredictionRequest)
  })
_sym_db.RegisterMessage(CreateBatchPredictionRequest)

ListBatchPredictionsRequest = _reflection.GeneratedProtocolMessageType('ListBatchPredictionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTBATCHPREDICTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListBatchPredictionsRequest)
  })
_sym_db.RegisterMessage(ListBatchPredictionsRequest)

ListBatchPredictionsResponse = _reflection.GeneratedProtocolMessageType('ListBatchPredictionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTBATCHPREDICTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListBatchPredictionsResponse)
  })
_sym_db.RegisterMessage(ListBatchPredictionsResponse)

BatchPredictionRequest = _reflection.GeneratedProtocolMessageType('BatchPredictionRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHPREDICTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchPredictionRequest)
  })
_sym_db.RegisterMessage(BatchPredictionRequest)

BatchGetBatchPredictionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetBatchPredictionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETBATCHPREDICTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetBatchPredictionsRequest)
  })
_sym_db.RegisterMessage(BatchGetBatchPredictionsRequest)

BatchGetBatchPredictionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetBatchPredictionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETBATCHPREDICTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetBatchPredictionsResponse)
  })
_sym_db.RegisterMessage(BatchGetBatchPredictionsResponse)

GetServingDataRequest = _reflection.GeneratedProtocolMessageType('GetServingDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSERVINGDATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetServingDataRequest)
  })
_sym_db.RegisterMessage(GetServingDataRequest)

GetServingDataResponse = _reflection.GeneratedProtocolMessageType('GetServingDataResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSERVINGDATARESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetServingDataResponse)
  })
_sym_db.RegisterMessage(GetServingDataResponse)

GetTrainingDataRequest = _reflection.GeneratedProtocolMessageType('GetTrainingDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAININGDATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetTrainingDataRequest)
  })
_sym_db.RegisterMessage(GetTrainingDataRequest)

GetTrainingDataResponse = _reflection.GeneratedProtocolMessageType('GetTrainingDataResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAININGDATARESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetTrainingDataResponse)
  })
_sym_db.RegisterMessage(GetTrainingDataResponse)

BrowseDataWarehouseRequest = _reflection.GeneratedProtocolMessageType('BrowseDataWarehouseRequest', (_message.Message,), {
  'DESCRIPTOR' : _BROWSEDATAWAREHOUSEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BrowseDataWarehouseRequest)
  })
_sym_db.RegisterMessage(BrowseDataWarehouseRequest)

BrowseDataWarehouseResponse = _reflection.GeneratedProtocolMessageType('BrowseDataWarehouseResponse', (_message.Message,), {

  'TypesEntry' : _reflection.GeneratedProtocolMessageType('TypesEntry', (_message.Message,), {
    'DESCRIPTOR' : _BROWSEDATAWAREHOUSERESPONSE_TYPESENTRY,
    '__module__' : 'continual.rpc.management.v1.management_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BrowseDataWarehouseResponse.TypesEntry)
    })
  ,
  'DESCRIPTOR' : _BROWSEDATAWAREHOUSERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BrowseDataWarehouseResponse)
  })
_sym_db.RegisterMessage(BrowseDataWarehouseResponse)
_sym_db.RegisterMessage(BrowseDataWarehouseResponse.TypesEntry)

SeedRequest = _reflection.GeneratedProtocolMessageType('SeedRequest', (_message.Message,), {
  'DESCRIPTOR' : _SEEDREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SeedRequest)
  })
_sym_db.RegisterMessage(SeedRequest)

SeedResponse = _reflection.GeneratedProtocolMessageType('SeedResponse', (_message.Message,), {
  'DESCRIPTOR' : _SEEDRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SeedResponse)
  })
_sym_db.RegisterMessage(SeedResponse)

YamlDiffRequest = _reflection.GeneratedProtocolMessageType('YamlDiffRequest', (_message.Message,), {
  'DESCRIPTOR' : _YAMLDIFFREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.YamlDiffRequest)
  })
_sym_db.RegisterMessage(YamlDiffRequest)

YamlDiffResponse = _reflection.GeneratedProtocolMessageType('YamlDiffResponse', (_message.Message,), {
  'DESCRIPTOR' : _YAMLDIFFRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.YamlDiffResponse)
  })
_sym_db.RegisterMessage(YamlDiffResponse)

GetModelQueryRequest = _reflection.GeneratedProtocolMessageType('GetModelQueryRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELQUERYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelQueryRequest)
  })
_sym_db.RegisterMessage(GetModelQueryRequest)

GetModelQueryResponse = _reflection.GeneratedProtocolMessageType('GetModelQueryResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELQUERYRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelQueryResponse)
  })
_sym_db.RegisterMessage(GetModelQueryResponse)

GetSchemaGraphRequest = _reflection.GeneratedProtocolMessageType('GetSchemaGraphRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSCHEMAGRAPHREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSchemaGraphRequest)
  })
_sym_db.RegisterMessage(GetSchemaGraphRequest)

GetSchemaGraphResponse = _reflection.GeneratedProtocolMessageType('GetSchemaGraphResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSCHEMAGRAPHRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSchemaGraphResponse)
  })
_sym_db.RegisterMessage(GetSchemaGraphResponse)

CreateEnvironmentRequest = _reflection.GeneratedProtocolMessageType('CreateEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateEnvironmentRequest)
  })
_sym_db.RegisterMessage(CreateEnvironmentRequest)

UpdateEnvironmentRequest = _reflection.GeneratedProtocolMessageType('UpdateEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateEnvironmentRequest)
  })
_sym_db.RegisterMessage(UpdateEnvironmentRequest)

GetEnvironmentRequest = _reflection.GeneratedProtocolMessageType('GetEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEnvironmentRequest)
  })
_sym_db.RegisterMessage(GetEnvironmentRequest)

BatchGetEnvironmentsRequest = _reflection.GeneratedProtocolMessageType('BatchGetEnvironmentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETENVIRONMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEnvironmentsRequest)
  })
_sym_db.RegisterMessage(BatchGetEnvironmentsRequest)

BatchGetEnvironmentsResponse = _reflection.GeneratedProtocolMessageType('BatchGetEnvironmentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETENVIRONMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEnvironmentsResponse)
  })
_sym_db.RegisterMessage(BatchGetEnvironmentsResponse)

DeleteEnvironmentRequest = _reflection.GeneratedProtocolMessageType('DeleteEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteEnvironmentRequest)
  })
_sym_db.RegisterMessage(DeleteEnvironmentRequest)

ListEnvironmentsRequest = _reflection.GeneratedProtocolMessageType('ListEnvironmentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTENVIRONMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEnvironmentsRequest)
  })
_sym_db.RegisterMessage(ListEnvironmentsRequest)

ListEnvironmentsResponse = _reflection.GeneratedProtocolMessageType('ListEnvironmentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTENVIRONMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEnvironmentsResponse)
  })
_sym_db.RegisterMessage(ListEnvironmentsResponse)

TestEnvironmentDataStoreRequest = _reflection.GeneratedProtocolMessageType('TestEnvironmentDataStoreRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTENVIRONMENTDATASTOREREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestEnvironmentDataStoreRequest)
  })
_sym_db.RegisterMessage(TestEnvironmentDataStoreRequest)

TestEnvironmentDataStoreResponse = _reflection.GeneratedProtocolMessageType('TestEnvironmentDataStoreResponse', (_message.Message,), {
  'DESCRIPTOR' : _TESTENVIRONMENTDATASTORERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestEnvironmentDataStoreResponse)
  })
_sym_db.RegisterMessage(TestEnvironmentDataStoreResponse)

GetSystemMessageRequest = _reflection.GeneratedProtocolMessageType('GetSystemMessageRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMMESSAGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSystemMessageRequest)
  })
_sym_db.RegisterMessage(GetSystemMessageRequest)

GetSystemMessageResponse = _reflection.GeneratedProtocolMessageType('GetSystemMessageResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMMESSAGERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSystemMessageResponse)
  })
_sym_db.RegisterMessage(GetSystemMessageResponse)


DESCRIPTOR._options = None
_BROWSEDATAWAREHOUSERESPONSE_TYPESENTRY._options = None

_MANAGEMENTAPI = _descriptor.ServiceDescriptor(
  name='ManagementAPI',
  full_name='continual.rpc.management.v1.ManagementAPI',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=15839,
  serialized_end=36894,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetUser',
    full_name='continual.rpc.management.v1.ManagementAPI.GetUser',
    index=0,
    containing_service=None,
    input_type=_GETUSERREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/{name=users/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListUsers',
    full_name='continual.rpc.management.v1.ManagementAPI.ListUsers',
    index=1,
    containing_service=None,
    input_type=_LISTUSERSREQUEST,
    output_type=_LISTUSERSRESPONSE,
    serialized_options=b'\202\323\344\223\002\013\022\t/v1/users',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetUsers',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetUsers',
    index=2,
    containing_service=None,
    input_type=_BATCHGETUSERSREQUEST,
    output_type=_BATCHGETUSERSRESPONSE,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/users:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateUser',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateUser',
    index=3,
    containing_service=None,
    input_type=_UPDATEUSERREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER,
    serialized_options=b'\202\323\344\223\002\0372\027/v1/{user.name=users/*}:\004user',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteUser',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteUser',
    index=4,
    containing_service=None,
    input_type=_DELETEUSERREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\024*\022/v1/{name=users/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetViewer',
    full_name='continual.rpc.management.v1.ManagementAPI.GetViewer',
    index=5,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER,
    serialized_options=b'\202\323\344\223\002\014\022\n/v1/viewer',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Register',
    full_name='continual.rpc.management.v1.ManagementAPI.Register',
    index=6,
    containing_service=None,
    input_type=_REGISTERREQUEST,
    output_type=_REGISTERRESPONSE,
    serialized_options=b'\202\323\344\223\002\024\"\022/v1/users:register',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Login',
    full_name='continual.rpc.management.v1.ManagementAPI.Login',
    index=7,
    containing_service=None,
    input_type=_LOGINREQUEST,
    output_type=_LOGINRESPONSE,
    serialized_options=b'\202\323\344\223\002\021\"\017/v1/users:login',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Logout',
    full_name='continual.rpc.management.v1.ManagementAPI.Logout',
    index=8,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\022\"\020/v1/users:logout',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendEmailVerification',
    full_name='continual.rpc.management.v1.ManagementAPI.SendEmailVerification',
    index=9,
    containing_service=None,
    input_type=_SENDEMAILVERIFICATIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002!\"\037/v1/users:sendEmailVerification',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='VerifyEmail',
    full_name='continual.rpc.management.v1.ManagementAPI.VerifyEmail',
    index=10,
    containing_service=None,
    input_type=_VERIFYEMAILREQUEST,
    output_type=_VERIFYEMAILRESPONSE,
    serialized_options=b'\202\323\344\223\002\027\"\025/v1/users:verifyEmail',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendPasswordReset',
    full_name='continual.rpc.management.v1.ManagementAPI.SendPasswordReset',
    index=11,
    containing_service=None,
    input_type=_SENDPASSWORDRESETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\035\"\033/v1/users:sendPasswordReset',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResetPassword',
    full_name='continual.rpc.management.v1.ManagementAPI.ResetPassword',
    index=12,
    containing_service=None,
    input_type=_RESETPASSWORDREQUEST,
    output_type=_RESETPASSWORDRESPONSE,
    serialized_options=b'\202\323\344\223\002\032\"\030/v1/:users:resetPassword',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSession',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSession',
    index=13,
    containing_service=None,
    input_type=_GETSESSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SESSION,
    serialized_options=b'\202\323\344\223\002\037\022\035/v1/{name=users/*/sessions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListSessions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListSessions',
    index=14,
    containing_service=None,
    input_type=_LISTSESSIONSREQUEST,
    output_type=_LISTSESSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002\037\022\035/v1/{parent=users/*}/sessions\312E\010sessions\322E\004User',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteSession',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteSession',
    index=15,
    containing_service=None,
    input_type=_DELETESESSIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\037*\035/v1/{name=users/*/sessions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteAllSessions',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteAllSessions',
    index=16,
    containing_service=None,
    input_type=_DELETEALLSESSIONSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\037*\035/v1/{parent=users/*}/sessions',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.GetOrganization',
    index=17,
    containing_service=None,
    input_type=_GETORGANIZATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION,
    serialized_options=b'\202\323\344\223\002\034\022\032/v1/{name=organizations/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetOrganizations',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetOrganizations',
    index=18,
    containing_service=None,
    input_type=_BATCHGETORGANIZATIONSREQUEST,
    output_type=_BATCHGETORGANIZATIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002\034\022\032/v1/organizations:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListOrganizations',
    full_name='continual.rpc.management.v1.ManagementAPI.ListOrganizations',
    index=19,
    containing_service=None,
    input_type=_LISTORGANIZATIONSREQUEST,
    output_type=_LISTORGANIZATIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002\023\022\021/v1/organizations',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateOrganization',
    index=20,
    containing_service=None,
    input_type=_CREATEORGANIZATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION,
    serialized_options=b'\202\323\344\223\002!\"\021/v1/organizations:\014organization',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateOrganization',
    index=21,
    containing_service=None,
    input_type=_UPDATEORGANIZATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION,
    serialized_options=b'\202\323\344\223\00272\'/v1/{organization.name=organizations/*}:\014organization',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteOrganization',
    index=22,
    containing_service=None,
    input_type=_DELETEORGANIZATIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\034*\032/v1/{name=organizations/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetOnboarding',
    full_name='continual.rpc.management.v1.ManagementAPI.GetOnboarding',
    index=23,
    containing_service=None,
    input_type=_GETONBOARDINGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ONBOARDING,
    serialized_options=b'\202\323\344\223\002\'\022%/v1/{name=organizations/*/onboarding}\322E\014Organization\312E\nonboarding\332E\026name:name:$/onboarding',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateOnboarding',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateOnboarding',
    index=24,
    containing_service=None,
    input_type=_UPDATEONBOARDINGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ONBOARDING,
    serialized_options=b'\202\323\344\223\002>20/v1/{onboarding.name=organizations/*/onboarding}:\nonboarding',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListRoles',
    full_name='continual.rpc.management.v1.ManagementAPI.ListRoles',
    index=25,
    containing_service=None,
    input_type=_LISTROLESREQUEST,
    output_type=_LISTROLESRESPONSE,
    serialized_options=b'\202\323\344\223\002\013\022\t/v1/roles',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetRole',
    full_name='continual.rpc.management.v1.ManagementAPI.GetRole',
    index=26,
    containing_service=None,
    input_type=_GETROLEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ROLE,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/{name=roles/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetRoles',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetRoles',
    index=27,
    containing_service=None,
    input_type=_BATCHGETROLESREQUEST,
    output_type=_BATCHGETROLESRESPONSE,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/roles:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateAccessPolicy',
    index=28,
    containing_service=None,
    input_type=_CREATEACCESSPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY,
    serialized_options=b'\202\323\344\223\002/\"\036/v1/{parent=**}/accessPolicies:\raccess_policy',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.GetAccessPolicy',
    index=29,
    containing_service=None,
    input_type=_GETACCESSPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=**/accessPolicies/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListAccessPolicies',
    full_name='continual.rpc.management.v1.ManagementAPI.ListAccessPolicies',
    index=30,
    containing_service=None,
    input_type=_LISTACCESSPOLICIESREQUEST,
    output_type=_LISTACCESSPOLICIESRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=**}/accessPolicies\322EDOrganization,Project,FeatureSet,Model,ModelVersion,Experiment,Change\312E\016accessPolicies\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteAccessPolicy',
    index=31,
    containing_service=None,
    input_type=_DELETEACCESSPOLICYREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002 *\036/v1/{name=**/accessPolicies/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateAccessPolicy',
    index=32,
    containing_service=None,
    input_type=_UPDATEACCESSPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY,
    serialized_options=b'\202\323\344\223\002=2,/v1/{access_policy.name=**/accessPolicies/*}:\raccess_policy',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TestPermissions',
    full_name='continual.rpc.management.v1.ManagementAPI.TestPermissions',
    index=33,
    containing_service=None,
    input_type=_TESTPERMISSIONSREQUEST,
    output_type=_TESTPERMISSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002&\"!/v1/{resource=**}:testPermissions:\001*\322EDOrganization,Project,FeatureSet,Model,ModelVersion,Experiment,Change\312E\013permissions\332E\rname:resource',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateInvite',
    index=34,
    containing_service=None,
    input_type=_CREATEINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002\031\"\027/v1/{parent=**}/invites',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.GetInvite',
    index=35,
    containing_service=None,
    input_type=_GETINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002\031\022\027/v1/{name=**/invites/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListInvites',
    full_name='continual.rpc.management.v1.ManagementAPI.ListInvites',
    index=36,
    containing_service=None,
    input_type=_LISTINVITESREQUEST,
    output_type=_LISTINVITESRESPONSE,
    serialized_options=b'\202\323\344\223\002\031\022\027/v1/{parent=**}/invites',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateInvite',
    index=37,
    containing_service=None,
    input_type=_UPDATEINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002(2\036/v1/{invite.name=**/invites/*}:\006invite',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResendInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.ResendInvite',
    index=38,
    containing_service=None,
    input_type=_RESENDINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002 \"\036/v1/{name=**/invites/*}:resend',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteInvite',
    index=39,
    containing_service=None,
    input_type=_DELETEINVITEREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\031*\027/v1/{name=**/invites/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AcceptInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.AcceptInvite',
    index=40,
    containing_service=None,
    input_type=_ACCEPTINVITEREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\026\"\024/v1/confirm/{code=*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetProject',
    full_name='continual.rpc.management.v1.ManagementAPI.GetProject',
    index=41,
    containing_service=None,
    input_type=_GETPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=b'\202\323\344\223\002\027\022\025/v1/{name=projects/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetProjects',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetProjects',
    index=42,
    containing_service=None,
    input_type=_BATCHGETPROJECTSREQUEST,
    output_type=_BATCHGETPROJECTSRESPONSE,
    serialized_options=b'\202\323\344\223\002\027\022\025/v1/projects:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListProjects',
    full_name='continual.rpc.management.v1.ManagementAPI.ListProjects',
    index=43,
    containing_service=None,
    input_type=_LISTPROJECTSREQUEST,
    output_type=_LISTPROJECTSRESPONSE,
    serialized_options=b'\202\323\344\223\0027\022\014/v1/projectsZ\'\022%/v1/{parent=organizations/*}/projects\322E\014Organization\312E\010projects\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateProject',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateProject',
    index=44,
    containing_service=None,
    input_type=_CREATEPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=b'\202\323\344\223\002\027\"\014/v1/projects:\007project',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateProject',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateProject',
    index=45,
    containing_service=None,
    input_type=_UPDATEPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=b'\202\323\344\223\002(2\035/v1/{project.name=projects/*}:\007project',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteProject',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteProject',
    index=46,
    containing_service=None,
    input_type=_DELETEPROJECTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\027*\025/v1/{name=projects/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TestProjectDataStore',
    full_name='continual.rpc.management.v1.ManagementAPI.TestProjectDataStore',
    index=47,
    containing_service=None,
    input_type=_TESTPROJECTDATASTOREREQUEST,
    output_type=_TESTPROJECTDATASTORERESPONSE,
    serialized_options=b'\202\323\344\223\002%\"#/v1/{name=projects/*}:testDataStore',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TestDataStore',
    full_name='continual.rpc.management.v1.ManagementAPI.TestDataStore',
    index=48,
    containing_service=None,
    input_type=_TESTDATASTOREREQUEST,
    output_type=_TESTDATASTORERESPONSE,
    serialized_options=b'\202\323\344\223\002\027\"\025/v1/app:testDataStore',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSchemaGraph',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSchemaGraph',
    index=49,
    containing_service=None,
    input_type=_GETSCHEMAGRAPHREQUEST,
    output_type=_GETSCHEMAGRAPHRESPONSE,
    serialized_options=b'\322E\023Project,Environment\312E\013schemaGraph\332E\023name:name;name:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetFeatureSet',
    full_name='continual.rpc.management.v1.ManagementAPI.GetFeatureSet',
    index=50,
    containing_service=None,
    input_type=_GETFEATURESETREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._FEATURESET,
    serialized_options=b'\202\323\344\223\002%\022#/v1/{name=projects/*/featureSets/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetFeatureSets',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetFeatureSets',
    index=51,
    containing_service=None,
    input_type=_BATCHGETFEATURESETSREQUEST,
    output_type=_BATCHGETFEATURESETSRESPONSE,
    serialized_options=b'\202\323\344\223\002.\022,/v1/{parent=projects/*}/featureSets:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListFeatureSets',
    full_name='continual.rpc.management.v1.ManagementAPI.ListFeatureSets',
    index=52,
    containing_service=None,
    input_type=_LISTFEATURESETSREQUEST,
    output_type=_LISTFEATURESETSRESPONSE,
    serialized_options=b'\202\323\344\223\002%\022#/v1/{parent=projects/*}/featureSets\322E\023Project,Environment\312E\013featureSets\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteFeatureSet',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteFeatureSet',
    index=53,
    containing_service=None,
    input_type=_DELETEFEATURESETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002%*#/v1/{name=projects/*/featureSets/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ValidateFeatureSet',
    full_name='continual.rpc.management.v1.ManagementAPI.ValidateFeatureSet',
    index=54,
    containing_service=None,
    input_type=_VALIDATEFEATURESETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetEntity',
    full_name='continual.rpc.management.v1.ManagementAPI.GetEntity',
    index=55,
    containing_service=None,
    input_type=_GETENTITYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENTITY,
    serialized_options=b'\202\323\344\223\002\"\022 /v1/{name=projects/*/entities/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListEntities',
    full_name='continual.rpc.management.v1.ManagementAPI.ListEntities',
    index=56,
    containing_service=None,
    input_type=_LISTENTITIESREQUEST,
    output_type=_LISTENTITIESRESPONSE,
    serialized_options=b'\202\323\344\223\002\"\022 /v1/{parent=projects/*}:entities\322E\023Project,Environment\312E\010entities\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetEntityFeatureSets',
    full_name='continual.rpc.management.v1.ManagementAPI.GetEntityFeatureSets',
    index=57,
    containing_service=None,
    input_type=_GETENTITYFEATURESETSREQUEST,
    output_type=_GETENTITYFEATURESETSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetFeatureSetDatasetStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetFeatureSetDatasetStats',
    index=58,
    containing_service=None,
    input_type=_GETFEATURESETDATASETSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS,
    serialized_options=b'\202\323\344\223\0022\0220/v1/{name=projects/*/featureSets/*/datasetstats}\322E\nFeatureSet\312E\014datasetStats\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelDatasetStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelDatasetStats',
    index=59,
    containing_service=None,
    input_type=_GETMODELDATASETSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS,
    serialized_options=b'\202\323\344\223\002-\022+/v1/{name=projects/*/models/*/datasetstats}\322E\005Model\312E\014datasetStats\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelVersionDatasetStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelVersionDatasetStats',
    index=60,
    containing_service=None,
    input_type=_GETMODELVERSIONDATASETSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS,
    serialized_options=b'\202\323\344\223\0028\0226/v1/{name=projects/*/models/*/versions/*/datasetstats}\322E\014ModelVersion\312E\014datasetStats\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetBatchPredictionDatasetStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetBatchPredictionDatasetStats',
    index=61,
    containing_service=None,
    input_type=_GETBATCHPREDICTIONDATASETSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS,
    serialized_options=b'\202\323\344\223\002@\022>/v1/{name=projects/*/models/*/batchPredictions/*/datasetstats}\322E\017BatchPrediction\312E\014datasetStats\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPredictionStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPredictionStats',
    index=62,
    containing_service=None,
    input_type=_GETPREDICTIONSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS,
    serialized_options=b'\202\323\344\223\002C\022A/v1/{name=projects/*/models/*/batchPredictions/*/predictionstats}\322E\017BatchPrediction\312E\017predictionStats\332E\027name:name:$/predictions',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetFeatureSetStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetFeatureSetStats',
    index=63,
    containing_service=None,
    input_type=_GETFEATURESETSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._COLUMNSETSTATS,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{name=projects/*/featureSets/*/stats}\322E\nFeatureSet\312E\005stats\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelStats',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelStats',
    index=64,
    containing_service=None,
    input_type=_GETMODELSTATSREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._COLUMNSETSTATS,
    serialized_options=b'\202\323\344\223\002&\022$/v1/{name=projects/*/models/*/stats}\322E\005Model\312E\005stats\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ComputeColumnStats',
    full_name='continual.rpc.management.v1.ManagementAPI.ComputeColumnStats',
    index=65,
    containing_service=None,
    input_type=_COMPUTECOLUMNSTATSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002)\"\'/v1/{name=projects/*/*/*}/stats:compute',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetData',
    full_name='continual.rpc.management.v1.ManagementAPI.GetData',
    index=66,
    containing_service=None,
    input_type=_GETDATAREQUEST,
    output_type=_GETDATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateFeatureSetCSVSignedURL',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateFeatureSetCSVSignedURL',
    index=67,
    containing_service=None,
    input_type=_CREATEFEATURESETCSVSIGNEDURLREQUEST,
    output_type=_CREATEFEATURESETCSVSIGNEDURLRESPONSE,
    serialized_options=b'\202\323\344\223\002#\"!/v1/{parent=projects/*}:uploadCsv',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateChange',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateChange',
    index=68,
    containing_service=None,
    input_type=_CREATECHANGEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._CHANGE,
    serialized_options=b'\202\323\344\223\002 \"\036/v1/{parent=projects/*}:change',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ReRun',
    full_name='continual.rpc.management.v1.ManagementAPI.ReRun',
    index=69,
    containing_service=None,
    input_type=_RERUNREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._CHANGE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetChange',
    full_name='continual.rpc.management.v1.ManagementAPI.GetChange',
    index=70,
    containing_service=None,
    input_type=_GETCHANGEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._CHANGE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=projects/*}/change/*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListChanges',
    full_name='continual.rpc.management.v1.ManagementAPI.ListChanges',
    index=71,
    containing_service=None,
    input_type=_LISTCHANGESREQUEST,
    output_type=_LISTCHANGESRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=projects/*}/change\312E\005plans\322E\023Project,Environment\340E\000\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CancelChange',
    full_name='continual.rpc.management.v1.ManagementAPI.CancelChange',
    index=72,
    containing_service=None,
    input_type=_CANCELCHANGEREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\'\"%/v1/{name=projects/*/change/*}:cancel',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PromoteModelVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.PromoteModelVersion',
    index=73,
    containing_service=None,
    input_type=_PROMOTEMODELVERSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION,
    serialized_options=b'\202\323\344\223\0023\0221/v1/{name=projects/*/models/*/versions/*}:promote',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListPromotions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListPromotions',
    index=74,
    containing_service=None,
    input_type=_LISTPROMOTIONSREQUEST,
    output_type=_LISTPROMOTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002g\022+/v1/{parent=projects/*/models/*}/promotionsZ8\0226/v1/{parent=projects/*/models/*/versions/*}/promotions\312E\npromotions\322E\022ModelVersion,Model\340E\000\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPromotion',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPromotion',
    index=75,
    containing_service=None,
    input_type=_GETPROMOTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION,
    serialized_options=b'\202\323\344\223\002-\022+/v1/{name=projects/*/models/*/promotions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetPromotions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetPromotions',
    index=76,
    containing_service=None,
    input_type=_BATCHGETPROMOTIONSREQUEST,
    output_type=_BATCHGETPROMOTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\0026\0224/v1/{parent=projects/*/models/*}/promotions:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListEvents',
    full_name='continual.rpc.management.v1.ManagementAPI.ListEvents',
    index=77,
    containing_service=None,
    input_type=_LISTEVENTSREQUEST,
    output_type=_LISTEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=projects/*}:events',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetEvent',
    full_name='continual.rpc.management.v1.ManagementAPI.GetEvent',
    index=78,
    containing_service=None,
    input_type=_GETEVENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=projects/*/events/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetEvents',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetEvents',
    index=79,
    containing_service=None,
    input_type=_BATCHGETEVENTSREQUEST,
    output_type=_BATCHGETEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002)\022\'/v1/{parent=projects/*}/events:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSubscription',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSubscription',
    index=80,
    containing_service=None,
    input_type=_GETSUBSCRIPTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SUBSCRIPTION,
    serialized_options=b'\202\323\344\223\002\'\022%/v1/{name=organizations/subscription}\322E\014Organization\312E\014subscription\332E\030name:name:$/subscription',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetBillingAccount',
    full_name='continual.rpc.management.v1.ManagementAPI.GetBillingAccount',
    index=81,
    containing_service=None,
    input_type=_GETBILLINGACCOUNTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BILLINGACCOUNT,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{name=organizations/*/billingAccount}\322E\014Organization\312E\016billingAccount\332E\032name:name:$/billingAccount',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUsage',
    full_name='continual.rpc.management.v1.ManagementAPI.GetUsage',
    index=82,
    containing_service=None,
    input_type=_GETUSAGEREQUEST,
    output_type=_GETUSAGERESPONSE,
    serialized_options=b'\202\323\344\223\002$\"\"/v1/{parent=organizations/*}:usage\322E\014Organization\312E\005usage\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListInvoices',
    full_name='continual.rpc.management.v1.ManagementAPI.ListInvoices',
    index=83,
    containing_service=None,
    input_type=_LISTINVOICESREQUEST,
    output_type=_LISTINVOICESRESPONSE,
    serialized_options=b'\202\323\344\223\002\'\022%/v1/{parent=organizations/*}/invoices\322E\014Organization\312E\010invoices\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateBillingAccount',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateBillingAccount',
    index=84,
    containing_service=None,
    input_type=_UPDATEBILLINGACCOUNTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BILLINGACCOUNT,
    serialized_options=b'\202\323\344\223\002?2,/v1/{billing_account.name=**/billingAccount}:\017billing_account',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateSubscription',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateSubscription',
    index=85,
    containing_service=None,
    input_type=_UPDATESUBSCRIPTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SUBSCRIPTION,
    serialized_options=b'\202\323\344\223\00272\'/v1/{subscription.name=**/subscription}:\014subscription',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListPrices',
    full_name='continual.rpc.management.v1.ManagementAPI.ListPrices',
    index=86,
    containing_service=None,
    input_type=_LISTPRICESREQUEST,
    output_type=_LISTPRICESRESPONSE,
    serialized_options=b'\202\323\344\223\002%\022#/v1/{parent=organizations/*}/prices\322E\014Organization\312E\006prices\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPrice',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPrice',
    index=87,
    containing_service=None,
    input_type=_GETPRICEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PRICE,
    serialized_options=b'\202\323\344\223\002%\022#/v1/{name=organizations/*/prices/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPaymentMethod',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPaymentMethod',
    index=88,
    containing_service=None,
    input_type=_GETPAYMENTMETHODREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PAYMENTMETHOD,
    serialized_options=b'\202\323\344\223\002-\022+/v1/{name=organizations/*/paymentMethods/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUpcomingInvoice',
    full_name='continual.rpc.management.v1.ManagementAPI.GetUpcomingInvoice',
    index=89,
    containing_service=None,
    input_type=_GETUPCOMINGINVOICEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVOICE,
    serialized_options=b'\202\323\344\223\002,\022*/v1/{name=organizations/*/upcomingInvoice}\322E\014Organization\312E\017upcomingInvoice\332E\033name:name:$/upcomingInvoice',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetInferredSchema',
    full_name='continual.rpc.management.v1.ManagementAPI.GetInferredSchema',
    index=90,
    containing_service=None,
    input_type=_GETINFERREDSCHEMAREQUEST,
    output_type=_GETINFERREDSCHEMARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetServingData',
    full_name='continual.rpc.management.v1.ManagementAPI.GetServingData',
    index=91,
    containing_service=None,
    input_type=_GETSERVINGDATAREQUEST,
    output_type=_GETSERVINGDATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetTrainingData',
    full_name='continual.rpc.management.v1.ManagementAPI.GetTrainingData',
    index=92,
    containing_service=None,
    input_type=_GETTRAININGDATAREQUEST,
    output_type=_GETTRAININGDATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BrowseDataWarehouse',
    full_name='continual.rpc.management.v1.ManagementAPI.BrowseDataWarehouse',
    index=93,
    containing_service=None,
    input_type=_BROWSEDATAWAREHOUSEREQUEST,
    output_type=_BROWSEDATAWAREHOUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Seed',
    full_name='continual.rpc.management.v1.ManagementAPI.Seed',
    index=94,
    containing_service=None,
    input_type=_SEEDREQUEST,
    output_type=_SEEDRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModel',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModel',
    index=95,
    containing_service=None,
    input_type=_GETMODELREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=projects/*/models/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetModels',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetModels',
    index=96,
    containing_service=None,
    input_type=_BATCHGETMODELSREQUEST,
    output_type=_BATCHGETMODELSRESPONSE,
    serialized_options=b'\202\323\344\223\002)\022\'/v1/{parent=projects/*}/models:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListModels',
    full_name='continual.rpc.management.v1.ManagementAPI.ListModels',
    index=97,
    containing_service=None,
    input_type=_LISTMODELSREQUEST,
    output_type=_LISTMODELSRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=projects/*}/models\322E\023Project,Environment\312E\006models\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteModel',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteModel',
    index=98,
    containing_service=None,
    input_type=_DELETEMODELREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002 *\036/v1/{name=projects/*/models/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ValidateModel',
    full_name='continual.rpc.management.v1.ManagementAPI.ValidateModel',
    index=99,
    containing_service=None,
    input_type=_VALIDATEMODELREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelQuery',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelQuery',
    index=100,
    containing_service=None,
    input_type=_GETMODELQUERYREQUEST,
    output_type=_GETMODELQUERYRESPONSE,
    serialized_options=b'\322E\005Model\312E\nmodelQuery\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelVersion',
    index=101,
    containing_service=None,
    input_type=_GETMODELVERSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{name=projects/*/models/*/versions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelVersionQuery',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelVersionQuery',
    index=102,
    containing_service=None,
    input_type=_GETMODELVERSIONQUERYREQUEST,
    output_type=_GETMODELVERSIONQUERYRESPONSE,
    serialized_options=b'\202\323\344\223\0024\0222/v1/{name=projects/*/models/*/versions/*}:getQuery',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetModelVersions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetModelVersions',
    index=103,
    containing_service=None,
    input_type=_BATCHGETMODELVERSIONSREQUEST,
    output_type=_BATCHGETMODELVERSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\0024\0222/v1/{parent=projects/*/models/*}/versions:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListModelVersions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListModelVersions',
    index=104,
    containing_service=None,
    input_type=_LISTMODELVERSIONSREQUEST,
    output_type=_LISTMODELVERSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{parent=projects/*/models/*}/versions\312E\rmodelVersions\322E\031Model,Project,Environment\340E\000\332E9name:parent;name:parent:$/models/-;name:parent:$/models/-',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TrainModel',
    full_name='continual.rpc.management.v1.ManagementAPI.TrainModel',
    index=105,
    containing_service=None,
    input_type=_TRAINMODELREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION,
    serialized_options=b'\202\323\344\223\0021\"//v1/{name=projects/*/models/*/versions/*}:train',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CancelTraining',
    full_name='continual.rpc.management.v1.ManagementAPI.CancelTraining',
    index=106,
    containing_service=None,
    input_type=_CANCELTRAININGREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002:\"8/v1/{name=projects/*/models/*/versions/*}:cancelTraining',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchPredict',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchPredict',
    index=107,
    containing_service=None,
    input_type=_CREATEBATCHPREDICTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION,
    serialized_options=b'\202\323\344\223\002.\",/v1/{model=projects/*/models/*}:batchPredict',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetBatchPrediction',
    full_name='continual.rpc.management.v1.ManagementAPI.GetBatchPrediction',
    index=108,
    containing_service=None,
    input_type=_BATCHPREDICTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION,
    serialized_options=b'\202\323\344\223\0023\0221/v1/{name=projects/*/models/*/batchPredictions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetBatchPredictions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetBatchPredictions',
    index=109,
    containing_service=None,
    input_type=_BATCHGETBATCHPREDICTIONSREQUEST,
    output_type=_BATCHGETBATCHPREDICTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002<\022:/v1/{parent=projects/*/models/*}/batchPredictions:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListBatchPredictions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListBatchPredictions',
    index=110,
    containing_service=None,
    input_type=_LISTBATCHPREDICTIONSREQUEST,
    output_type=_LISTBATCHPREDICTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\0023\0221/v1/{parent=projects/*/models/*}/batchPredictions\312E\020batchPredictions\322E\031Project,Model,Environment\340E\000\332E#name:parent;name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CancelBatchPrediction',
    full_name='continual.rpc.management.v1.ManagementAPI.CancelBatchPrediction',
    index=111,
    containing_service=None,
    input_type=_BATCHPREDICTIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002Q\"O/v1/{name=projects/*/models/*/versions/*/batchPredictions/*}:cancelBatchPredict',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetTrainingMetrics',
    full_name='continual.rpc.management.v1.ManagementAPI.GetTrainingMetrics',
    index=112,
    containing_service=None,
    input_type=_GETTRAININGMETRICSREQUEST,
    output_type=_GETTRAININGMETRICSRESPONSE,
    serialized_options=b'\202\323\344\223\002D\022B/v1/{experiment=projects/*/models/*/versions/*}:getTrainingMetrics',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListTrainingMetrics',
    full_name='continual.rpc.management.v1.ManagementAPI.ListTrainingMetrics',
    index=113,
    containing_service=None,
    input_type=_LISTTRAININGMETRICSREQUEST,
    output_type=_LISTTRAININGMETRICSRESPONSE,
    serialized_options=b'\202\323\344\223\002@\022>/v1/{model=projects/*/models/*/versions/*}:listTrainingMetrics',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetExperiment',
    full_name='continual.rpc.management.v1.ManagementAPI.GetExperiment',
    index=114,
    containing_service=None,
    input_type=_GETEXPERIMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT,
    serialized_options=b'\202\323\344\223\0029\0227/v1/{name=projects/*/models/*/versions/*/experiments/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetExperiments',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetExperiments',
    index=115,
    containing_service=None,
    input_type=_BATCHGETEXPERIMENTSREQUEST,
    output_type=_BATCHGETEXPERIMENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002B\022@/v1/{parent=projects/*/models/*/versions/*}/experiments:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListExperiments',
    full_name='continual.rpc.management.v1.ManagementAPI.ListExperiments',
    index=116,
    containing_service=None,
    input_type=_LISTEXPERIMENTSREQUEST,
    output_type=_LISTEXPERIMENTSRESPONSE,
    serialized_options=b'\202\323\344\223\0029\0227/v1/{parent=projects/*/models/*/versions/*}/experiments\312E\013experiments\322E\014ModelVersion\340E\000\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetYamlDiff',
    full_name='continual.rpc.management.v1.ManagementAPI.GetYamlDiff',
    index=117,
    containing_service=None,
    input_type=_YAMLDIFFREQUEST,
    output_type=_YAMLDIFFRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.GetEnvironment',
    index=118,
    containing_service=None,
    input_type=_GETENVIRONMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetEnvironments',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetEnvironments',
    index=119,
    containing_service=None,
    input_type=_BATCHGETENVIRONMENTSREQUEST,
    output_type=_BATCHGETENVIRONMENTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListEnvironments',
    full_name='continual.rpc.management.v1.ManagementAPI.ListEnvironments',
    index=120,
    containing_service=None,
    input_type=_LISTENVIRONMENTSREQUEST,
    output_type=_LISTENVIRONMENTSRESPONSE,
    serialized_options=b'\322E\007Project\312E\014environments\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateEnvironment',
    index=121,
    containing_service=None,
    input_type=_CREATEENVIRONMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateEnvironment',
    index=122,
    containing_service=None,
    input_type=_UPDATEENVIRONMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteEnvironment',
    index=123,
    containing_service=None,
    input_type=_DELETEENVIRONMENTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TestEnvironmentDataStore',
    full_name='continual.rpc.management.v1.ManagementAPI.TestEnvironmentDataStore',
    index=124,
    containing_service=None,
    input_type=_TESTENVIRONMENTDATASTOREREQUEST,
    output_type=_TESTENVIRONMENTDATASTORERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSystemMessage',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSystemMessage',
    index=125,
    containing_service=None,
    input_type=_GETSYSTEMMESSAGEREQUEST,
    output_type=_GETSYSTEMMESSAGERESPONSE,
    serialized_options=b'\202\323\344\223\002\023\"\021/v1/systemMessage',
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_MANAGEMENTAPI)

DESCRIPTOR.services_by_name['ManagementAPI'] = _MANAGEMENTAPI

# @@protoc_insertion_point(module_scope)
